/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.compiler;

import java.util.ArrayList;
import java.util.List;

import antlr.Token;
import antlr.TokenStream;

import at.dms.util.MessageDescription;
import at.dms.util.Utils;

/**
 * This class describes the capabilities of lexers (token streams)
 * to share a common input buffer and to handle line numbers.
 */
public abstract class Scanner
    implements TokenStream
{
    /**
     * Special token to indicate end-of-file condition.
     */
    private static final CToken  EOF = new CToken(1, "End of file");

    /**
     * ...
     */
    private final Compiler compiler;

    /**
     * The data source for this scanner.
     */
    protected final InputBuffer buffer;

    /**
     * ...
     */
    private final List comments;

    /**
     * Constructs a new hand written scanner.
     */
    protected Scanner(Compiler compiler, InputBuffer buffer)
    {
        this.compiler = compiler;
        this.buffer = buffer;
        this.comments = new ArrayList();
    }

    /**
     * Returns the input buffer.
     */
    public final InputBuffer getBuffer()
    {
        return buffer;
    }

    /**
     * Returns the name of the source code file.
     */
    public final String getFile()
    {
        return buffer.file;
    }

    /**
     * Returns the current line number in the source code.
     */
    public final int getLine()
    {
        return buffer.line;
    }

    /**
     * Sets the current line number in the source code.
     */
    public final void setLine(int line)
    {
        buffer.line = line;
    }

    /**
     * Returns the current position in the input stream.
     */
    public final int getPosition()
    {
        return buffer.pos;
    }

    /**
     * Sets the current position in the input stream.
     */
    public final void setPosition(int pos)
    {
        buffer.pos = pos;
    }

    /**
     * Returns the reference of the current token in the source file.
     */
    public final TokenReference getTokenReference()
    {
        return TokenReference.build(buffer.file, buffer.line);
    }

    /**
     * Reports that an error has been detected in the lexical analyzer.
     * The handling is delegated to the compiler driver.
     *
     * @param error
     * The error to report.
     */
    protected final void reportTrouble(PositionedError trouble)
    {
        compiler.reportTrouble(trouble);
    }

    /**
     * Reports that an error has been detected in the lexical analyzer.
     * The handling is delegated to the compiler driver.
     *
     * @param mess    the error message.
     * @param param1  the first message parameter.
     * @param param2  the second message parameter.
     */
    protected final void reportTrouble(MessageDescription mess,
                                       Object param1, Object param2)
    {
        reportTrouble(new PositionedError(getTokenReference(),
                                          mess, param1, param2));
    }

    /**
     * Returns the next token in the input.
     */
    protected abstract Token getToken()
        throws PositionedError;

    /**
     *  Identifiers/Keywords/true/false/null (start with java letter)
     *  numeric literal (start with number)
     *  character literal (start with single quote)
     *  string (start with double quote)
     *  separator (parens, braces, brackets, semicolon, comma, period)
     *  operator (equals, plus, minus, etc)
     *  whitespace
     *  comment (start with slash)
     */
    public final Token nextToken()
    {
        while (true) {
            if (buffer.isEndOfData()) {
                return EOF;
            }
            else {
                switch (buffer.data[buffer.pos]) {

                    // White space:
                case '\n': // LineTerminator NL
                    buffer.line++;
                    buffer.pos++;
                    continue;

                case '\r': // LineTerminator CR
                    // No need to handle CR-NL separatly here:
                    // NL will be caught in the next loop cycle
                case ' ': // ASCII SP
                case '\t': // ASCII HT
                case '\f': // ASCII FF
                    buffer.pos++;
                    continue;
   
                    // EOF character:
                case '\020': // ASCII SUB
                    buffer.pos++;
                    return EOF;
   
                    // else, a Token
                default:
                    try {
                        Token tok = getToken();

                        if (tok != null) {
                            return tok;
                        }
                        continue;
                    }
                    catch (PositionedError e) {
                        reportTrouble(e);
                        buffer.pos++;
                    }
                    catch (Throwable e) {
                        if (buffer.pos < buffer.size) {
                            reportTrouble(CompilerMessages
                                          .NO_VIABLE_ALT_FOR_CHAR,
                                          String.valueOf
                                          (buffer.data[buffer.pos]),
                                          null);
                        }
                        else {
                            // not a very pretty way of handling unexpected EOF
                            reportTrouble(CompilerMessages.UNEXPECTED_EOF,
                                          null, null);
                            return EOF;
                        }
                        buffer.pos++;
                    }
                }
            }
        }
    }


    /**
     * Scans an escape sequence.
     * NOTE: scanXxx procedures do not return tokens.
     *
     * @return the scanned character.
     */
    protected char scanEscapeSequence()
        throws PositionedError
    {
        if (buffer.data[buffer.pos++] != '\\') {
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR, "\\\\");
        }

        switch (buffer.data[buffer.pos]) {
        case 'b':
            buffer.pos++; return '\b';
        case 't':
            buffer.pos++; return '\t';
        case 'n':
            buffer.pos++; return '\n';
        case 'f':
            buffer.pos++; return '\f';
        case 'r':
            buffer.pos++; return '\r';
        case '\"':
            buffer.pos++; return '\"';
        case '\'':
            buffer.pos++; return '\'';
        case '\\':
            buffer.pos++; return '\\';
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
            return (char)scanOctCharacter();
        default:
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR,
                                      String.valueOf(buffer.data[buffer.pos]));
        }
    }

    /**
     * Interprets an octal character from the input.
     */
    private char scanOctCharacter()
        throws PositionedError
    {
        char char1 = buffer.data[buffer.pos];
        /* Only two digits max if char1 > '3' */
        if (char1 >= '4' && char1 <= '7') {
            char char2 = buffer.data[buffer.pos + 1];
            if (char2 >= '0' && char2 <= '7') {
                buffer.pos += 2;
                return (char)((char1 - '0') * 8 + (char2 - '0'));
            }
            else {
                buffer.pos += 1;
                return (char)(char1 - '0');
            }
        }
        else {
            char char2 = buffer.data[buffer.pos + 1];
            if (char2 >= '0' && char2 <= '7') {
                char char3 = buffer.data[buffer.pos + 2];
                if (char3 >= '0' && char3 <= '7') {
                    buffer.pos += 3;
                    return (char)((char1 - '0') * 64 + (char2 - '0') * 8
                                  + (char3 - '0'));
                }
                else {
                    buffer.pos += 2;
                    return (char)((char1 - '0') * 8 + (char2 - '0'));
                }
            }
            else {
                buffer.pos += 1;
                return (char)(char1 - '0');
            }
        }
    }

    /**
     * Interprets a hexadecimal character from the input.
     */
    private char scanHexCharacter()
        throws PositionedError
    {
        int  val = 0;
        int  i;

        for (i = 0;
             i < 4 && (Character.digit(buffer.data[buffer.pos], 16) != -1);
             i++) {

            val = (16*val) + Character.digit(buffer.data[buffer.pos++], 16);
        }

        if (i == 0) {
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR,
                                      "" + buffer.data[buffer.pos]);
        }

        return (char)val;
    }

    /**
     * Determines if the specified character is permissible as
     * the first character in a Java identifier.
     *
     * @return true iff the character may start a Java identifier
     */
    protected static boolean isJavaIdentifierStart(char c)
    {
        switch (c) {
        case '$':  case 'A':  case 'B':  case 'C':
        case 'D':  case 'E':  case 'F':  case 'G':
        case 'H':  case 'I':  case 'J':  case 'K':
        case 'L':  case 'M':  case 'N':  case 'O':
        case 'P':  case 'Q':  case 'R':  case 'S':
        case 'T':  case 'U':  case 'V':  case 'W':
        case 'X':  case 'Y':  case 'Z':  case '_':
        case 'a':  case 'b':  case 'c':  case 'd':
        case 'e':  case 'f':  case 'g':  case 'h':
        case 'i':  case 'j':  case 'k':  case 'l':
        case 'm':  case 'n':  case 'o':  case 'p':
        case 'q':  case 'r':  case 's':  case 't':
        case 'u':  case 'v':  case 'w':  case 'x':
        case 'y':  case 'z':
            return true;
        default:
            return Character.isJavaIdentifierStart(c);
        }
    }

    /**
     * Reads a Java style comment (with leading // or <slash>* or <slash>**).
     * @return true iff a comment was read (the leading '/' is consumed)
     */
    protected boolean readJavaComment()
        throws PositionedError
    {
        // line.charAt(line_buffer.pos+0) is '/'

        switch (buffer.data[buffer.pos + 1]) {
        case '/':
            readEndOfLineComment();
            return true;
        case '*':
            readTraditionalComment();
            return true;
        default:
            // it's a token, not a comment
            return false;
        }
    }

    /**
     * Reads '//' comment.
     */
    private void readEndOfLineComment()
        throws PositionedError
    {
        buffer.pos += 2; // skip leading "//"

        int start = buffer.pos;

        // read until end of line
        while (true) {
            if (buffer.isEndOfData()) {
                throw new PositionedError(getTokenReference(),
                                          CompilerMessages.UNEXPECTED_EOF);
            }

            char c = buffer.data[buffer.pos];
            if (c == '\n' || c == '\r') {
                break;
            }
            else {
                buffer.pos++;
            }
        }
      
        if (compiler.parseComments()) {
            addComment(new JavaStyleComment(new String(buffer.data,
                                                       start,
                                                       buffer.pos - start),
                                            true,
                                            spaceBefore(start),
                                            spaceAfter(buffer.pos)));
        }
    }

    /**
     * ...
     */
    private void readTraditionalComment()
        throws PositionedError
    {
        buffer.pos += 2; // skip leading "/*"

        int  start = buffer.pos;

        // read until "*/" found
    _loop_:
        while (true) {
            //!!! graf 000212: check for EOF (needs to declare a possible
            // exception)
            //!!! graf 000212: '\r' and '\r\n' are not handled properly
            //!!! graf 000212: add warning for "/*" inside comment
            switch (buffer.data[buffer.pos]) {
            case '*':
                buffer.pos++;
                while (buffer.data[buffer.pos] == '*') {
                    buffer.pos++;
                }
                if (buffer.data[buffer.pos] == '/') {
                    buffer.pos++;
                    break _loop_;
                }
                break;
            case '\n':
                buffer.line++;
                buffer.pos++;
                break;
            default:
                buffer.pos++;
            }
        }

        if (compiler.parseComments()) {
            if (buffer.data[start] == '*') {
                // documentation comment
                addComment(new JavadocComment(new String(buffer.data,
                                                         start + 1,
                                                         buffer.pos - start-3),
                                              spaceBefore(start),
                                              spaceAfter(buffer.pos)));
            }
            else {
                // traditional comment
                addComment(new JavaStyleComment(new String(buffer.data, start,
                                                           buffer.pos-start-2),
                                                false,
                                                spaceBefore(start),
                                                spaceAfter(buffer.pos)));
            }
        }
    }

    /**
     * Adds a comment.
     */
    private final void addComment(JavaStyleComment comment)
    {
        comments.add(comment);
    }

    /**
     *
     */
    public JavaStyleComment[] getStatementComment()
    {
        JavaStyleComment[] comm = (JavaStyleComment[])
            comments.toArray(new JavaStyleComment[comments.size()]);

        comments.clear();

        return comm;
    }


    /**
     *
     */
    public JavadocComment getJavadocComment()
    {
        for (int i = comments.size() - 1; i >= 0; i--) {
            if (comments.get(i) instanceof JavadocComment) {
                JavadocComment jdoc = (JavadocComment)comments.get(i);

                comments.set(i, null);
                return jdoc;
            }
        }
        return null;
    }

    private int previousReturn(int i)
    {
        for (; i >= 0; i--) {
            if (buffer.data[i] == '\n') {
                return i;
            }
        }
        return 0;
    }

    private int previousNonBlank(int i)
    {
        for (; i >= 0; i--) {
            if (buffer.data[i] != ' ' && buffer.data[i] != '\t') {
                return i;
            }
        }
        return 0;
    }

    private int nextReturn(int i)
    {
        for (; i < buffer.data.length; i++) {
            if (buffer.data[i] == '\n' || buffer.data[i] == 0) {
                return i;
            }
        }

        return buffer.data.length;
    }

    private int nextNonBlank(int i)
    {
        for (; i < buffer.data.length; i++) {
            if ( (buffer.data[i] != ' ' && buffer.data[i] != '\t')
                 || buffer.data[i] == 0 ) {
                return i;
            }
        }

        return buffer.data.length;
    }

    private boolean spaceBefore(int pos)
    {
        int  previousReturn = previousReturn(pos);

        return ( previousReturn == 0
                 ? false
                 : ( previousNonBlank(previousReturn - 1)
                     <= previousReturn(previousReturn - 1) ) );
    }

    private boolean spaceAfter(int pos)
    {
        int    nextReturn = nextReturn(pos);

        return nextNonBlank(nextReturn + 1) >= nextReturn(nextReturn + 1);
    }
}
