/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;

/**
 * This class represents the type of an array in the type structure.
 */
public class CArrayType
    extends CClassType
{
    /**
     * Base type of this array type.
     */
    private CType baseType;

    /**
     * The number of dimensions of this array type.
     */
    private int arrayBound;

    /**
     * Constructs an array type.
     *
     * @param baseType
     * The base type of the array.
     *
     * @param arrayBound
     * The number of dimensions of the array.
     */
    public CArrayType(CType baseType, int arrayBound)
    {
        super();

        type = TID_ARRAY;
        assert(baseType != null);
        assert(arrayBound > 0);

        if (baseType.isArrayType()) {
            this.arrayBound = arrayBound + ((CArrayType)baseType).arrayBound;
            this.baseType = ((CArrayType)baseType).baseType;
        }
        else {
            this.arrayBound = arrayBound;
            this.baseType = baseType;
        }
    }

    /**
     * Transforms this type to a string
     * @return the ksm form of this type
     */
    public String toString()
    {
        String res = baseType.toString();
        for (int i = 0; i < arrayBound; i++) {
            res += "[]";
        }
        return res;
    }

    /**
     * Returns the signature.
     */
    public String getQualifiedName()
    {
        return getSignature();
    }

    /**
     * Appends the classfile signature of this type to the <code>buffer</code>
     * parameter.
     */
    public void appendSignature(FastStringBuffer buffer)
    {
        for (int i = 0; i < arrayBound; i++) {
            buffer.append('[');
        }

        baseType.appendSignature(buffer);
    }

    /**
     * Returns the size used in stack by value of this type.
     *
     * @return
     * The size used in stack by value of this type.
     */
    public int getSize()
    {
        return 1;
    }

    /**
     * Returns true because an array always has a reference type.
     *
     * @return true because an array always has a reference type.
     */
    public boolean isReference()
    {
        return true;
    }

    /**
     * Returns true, because this is the class for array types.
     *
     * @return true, because this is the class for array types.
     */
    public boolean isArrayType()
    {
        return true;
    }

    /**
     * Returns the base class of this type.
     *
     * @return the base class of this type.
     */
    public CType getBaseType()
    {
        assert(baseType != null);
        assert(!(baseType instanceof CArrayType));
        return baseType;
    }

    /**
     * Returns the type that has one dimension less than this array type.
     *
     * @return
     * The type that has one dimension less than this array type.
     */
    public CType getIndexedType()
    {
        assert(baseType != null);
        assert(!(baseType instanceof CArrayType));
        if (arrayBound == 1) {
            return baseType;
        }
        else {
            return new CArrayType(baseType, arrayBound - 1);
        }
    }

    /**
     * Returns the number of dimensions of this array type.
     *
     * @return the number of dimensions of this array type.
     */
    public int getArrayBound()
    {
        return arrayBound;
    }

    /**
     * Returns true if this type is checked and valid.
     *
     * @return true if this type is checked and valid.
     */
    public boolean checked()
    {
        return baseType.checked();
    }

    /**
     * Returns true if this type is equal to another CType.
     *
     * XXX: it is bad style and a cause of many errors to overload the equals
     * method with another type than "Object" as its single argument.
     * TODO: rename method in all CType classes to equalsCType and make
     * an equals(Object) that calls to equalsCType if the argument is of that
     * type. Java does _not_ have dynamic dispatching on argument types, never
     * forget!
     */
    public boolean equals(CType other)
    {
        if (!other.isArrayType()) {
            return false;
        }
        else {
            CArrayType arr = (CArrayType)other;

            return ( baseType.equals(arr.baseType) &&
                     arrayBound == arr.arrayBound );
        }
    }

    /**
     * Check that type is valid.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void checkType(CContext context)
        throws UnpositionedError
    {
        if (!isChecked()) {
            setClass(CStdType.Object.getCClass());
            baseType.checkType(context);
        }
    }

    /**
     * Returns the class object associated with this type
     *
     * If this type was never checked (read from class files)
     * check it!
     *
     * @return the class object associated with this type
     */
    public CClass getCClass()
    {
        if (!isChecked()) {
            try {
                checkType(null);
            }
            catch (UnpositionedError cue) {
                // Can only fail because the base type doesn't check out.
                // Then the super.getCClass will return BAC_CLASS.
            }
        }

        return super.getCClass();
    }

    /**
     * isAssignableTo
     *
     * @param dest
     * The destination type.
     *
     * @return
     * Whether assignment is possible.
     */
    public boolean isAssignableTo(CType dest)
    {
        if (dest.isArrayType()) {
            if (arrayBound == ((CArrayType)dest).arrayBound &&
                baseType.isAssignableTo(((CArrayType)dest).baseType)) {
                // normal folks
                return true;
            }
            else {
                if (arrayBound < ((CArrayType)dest).arrayBound) {
                    // int[][] i = new int[]; ???
                    return false;
                }
                // May be the unusal case Object[] o = new String[][]...;
                return (((CArrayType)dest).baseType).equals(CStdType.Object);
            }
        }
        else {
            return dest.equals(CStdType.Object);
        }
    }

    /**
     * Can this type be converted to the specified type by casting 
     * conversion (JLS 5.5) ?
     * @param dest  the destination type
     * @return true iff the casting conversion is valid
     */
    public boolean isCastableTo(CType dest)
    {
        // test for array first because array types are classes

        if (dest.isArrayType()) {
            // optimizes recursive call with base types (indexed types)

            CType  destBase = ((CArrayType)dest).baseType;
            int  destBound = ((CArrayType)dest).arrayBound;

            if (arrayBound == destBound) {
                if (baseType.isPrimitive()) {
                    return baseType == destBase;
                }
                else {
                    return baseType.isCastableTo(destBase);
                }
            }
            else if (arrayBound < destBound) {
                return baseType.isCastableTo
                    (new CArrayType(destBase, destBound - arrayBound));
            }
            else {
                // arrayBound > destBound
                return new CArrayType(baseType, arrayBound - destBound)
                    .isCastableTo(destBase);
            }
        }
        else if (dest.isClassType()) {
            if (dest.equals(CStdType.Object)) {
                // if T is a class type, then if T is not Object,
                // then a compile-time error occurs
                return true;
            }
            else if (dest.getCClass().getQualifiedName()
                     .equals(JAV_CLONEABLE)) {
                // if T is an interface type, then a compile-time error
                // occurs unless T is the interface type Cloneable
                return true;
            }
            else {
                return false;
            }
        }
        else {
            return false;
        }
    }

    /**
     * implicitlyCastTo
     * @param dest the expression type to cast to
     * @return cast is valid ?
     */
    public boolean implicitlyCastTo(CType dest)
    {
        return isAssignableTo(dest);
    }
}
