/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashSet;

/**
 * This class represents a local context during checkBody.
 * It follows the control flow and maintain informations about
 * variable (initialized, used, allocated) and exceptions (thrown, catched.)
 * It also verifies if the context is still reachable.
 *
 * There is a set of utilities method to access fields, methods and class
 * with the name by clamping the parsing tree
 * @see CContext
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CConstructorContext
 * @see CInitializerContext
 * @see CBodyContext
 * @see CBlockContext
 */
public class CBodyContext
    extends CAbstractBodyContext
{
    /**
     * Construct a body context.
     *
     * @param parent
     * The parent context; may not be null except if called by the top level.
     */
    protected CBodyContext(CAbstractBodyContext parent)
    {
        super(parent);
    }

    /**
     * Construct a body context.
     *
     * @param parent
     * The parent context; may not be null except if called by the top level.
     */
    private CBodyContext(CAbstractBodyContext parent, CBodyContext clone)
    {
        super(parent);

        flowState = clone.flowState;

        if (clone.variableInfo != null) {
            this.variableInfo = (CVariableInfo)clone.variableInfo.clone();
        }

        if (clone.fieldInfo != null) {
            this.fieldInfo = (CVariableInfo)clone.fieldInfo.clone();
        }
    }

    /**
     * Creates a class context (for inner/local classes) with this body
     * context as its parent.
     */
    public CClassContext createClassContext(CSourceClass clazz)
    {
        return new CClassContext(this, clazz, null);
    }

    /**
     * Creates a block context with this body context as its parent.
     */
    public CBlockContext createBlockContext()
    {
        return new CBlockContext(this);
    }
  
    /**
     * Creates a block context with this body context as its parent.
     */
    public CBlockContext createBlockContext(int params)
    {
        return new CBlockContext(this, params);
    }

    /**
     * Creates a loop context with this body context as its parent.
     */
    public CLoopContext createLoopContext(JLoopStatement self)
    {
        return new CLoopContext(this, self);
    }
  
    /**
     * Creates a try context with this body context as its parent.
     */
    public CTryContext createTryContext()
    {
        return new CTryContext(this);
    }

    /**
     * Creates an expression context with this body context as its parent.
     */
    public CExpressionContext createExpressionContext()
    {
        return new CExpressionContext(this);
    }

    /**
     * Returns a clone of this body context.
     */
    public CBodyContext cloneContext()
    {
        return new CBodyContext((CAbstractBodyContext)parent, this);
    }

    /**
     * Creates a snapshot. A snapshot is a clone without throwables
     * information.
     */
    public CBodyContext makeSnapshot()
    {
        CBodyContext copy = cloneContext();

        copy.throwables = new HashSet();
        return copy;
    }

    /**
     * Merge variable and field assignment information from another context
     * with this one. Merging means that unless some variable is definately
     * assigned in both cases, it is never definately assigned afterwards.
     */
    // XXX: rename, to ? mergeFlowAndAssignmentInfo ?
    public void merge(CBodyContext context)
    { 
        flowState = flowState & context.flowState;

        if (variableInfo != null || context.variableInfo != null) {
            int parentPosition = getBlockContext().getMaxLocalsIndex();
      
            for (int i = 0; i < parentPosition; i++) {
                int infoA = getVariableInfo(i);
                int infoB = context.getVariableInfo(i);
 
                if ( (infoA | infoB) != 0) {
                    setVariableInfo(i, CVariableInfo.merge(infoA, infoB));
                }
            }
        }

        if (fieldInfo != null || context.fieldInfo != null) {
            int parentPosition =
                parent.getClassContext().getCClass().getFieldCount();
      
            for (int i = 0; i < parentPosition; i++) {
                int infoA = getFieldInfo(i);
                int infoB = context.getFieldInfo(i);
 
                if ( (infoA | infoB) != 0) {
                    setFieldInfo(i, CVariableInfo.merge(infoA, infoB));
                }
            }    
        }
    }

    /**
     * Adopt variable and field assignment information from another context.
     * Adopting means that if the other context has some assignment
     * information about some field or variable, it overrules any assignment
     * information about it that is already present in this context.
     */
    // XXX: rename, to ? adoptFlowAndAssignmentInfo ?
    public void adopt(CBodyContext context)
    {
        flowState = context.flowState;

        int parentPosition = getBlockContext().getMaxLocalsIndex();
      
        for (int i = 0; i < parentPosition; i++) {
            int info = context.getVariableInfo(i);
      
            if (info != 0) {
                addVariableInfo(i, info);
            }
        }

        parentPosition = parent.getClassContext().getCClass().getFieldCount();
    
        for (int i = 0; i < parentPosition; i++) {
            int info = context.getFieldInfo(i);
      
            if (info != 0) {
                addFieldInfo(i, info);
            }    
        }
    }
}
