/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.io.File;
import java.io.IOException;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import at.dms.classfile.*;
import at.dms.compiler.Compiler;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.Utils;

/**
 * This class represents the exported members of a class. (inner classes,
 * methods and fields)
 */
public abstract class CClass
    extends CMember
{
    /**
     * Special constant that represents an "undefined" class that is used
     * when some class object is needed as a placeholder to generate better
     * error messages.
     */
    protected static CClass CLS_UNDEFINED =
        new CSourceClass(null, null, null, TokenReference.NO_REF, 0,
                         "<gen>", "<gen>", false);

    /**
     * Compiler that generated this class export.
     */
    private final Compiler compiler;

    /**
     * Name of file from which this class was compiled (possibly this came
     * from the .class file, or this information is supplied by the compiler
     * in the case of a CSourceClass)
     */
    private final String sourceFile;

    /**
     * The fully qualified name of the class.
     */
    private final String qualifiedName;

    /**
     * The package name of the class.
     */
    private final String packageName;

    /**
     * The interfaces implemented by the class.
     */
    private CClassType[] interfaces;

    /**
     * Type object that represents this class as a type.
     */
    private CClassType type;

    /**
     * The inner classes that are contained in this class.
     */
    private CClassType[] innerClasses;

    /**
     * The superclass of this class.
     */
    private CClassType superClass;

    /**
     * The fields in this class.
     */
    private CFieldTable fields;

    /**
     * The methods in this class.
     */
    private CMethod[] methods;

    /**
     * Constructs a class export from file.
     */
    public CClass(Compiler compiler,
                  CClass owner,
                  String sourceFile,
                  int modifiers,
                  String ident,
                  String qualifiedName,
                  CClassType superClass,
                  boolean deprecated)
    {
        super(owner, modifiers, ident, deprecated);

        this.compiler = compiler;
        this.sourceFile = sourceFile;
        this.qualifiedName = qualifiedName;
        this.superClass = superClass;

        int cursor = qualifiedName.lastIndexOf('/');

        if (cursor > 0) {
            this.packageName = qualifiedName.substring(0, cursor).intern();
        }
        else {
            this.packageName = "";
        }
    }

    /**
     * Ends the definition of this class.
     */
    public void close(CClassType[] interfaces,
                      CField[] fields,
                      CMethod[] methods)
    {
        this.interfaces = interfaces;
        this.fields = new CFieldTable(fields);
        this.methods = methods;
        assert(interfaces != null);
    }

    /**
     * Returns the compiler.
     */
    public Compiler getCompiler()
    {
        return compiler;
    }

    /**
     * Returns <code>this</code>.
     *
     * @return <code>this</code>.
     */
    public CClass getCClass()
    {
        return this;
    }

    /**
     * Returns the size of the outer-local variables in the variable table.
     */
    public int getOuterLocalsSize()
    {
        return 0;
    }

    /**
     * @return true if this class is an inner class.
     */
    public boolean isInnerClass()
    {
        return getOwner() != null;
    }

    /**
     * @return true if this member is abstract.
     */
    public boolean isAbstract()
    {
        return hasFlag(getModifiers(), ACC_ABSTRACT);
    }

    /**
     * @return true if this class is anonymous.
     */
    public boolean isAnonymous()
    {
        return getIdent().length() == 0;
    }

    /**
     * @return true if this class is has an outer this.
     */
    public boolean hasOuterThis()
    {
        return isInnerClass() && !isStatic() && !isInterface();
    }

    /**
     * Sets hasOuterThis. 
     */
    public void setHasOuterThis(boolean hasOuterThis)
    {
        if (hasOuterThis != hasOuterThis()) {
            throw new IllegalArgumentException
                ("Cannot set outer this anymore; value must equals result of "
                 + "hasOuterThis() which is " + hasOuterThis());
        }
    }

    /**
     * Returns true if this class is an interface.
     *
     * @return true if this class is an interface.
     */
    public boolean isInterface()
    {
        return hasFlag(getModifiers(), ACC_INTERFACE);
    }

    /**
     * Returns the full name of this class.
     *
     * @return the full name of this class.
     */
    public String getQualifiedName()
    {
        return qualifiedName;
    }

    /**
     * Returns the name of the package of the package this class belongs to.
     *
     * @return the name of the package of the package this class belongs to.
     */
    public String getPackage()
    {
        return packageName;
    }

    /**
     * Returns the super class of this class.
     */
    public CClass getSuperClass()
    {
        return superClass == null ? null : superClass.getCClass();
    }

    /**
     * Sets the super class of this class.
     */
    public void setSuperClass(CClassType superClass)
    {
        this.superClass = superClass;
    }

    /**
     * Returns the type of this class.
     */
    public CClassType getType()
    {
        if (type == null) {
            type = CClassType.lookup(getQualifiedName());
        }
        return type;
    }

    /**
     * Returns the source file of this class.
     *
     * @return
     * The source file of this class.
     */
    public String getSourceFile()
    {
        return sourceFile;
    }

    /**
     * Returns the named field.
     *
     * @param ident
     * The name of the field to get.
     *
     * @return
     * The field with the same name as the value of <code>ident</code>.
     */
    public CField getField(String ident)
    {
        assert(this != CLS_UNDEFINED, "This is the CLS_UNDEFINED class");
        assert(fields != null, getQualifiedName());
        return fields.lookup(ident);
    }

    /**
     * Returns the number of fields in this class.
     */
    public int getFieldCount()
    {
        return fields.size();
    }

    /**
     * Returns the InnerClasses of this clas.
     *
     * @return the InnerClasses.
     */
    public CClassType[] getInnerClasses()
    {
        return innerClasses;
    }

    /**
     * End of first pass, we need inner classes/interfaces.
     */
    public void setInnerClasses(CClassType[] inners)
    {
        this.innerClasses = inners;
    }

    /**
     * Add synthetic method class.
     */
    public void addMethod(CSourceMethod newMethod)
    {
        for (int i = 0; i < methods.length; i++) {
            if (methods[i].equals(newMethod)) {
                throw new IllegalArgumentException("Duplicate method "
                                                   + newMethod.getIdent());
            }
        }
        CMethod[] newMethods = new CMethod[methods.length + 1];
        System.arraycopy(methods, 0, newMethods, 0, methods.length);
        newMethods[methods.length] = newMethod;
        this.methods = newMethods;
    }

    /**
     * Add synthetic inner class.
     */
    public void addInnerClass(CClassType newInner)
    {
        CClassType[] newInners = new CClassType[innerClasses.length + 1];
        System.arraycopy(innerClasses, 0, newInners, 0, innerClasses.length);
        newInners[innerClasses.length] = newInner;
        this.innerClasses = newInners;
    }

    /**
     * descendsFrom.
     *
     * @param from an other CClass
     *
     * @return true if this class inherit from "from" or equals "from"
     */
    public boolean descendsFrom(CClass from)
    {
        if (from == this) {
            return true;
        }
        else if (superClass == null) {
            // java/lang/object
            return false;
        }
        else if (getSuperClass().descendsFrom(from)) {
            return true;
        }
        else {
            for (int i = 0; i < interfaces.length; i++) {
                if (interfaces[i].getCClass().descendsFrom(from)) {
                    return true;
                }
            }

            return false;
        }
    }

    /**
     * Returns true iff this class is defined inside the specified class
     * @param outer  another class
     */
    public boolean isDefinedInside(CClass outer)
    {
        if (this == outer) {
            return true;
        }
        else if (getOwner() == null) {
            return false;
        }
        else {
            return getOwner().isDefinedInside(outer);
        }
    }

    /**
     * Look up member classes of this class, or of superclasses or interfaces.
     *
     * @param name
     * A TypeName (6.5.2)
     */
    public CClass lookupMemberClass(String name)
        throws UnpositionedError
    {
        if (innerClasses != null && (name.indexOf('/') == -1)) {
            for (int i = 0; i < innerClasses.length; i++) {
                if (innerClasses[i].getCClass().getIdent().equals(name)) {
                    return innerClasses[i].getCClass();
                }
            }
        }
        CClass memberClazz = null;

        if (superClass != null) {
            memberClazz = superClass.getCClass().lookupMemberClass(name);
        }

        if (interfaces == null) {
            return memberClazz;
        }

        for (int i = 0; i < interfaces.length; i++) {
            CClass ifaceMemberClazz =
                interfaces[i].getCClass().lookupMemberClass(name);

            if (ifaceMemberClazz != null) {
                if (memberClazz != null
                    && memberClazz.getOwner() != ifaceMemberClazz.getOwner()) {

                    throw new UnpositionedError
                        (KjcMessages.TYPE_NAME_AMBIGUOUS,
                         new Object[] {
                             name,
                             ifaceMemberClazz.getQualifiedName(),
                             memberClazz.getQualifiedName() });
                }
                
                memberClazz = ifaceMemberClazz;
            }
        }

        return memberClazz;
    }

    /**
     * Searches a field in current class and parent hierarchy as needed.
     *
     * @param name
     * The simple name of the field.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CField lookupField(String name)
        throws UnpositionedError
    {
        CField field = getField(name);

        if (field != null) {
            return field;
        }
        else {
            field = lookupSuperField(name);
            return field;
        }
    }

    /**
     * Searches a field in current class and parent hierarchy as needed.
     *
     * @param name  the simple name of the field
     * @exception UnpositionedError this error will be positioned soon
     */
    public CField lookupSuperField(String name)
        throws UnpositionedError
    {
        CField field = getField(name);

        if (superClass != null) {
            // java.lang.object / interfaces
            field = getSuperClass().lookupField(name);
        }

        for (int i = 0; i < interfaces.length; i++) {
            CField ifaceField = interfaces[i].getCClass().lookupField(name);

            if (ifaceField != null) {
                if (field != null
                    && field.getOwner() != ifaceField.getOwner()) {

                    throw new UnpositionedError(KjcMessages.FIELD_AMBIGUOUS,
                                                name);
                }
                
                field = ifaceField;
            }
        }

        return field;
    }

    /**
     * Searches for an accessible, matching method with the provided type
     * parameters. Looks in parent hierarchy as needed.
     *
     * @param caller
     * The class of the method call expression, which is used to determine
     * whether a matching method is accessible. If it is not accessible,
     * the search goes on for another matching method. This parameter may
     * be <code>null</code> to look up any matching method, regardless of
     * accessibility.
     *
     * @param name
     * Method name.
     *
     * @param params
     * Method parameters.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CMethod lookupMethod(CClass caller, String name, CType[] params)
        throws UnpositionedError
    {
        CMethod method = getMatchingMethod(caller, name, params);

        if (method != null) {
            return method;
        }
        else {
            return lookupSuperMethod(caller, name, params);
        }
    }

    /**
     * Searches for a matching method with the provided type parameters
     * in the superclass and/or interfaces of this type.
     *
     * @param caller
     * The class of the method call expression, which is used to determine
     * whether a matching method is accessible. If it is not accessible,
     * the search goes on for another matching method. This parameter may
     * be <code>null</code> to look up any matching method, regardless of
     * accessibility.
     *
     * @param name
     * Method name.
     *
     * @param params
     * Method parameters.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CMethod lookupSuperMethod(CClass caller,
                                     String name, CType[] params)
        throws UnpositionedError
    {
        CMethod method;

        if (superClass == null) {
            method = null;
        }
        else {
            method = getSuperClass().lookupMethod(caller, name, params);
        }

        for (int i = 0; i < interfaces.length && method == null; i++) {
            method = interfaces[i].getCClass().lookupMethod(caller,
                                                            name, params);
        }

        return method;
    }

    /**
     * getMatchingMethod.
     *
     * Search for a matching method with the provided type parameters.
     * Linear search.
     * (if method exists)
     *
     * @param caller
     * If non-null, then a method only matches if it is accessible from the
     * caller class.
     *
     * @param ident
     * Method name.
     *
     * @param params
     * Method parameters.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    private CMethod getMatchingMethod(CClass caller,
                                      String ident, CType[] params)
        throws UnpositionedError
    {
        CMethod method = null;
        boolean lastMethodAmbiguous = false;

        for (int i = 0; i < methods.length; i++) {
            CMethod current = methods[i];

            if (current.isMatching(ident, params)) {
                if (caller != null && !current.isAccessible(caller)) {
                    continue;
                }
                if (method == null || current.isMoreSpecificThan(method)) {
                    method = current;
                    lastMethodAmbiguous = false;
                }
                else if (method.isMoreSpecificThan(current)) {
                    lastMethodAmbiguous = false;
                }
                else {
                    lastMethodAmbiguous = true;
                }
            }
        }

        if (lastMethodAmbiguous) {
            throw new UnpositionedError(KjcMessages.METHOD_AMBIGOUS, ident);
        }

        return method;
    }

    /**
     * Returns a list of abstract methods
     */
    public CMethod[] getAbstractMethods()
    {
        List methodsList = collectAbstractMethods();
        CMethod[] methods = new CMethod[methodsList.size()];
        methods = (CMethod[])methodsList.toArray(methods);
        return methods;
    }

    /**
     * Returns a list of interface methods
     */
    public CMethod[] getInterfaceMethods() {
        List l = new ArrayList(5);
        collectInterfaceMethods(l);
        CMethod[] methods = new CMethod[l.size()];
        return (CMethod[])l.toArray(methods);
    }

    /**
     * collectAbstractMethods
     */
    public List collectAbstractMethods()
    {
        List collection = new ArrayList(methods.length);
        for (int i = 0; i < methods.length; i++) {
            CMethod current = methods[i];

            if (current.isAbstract()) {
                collection.add(current);
            }
        }

        if (superClass != null) {
            collection.addAll(getSuperClass().collectAbstractMethods());
        }
        return collection;
    }

    /**
     * Collects all methods in this type that are declared in an interface.
     * Looks in parent hierarchy as needed.
     *
     * @param list
     * Where the collected methods are placed.
     */
    public void collectInterfaceMethods(List list) {
        if (isInterface()) {
            for (int i = 0; i < methods.length; i++) {
                list.add(methods[i]);
            }
        }

        for (int i = 0; i < interfaces.length; i++) {
            CClass parent;
            parent = interfaces[i].getCClass();
            assert(parent != null);
            parent.collectInterfaceMethods(list);
        }

        if (superClass != null) {
            /* We're not in java.lang.Object */
            getSuperClass().collectInterfaceMethods(list);
        }
    }

    /**
     * Returns the unqualified, package less and outerclass-name-less name
     * of this class.
     *
     * @return
     * The unqualified, package less and outerclass-name-less name of this
     * class.
     */
    protected static String getIdent(String qualifiedName)
    {
        assert(qualifiedName != null);
        String shortName;

        int cursor = qualifiedName.lastIndexOf('/');
        if (cursor > 0) {
            shortName = qualifiedName.substring(cursor + 1);
        }
        else {
            shortName = qualifiedName;
        }

        cursor = shortName.lastIndexOf('$');
        if (cursor > 0) {
            return shortName.substring(cursor + 1);
        }
        else {
            return shortName;
        }
    }

    /**
     * Checks if this class is accessible from another class.
     *
     * @return true
     * If this class is accessible from the class <code>from</code>.
     */
    public boolean isAccessible(CClass from)
    {
        if ((from.isDefinedInside(this)) || isPublic()
            || isDefinedInside(from)) {
            return true;
        }
        else if (from.isInnerClass() && isAccessible(from.getOwner())) {
            return true;
        }
        else if (!isPrivate() && getPackage() == from.getPackage()) {
            return true;
        }
        else if (isInnerClass() && isProtected()
                 && from.descendsFrom(getOwner())) {
            return true;
        }
        else {
            return false;
        }
    }

    /**
     * Add a field.
     */
    public void addField(CSourceField field)
    {
        fields.addField(field);
    }

    /**
     * Generate the code in a class file.
     *
     * @param destination the root directory of the class hierarchy.
     */
    public void genClassFile(String destination)
        throws IOException, ClassFileFormatException
    {
        ClassConstant[] ifaceConstants = new ClassConstant[interfaces.length];

        for (int i = 0; i < this.interfaces.length; i++) {
            ifaceConstants[i] =
                new ClassConstant(interfaces[i].getQualifiedName());
        }

        int count = 0;
        MethodInfo[] pmethods = new MethodInfo[methods.length];

        for (int i = 0; i < methods.length; i++) {
            pmethods[count++] = methods[i].genMethodInfo();

        }

        FieldInfo[] pfields = fields.buildFieldInfo();

        InnerClassInfo[] inners = new InnerClassInfo[innerClasses.length];

        for (int i = 0; i < inners.length; i++) {
            CClass innerCClass = innerClasses[i].getCClass();
            inners[i] = new InnerClassInfo(innerCClass.getQualifiedName(),
                                           getQualifiedName(),
                                           innerCClass.getIdent(),
                                           (short)innerCClass.getModifiers());
        }

        String[] sourceSplit = Utils.splitQualifiedName(sourceFile,
                                                        File.separatorChar);

        ClassInfo classInfo =
            new ClassInfo((short)
                          (getModifiers() & (~ACC_STATIC)),
                          getQualifiedName(),
                          (superClass == null
                           ? null
                           : superClass.getCClass().getQualifiedName()),
                          ifaceConstants,
                          pfields,
                          pmethods,
                          inners,
                          sourceSplit[1],
                          false);

        String[] nameSplit = Utils.splitQualifiedName(getQualifiedName(), '/');

        File writeDir;
        String writeFile;

        if (destination == null || destination.equals("")) {
            /* No destination dir: Put class in same dir as source dir. */
            writeDir = new File(".", sourceSplit[0]);
        }
        else {
            writeDir = new File(destination, nameSplit[0]);
        }

        writeFile = nameSplit[1] + ".class";

        classInfo.write(writeDir, writeFile);

    }

    /**
     * Add synthetic parameters to method def.
     */
    public CType[] genConstructorArray(CType[] params)
    {
        throw new RuntimeException();
    }

    /**
     * Add synthetic parameters to method call.
     */
    public void genSyntheticParams(CodeSequence code, JExpression explicitThis)
    {
        if (isInnerClass() && hasOuterThis()) {
            if (explicitThis == null) {
                code.plantLoadThis();
            }
            else {
                explicitThis.genCode(code);
                /* XXX: JLS2: 15.9.4: check if null, then throw
                 * NullPointerException! Don't give a null outer-this to the
                 * constructor method. TODO!
                 */
            }
        }
    }

    /**
     * Add synthetic parameters to method call.
     */
    public void genSyntheticParamsFromExplicitSuper(JExpression[] params,
                                                    CodeSequence code)
    {
        if (isInnerClass() && hasOuterThis()) {
            JGeneratedLocalVariable var;
            var = new JGeneratedLocalVariable(null, 0,
                                              getOwner().getType(),
                                              JAV_OUTER_THIS, null);
            var.setPosition(1);
            new JLocalVariableExpression(TokenReference.NO_REF, var)
                .genCode(code);
        }
    }
}
