/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.CWarning;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a class context during check.
 *
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CContext
 */
public class CClassContext
    extends CContext
{
    /**
     * Field assignmnent information.
     */
    private CVariableInfo fieldInfo = new CVariableInfo();

    /**
     * Type declaration of class that provides this context.
     */
    private JTypeDeclaration decl;

    /**
     * Whether this class has an initializer.
     */
    private boolean hasInitializer;

    /**
     * Next anonymous index.
     */
    private int anonymous = 0;

    /**
     * The class that provides this context.
     */
    protected CSourceClass self;

    /**
     * Constructs a class context.
     *
     * @param parent
     * The parent context or null at top level.
     *
     * @param clazz
     * The corresponding clazz.
     */
    CClassContext(CContext parent, CSourceClass clazz, JTypeDeclaration decl)
    {
        super(parent);

        this.self = clazz;
        this.decl = decl;
    }

    /**
     * Verify all final fields are initialized.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void close(JTypeDeclaration decl,
                      CVariableInfo staticInfo,
                      CVariableInfo instanceInfo,
                      CVariableInfo[] constructorsInfo)
        throws UnpositionedError
    {
        JFieldDeclaration[] fields = decl.getFields();

        for (int i = 0; i < fields.length; i++) {
            CSourceField field = (CSourceField)fields[i].getField();

            if (isFieldRedefined(field.getIdent())) {
                reportTrouble(new CWarning(decl.getTokenReference(),
                                           KjcMessages.FIELD_RENAME_SUPER,
                                           field.getIdent(),
                                           null));
            }

            if (field.isFinal()) {
                if (field.isStatic()) {
                    check(staticInfo != null
                          && CVariableInfo.isInitialized
                          (staticInfo.getInfo(field.getPosition())),
                          KjcMessages.UNINITIALIZED_FINAL_FIELD,
                          field.getIdent());
                }
                else if ( !(instanceInfo != null
                            && CVariableInfo.isInitialized
                            (instanceInfo.getInfo(field.getPosition())))) {
                    // should be initialized in all constructors
                    for (int j = 0; j < constructorsInfo.length; j++) {
                        check(CVariableInfo.isInitialized
                              (constructorsInfo[j].getInfo(field.getPosition())),
                              KjcMessages.UNINITIALIZED_FINAL_FIELD,
                              field.getIdent());
                    }
                }
            }
            else if (field.isStatic()) {
                if (staticInfo != null
                    && !CVariableInfo.isInitialized
                    (staticInfo.getInfo(field.getPosition()))) {
                    reportTrouble(new CWarning(decl.getTokenReference(),
                                               KjcMessages.UNINITIALIZED_FIELD,
                                               field.getIdent()));     
                }
            }
            else if (!(instanceInfo != null
                       && CVariableInfo.isInitialized
                       (instanceInfo.getInfo(field.getPosition())))) {
                // should be initialized in all constructors
                for (int j = 0; j < constructorsInfo.length; j++) {
                    if (!CVariableInfo.isInitialized
                        (constructorsInfo[j].getInfo(field.getPosition()))) {
                        reportTrouble
                            (new CWarning(decl.getTokenReference(),
                                          KjcMessages.UNINITIALIZED_FIELD,
                                          field.getIdent()));
  
                        break;
                    }
                }
            }
        }
        
        //!!! graf 000123: use self.isAbstract()
        if (! self.isAbstract()) {
            // check all abstract methods are implemented
            CMethod[] methods = self.getAbstractMethods();
            for (int i = 0; i < methods.length; i++) {
                CMethod m = self.lookupMethod(null,
                                              methods[i].getIdent(),
                                              methods[i].getParameters());

                check(! m.isAbstract(),
                      KjcMessages.CLASS_SHOULD_BE_ABSTRACT,
                      self.getIdent(), m.getIdent());
            }
        }

        // check all interfaces methods are implemented
        CMethod[] methods = self.getInterfaceMethods();
        for (int i = 0; i < methods.length; i++) {
            CMethod m = self.lookupMethod(null,
                                          methods[i].getIdent(),
                                          methods[i].getParameters());
      
            if (m == null) {
                System.err.println("Trouble: can't find "
                                   + methods[i].getIdent()
                                   + " in " + self.getIdent());
            }
            if (m.getIdent() != JAV_STATIC_INIT) {
                if (m.getOwner().isInterface()) {
                    if (!self.isAbstract()) {
                        check(false,
                              KjcMessages.CLASS_SHOULD_IMPLEMENTS,
                              self.getIdent(), m.getIdent());
                    }
                    else {
                        // add method as an abstract method in generated class
                        // !!! TODO
                    }
                }
            }
        }
    }

    /**
     * Returns true iff a field with same name is already defined in a
     * superclass or an implemented interface.
     *
     * @param ident
     * The name of the field.
     */
    private boolean isFieldRedefined(String ident)
        throws UnpositionedError
    {
        try {
            // defined once in super classes ?
            return self.lookupSuperField(ident) != null;
        }
        catch (UnpositionedError e) {
            if (! e.hasDescription(KjcMessages.FIELD_AMBIGUOUS)) {
                throw e;
            }
            // defined more than once!
            return true;
        }
    }

    /**
     * Makes new method context.
     */
    public CMethodContext createMethodContext(CMethod self)
    {
        return new CMethodContext(this, self);
    }

    /**
     * Makes new constructor-method context.
     */
    public CConstructorContext createConstructorContext(CMethod self)
    {
        return new CConstructorContext(this, self);
    }

    /**
     * Add an initializer to this context.
     */
    public void addInitializer()
    {
        hasInitializer = true;
    }

    /**
     * Returns true if this class need initializers
     */
    public boolean hasInitializer()
    {
        return hasInitializer;
    }

    /**
     * @see CSourceClass#getNextSyntheticIndex
     */
    public int getNextSyntheticIndex()
    {
        return self.getNextSyntheticIndex();
    }

    /**
     * @see CSourceClass#getNextAnonymousIndex
     */
    public int getNextAnonymousIndex()
    {
        return self.getNextAnonymousIndex();
    }

    /**
     * @param var  the definition of a variable
     * @return a variable info from a variable definition in current context
     */
    public int getVariableInfo(int index)
    {
        return ((CBodyContext)parent).getVariableInfo(index);
    } 

    /**
     * Returns information about a specific field.
     *
     * @param index
     * The field position in the field array.
     *
     * @return
     * All informations we have about this field.
     */
    public int getFieldInfo(int index)
    {
        return fieldInfo.getInfo(index);
    }

    /**
     * Sets field info about a specific field.
     *
     * @param index
     * The field position in the field array.
     * @param info
     * The information to add.
     *
     * We make it a local copy of this information and at the end of this
     * context we will transfert it to the parent context according to
     * controlFlow.
     */
    public void addFieldInfo(int index, int info)
    {
        fieldInfo.addInfo(index, info);
    }

    /**
     * Returns old field info and resets field info to a clean slate.
     */
    public CVariableInfo getVariableInfo()
    {
        CVariableInfo tmp = fieldInfo;

        fieldInfo = new CVariableInfo(); // ==> rename it reset

        return tmp;
    }

    /**
     * Marks all fields as 'initialized'
     */
    public void markAllFieldsAsInitialized()
    {
        for (int i = getCClass().getFieldCount() - 1; i >= 0; i--) {
            fieldInfo.addInfo(i, CVariableInfo.initialize());
        }
    }

    /**
     * Sets assignment information for all fields at once.
     */
    public void setVariableInfo(CVariableInfo info)
    {
        fieldInfo = info;
    }

    /**
     * Search for a class with the provided name.
     *
     * @param name
     * Method name
     * @return
     * The class if found, null otherwise.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CClassType lookupClass(String name)
        throws UnpositionedError
    {
        CClass clazz = self.lookupMemberClass(name);

        if (clazz != null) {
            return clazz.getType();
        }
        else {
            return super.lookupClass(name);
        }
    }

    /**
     * Searches for a matching method with the provided type parameters.
     *
     * @param caller
     * The class containing the statement that calls the method, used for
     * access checks. May be <code>null</code> to suppress access checks.
     *
     * @param ident
     * Method name.
     *
     * @param params
     * Method parameters.
     *
     * @return
     * The method or null if not found.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CMethod lookupMethod(CClass caller, String ident, CType[] params)
        throws UnpositionedError
    {
        CMethod method = self.lookupMethod(caller, ident, params);
        if (method != null) {
            return method;
        }

        CClassContext parent = getParentContext().getClassContext();
        if (parent != null) {
            return parent.lookupMethod(caller, ident, params);
        }
        else {
            return null;
        }
    }

    /**
     * lookupField
     * @param ident  the name of the field
     * @return a variable from an ident in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public CField lookupOuterField(String ident)
        throws UnpositionedError
    {
        CClassContext parent = getParentContext().getClassContext();

        if (parent != null) {
            CField field = parent.lookupField(ident);

            return field != null ? field : parent.lookupOuterField(ident);
        }

        return null;
    }

    /**
     * lookupField
     * @param ident  the name of the field
     * @return a variable from an ident in current context
     * @exception UnpositionedError this error will be positioned soon
     */
    public CField lookupField(String ident)
        throws UnpositionedError
    {
        CField field = getCClass().lookupField(ident);

        return field;
    }

    /**
     * lookupOuterLocalVariable
     * @param ident  the name of the outer variable
     * @return a variable from an ident in upperclass context
     */
    public JExpression lookupOuterLocalVariable(TokenReference ref,
                                                String ident)
    {
        if (parent instanceof CBodyContext ||
            parent instanceof CExpressionContext) {

            JLocalVariable var = parent.lookupLocalVariable(ident);
            if (var != null) {
                return new JOuterLocalVariableExpression(ref, var,
                                                         getCClass());
            }
            else {
                CClassContext ctxt = getParentContext().getClassContext();
                return (ctxt == null
                        ? null : ctxt.lookupOuterLocalVariable(ref, ident));
            }
        }
        return null;
    }

    /**
     * lookupLocalVariable
     * @param ident  the name of the local variable
     * @return a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident)
    {
        return null;
    }

    /**
     * getClassContext
     * @return the near parent of type CClassContext
     */
    public CClassContext getClassContext()
    {
        return this;
    }

    /**
     * Returns the nearest class parent in this context, which is of type
     * CSourceClass.
     * @return the nearest class parent
     */
    public CSourceClass getCClass()
    {
        return self;
    }

    /**
     * getMethod
     * @return the near parent of type CMethodContext
     */
    public CMethodContext getMethod()
    {
        return getParentContext().getMethodContext();
    }

    /**
     * getTypeDeclaration
     * @return the near parent of type CMethodContext
     */
    public JTypeDeclaration getTypeDeclaration()
    {
        return decl;
    }
}
