/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.InnerClassInfo;
import at.dms.compiler.UnpositionedError;

/**
 * ClassTypes that are referred to by name and that are resolved when needed.
 */
public class CClassNameType
    extends CClassType
{
    /**
     * null => checked
     */
    private String name;

    /**
     * InnerClassInfo for class that has to be resolved.
     */
    private InnerClassInfo extraInfo;

    /**
     * Construct a class type.
     *
     * @param name
     * The qualified name of the class, possible qualified.
     */
    public CClassNameType(String name)
    {
        this(name, null);
    }

    /**
     * Construct a class type.
     *
     * @param name
     * The qualified name of the class, possible qualified.
     * @param extraInfo
     * The extra innerclassinfo that described the classes being looked for.
     */
    public CClassNameType(String name, InnerClassInfo extraInfo)
    {
        super();

        this.extraInfo = extraInfo;
        this.name = name.intern();
    }

    /**
     * Sets the qualified name of the type being looking for.
     */
    public void setQualifiedName(String qname)
    {
        this.name = qname;
    }

    /**
     * Transforms this type to a string
     */
    public String toString()
    {
        return (name != null ? name.replace('/', '.') : super.toString());
    }

    /**
     *
     */
    public String getQualifiedName()
    {
        return (name == null ? super.getQualifiedName() : name);
    }

    /**
     * Returns the class object associated with this type
     *
     * If this type was never checked (read from class files)
     * check it!
     *
     * @return the class object associated with this type
     */
    public CClass getCClass()
    {
        if (name != null) {
            try {
                checkType(null);
            }
            catch (UnpositionedError cue) {
                setClass(new CBadClass(name));
                setQualifiedName(null);
                return getCClass();
            }
        }

        return super.getCClass();
    }

    /**
     * Try to resolve an inner type name!
     */
    private boolean tryResolveInnerType(CContext context)
        throws UnpositionedError
    {
        // May be a innerclass
        int index = name.lastIndexOf("/");
        if (index > 0) {
            try {
                CClassNameType outer =
                    new CClassNameType(name.substring(0, index)
                                       .intern());

                /* DBG: 
                System.err.print("Checking outer "
                                 + name.substring(0, index)
                                 + "...");
                /* */

                outer.checkType(context);

                /* DBG: 
                System.err.println(name.substring(0, index)
                                   + " succeeded!");
                /* */

                String innerName = name.substring(index + 1).intern();

                CClass innerClazz =
                    outer.getCClass().lookupMemberClass(innerName);

                if (innerClazz != null) {
                    setClass(innerClazz);
                    setQualifiedName(null);
                    return true;
                }
                else {
                    return false;
                }
            }
            catch (UnpositionedError ce) {
                if ( ce.getFormattedMessage().getDescription() ==
                     KjcMessages.TYPE_UNKNOWN ) {

                    return false;
                }
                else {
                    throw ce;
                }

                /* DBG: 
                if (name == null) {
                    ce.printStackTrace();
                }
                System.err.println("Failed: " + ce +
                                   (name == null ? " [qname == null]" : ""));
                /* */

            }
        }
        else {
            return false;
        }
    }

    /**
     * Check that type is valid.
     * Necessary to resolve String into java/lang/String.
     *
     * @param context
     * The context (may be be null.)
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void checkType(CContext context)
        throws UnpositionedError
    {
        if (name == null) {
            /* Class has already been checked and (hopefully) resolved. */
            return;
        }

        if (extraInfo != null) {
            /* Extra info is present, so it is exactly known which class
             * is being looked for, and the name is known to be fully qualified
             */
        }

        //DBG: System.err.println("Checking type " + name);

        String saved = name;

        /* First look it up as a member class */
        if (tryResolveInnerType(context)) {
            return;
        }

        if (context != null) {
            /* DBG:
            if (name == null) {
                System.err.println("Help, name is null and type is "
                                   + super.getCClass().getIdent() +
                                   " and orig qname is " + saved);
            }
            */
            CClassType type = context.lookupClass(name);
            if (type != null) {
                if ((type instanceof CClassNameType) && (context != null)
                    && (((CClassNameType)type).name != null)) {
                    
                    setQualifiedName(((CClassNameType)type).name);
                    // we have found the good one but we dont want to
                    // really load the class here.
                    //setClass(type.getCClass());
                    // setQualifiedName(null);
                }
                else {
                    setClass(type.getCClass());
                    setQualifiedName(null);
                }
                return;
            }
        }

        /* Try to look it up at top level */
        if (CTopLevel.hasClassFile(name)) {
            setClassFromName(name, extraInfo);
            setQualifiedName(null);
            return;
        }
        // DBG: System.err.println("Not found: " + name);
        throw new UnpositionedError(KjcMessages.TYPE_UNKNOWN, name);
    }

}
