/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashMap;
import java.util.Map;

import at.dms.classfile.InnerClassInfo;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;

/**
 * This class represents class types in the type structure.
 */
public class CClassType
    extends CType
{
    /**
     * Constant for an empty aray of CClassType.
     */
    public static final CClassType[] EMPTY = new CClassType[0];

    /**
     * Constant that identifies unresolved or undefined class.
     */
    private static final CClass BAC_CLASS = new CBadClass("<NOT YET DEFINED>");

    /**
     * Static map of {qualifiedname:String -> CClassType}
     */
    private static Map allCClassType = new HashMap(2000);

    /**
     * The class export that this is the type for.
     */
    private CClass clazz;

    /**
     * Construct a class type that is yet unresolved; this constructor is
     * protected and for use by subclasses.
     */
    protected CClassType()
    {
        super(TID_CLASS);

        this.clazz = BAC_CLASS;
    }

    /**
     * Construct a class type.
     *
     * @param clazz
     * The class that will be represented by this type.
     */
    public CClassType(CClass clazz)
    {
        super(TID_CLASS);

        this.clazz = clazz;
   
        if (!(this instanceof CArrayType)) {
            allCClassType.put(clazz.getQualifiedName(), this);
        }
    }

    /**
     * Looks up CClassType object for the qualifiedName using a static Map.
     */
    public static CClassType lookup(String className)
    {
        return lookup(className, null);
    }

    /**
     * Looks up CClassType object for the qualifiedName using a static Map.
     */
    public static CClassType lookup(String className, InnerClassInfo extraInfo)
    {
        if (className.indexOf('/') >= 0) {
            /* The name is qualified. */
            CClassType type = (CClassType)allCClassType.get(className);
            
            if (type == null) {
                type = new CClassNameType(className, extraInfo);
                allCClassType.put(className, type);
            }
            
            return type;
        }
        else {
            return new CClassNameType(className, extraInfo);
        }
    }

    /**
     * Returns whether the type is checked.
     */
    protected boolean isChecked()
    {
        return clazz != BAC_CLASS;
    }

    /**
     * Sets the CClass export belonging to this CClassType.
     */
    protected void setClass(CClass clazz)
    {
        this.clazz = clazz;
    }

    /**
     * Tries to set the CClass export belonging to this CClassType by loading
     * it from the top level.
     */
    protected void setClassFromName(String name, InnerClassInfo extraInfo)
    {
        this.clazz = CTopLevel.loadClass(name, extraInfo);
    }

    /**
     * equals.
     */
    public boolean equals(CType other)
    {
        if ( other.isClassType() && !other.isArrayType() ) {
            return ((CClassType)other).getCClass() == getCClass();
        }
        else {
            return false;
        }
    }

    /**
     * Transforms this type to a string.
     */
    public String toString()
    {
        if (clazz.isInnerClass()) {
            if (clazz.isSynthetic()) {
                return "$" + clazz.getIdent();
            }
            else {
                return clazz.getIdent();
            }
        }
        else {
            return getQualifiedName().replace('/', '.');
        }
    }

    /**
     * Transforms this type to a string.
     */
    protected void appendSignature(FastStringBuffer buffer)
    {
        buffer.append('L');
        buffer.append(getQualifiedName());
        buffer.append(';');
    }

    /**
     * @return the short name of this class
     */
    public String getIdent()
    {
        return getCClass().getIdent();
    }

    /**
     * Returns the qualified name of the class of this class type.
     */
    public String getQualifiedName()
    {
        return getCClass().getQualifiedName();
    }

    /**
     * Returns the size used in stack by value of this type.
     */
    public int getSize()
    {
        return 1;
    }

    /**
     * Check if a type is a reference.
     *
     * @return whether this is a reference type.
     */
    public boolean isReference()
    {
        return true;
    }

    /**
     * Check if a type is a class type.
     *
     * @return whether this is a class type.
     */
    public boolean isClassType()
    {
        return true;
    }

    /**
     * Returns the class object associated with this type
     *
     * If this type was never checked (read from class files)
     * check it!
     *
     * @return the class object associated with this type
     */
    public CClass getCClass()
    {
        // !!! graf 000213
        // !!! should have been checked (see JClassFieldExpression)
        assert(clazz != BAC_CLASS, "" + getClass().getName());
        // !!! graf 000213

        return clazz;

        // Not clear why one would cache. -- EJB
        // return getCachedCClass();
    }

    /**
     * Check that type is valid.
     *
     * @param context
     * The context (may be be null)
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void checkType(CContext context)
        throws UnpositionedError
    {
    }

    /**
     * isAssignableTo
     *
     * @param destination
     * The destination type.
     * @return
     * Assignment is valid ?
     */
    public boolean isAssignableTo(CType destination)
    {
        return ( destination.isClassType() && !destination.isArrayType() &&
                 getCClass().descendsFrom(destination.getCClass()) );
    }

    /**
     * Can this type be converted to the specified type by casting conversion
     * (JLS 5.5) ?
     *
     * @param destination
     * The destination type.
     * @return
     * true iff the casting conversion is valid.
     */
    public boolean isCastableTo(CType destination)
    {
        CClass clazz = getCClass();
        CClass toClazz = destination.getCClass();

        // test for array first because array types are classes

        if ( clazz.isInterface() ) {
            if (! destination.isClassType() ) {
                return false;
            }
            else if (toClazz.isInterface()) {
                // if T is an interface type and if T and S contain methods
                // with the same signature (JLS 8.4.2) but different return
                // types, then a compile-time error occurs.
                //!!! graf 000512: FIXME: implemt this test
                return true;
            }
            else if (! toClazz.isFinal() ) {
                return true;
            }
            else {
                return toClazz.descendsFrom(clazz);
            }
        }
        else {
            // this is a class type
            if (destination.isArrayType()) {
                return equals(CStdType.Object);
            }
            else if (! destination.isClassType()) {
                return false;
            }
            else if (toClazz.isInterface()) {
                if (! clazz.isFinal()) {
                    return true;
                }
                else {
                    return clazz.descendsFrom(toClazz);
                }
            }
            else {
                return ( clazz.descendsFrom(toClazz) ||
                         toClazz.descendsFrom(clazz) );
            }
        }
    }

    /**
     * implicitlyCastTo.
     *
     * @param destination
     * The expression type to cast to.
     * @return
     * Cast is valid?
     */
    public boolean implicitlyCastTo(CType destination)
    {
        return isAssignableTo(destination);
    }

    /**
     * Whether this is an exception that should be declared.
     */
    public boolean isCheckedException()
    {
        return ( !isAssignableTo(CStdType.RuntimeException)
                 && !isAssignableTo(CStdType.Error) );
    }

    /**
     * Before a compile process, cached info should be flushed.
     * Unclear why this method is needed.
    private CClass __getCachedCClass()
    {
        clazz = CTopLevel.reloadClass(clazz);

        if (clazz == null) {
            if (this == CStdType.Object) {
                throw new RuntimeException("java.lang.Object is not in "
                                           + "the classpath !!!");
            }
            clazz = CStdType.Object.getCClass();
        }

        return clazz;
    }
     */

}
