/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.Map;

import at.dms.compiler.UnpositionedError;

/**
 * This class represents a compilation unit
 */
public class CCompilationUnit
{
    /**
     * Package name of this compilation unit.
     */
    private final String packageName;

    /**
     * The imported classes.
     */
    private final JClassImport[] importedClasses;

    /**
     * The imported packages (import statement ending in ".*")
     */
    private final JPackageImport[] importedPackages;

    /**
     *
     */
    private Map loadedClasses;

    /**
     * Construct a compilation unit context.
     */
    public CCompilationUnit(String packageName,
                            JClassImport[] importedClasses,
                            JPackageImport[] importedPackages,
                            Map loadedClasses)
    {
        this.packageName = packageName;
        this.importedClasses = importedClasses;
        this.importedPackages = importedPackages;
        this.loadedClasses = loadedClasses;
    }

    /**
     * Returns a class according to imports or null if error occur.
     *
     * @return
     * A class according to imports or null if error occur.
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public CClassType lookupClass(String name)
        throws UnpositionedError
    {
        // $$$ USE A STRING BUFFER FOR IMPORT
        if (name.lastIndexOf('/') == -1) {
            // 6.5.4.1 Simple Type Names

            CClassType cl;

            // First look for a type declared by a single-type-import of by a
            // type declaration
            if ((cl = (CClassType)loadedClasses.get(name)) != null) {

                // If type is declared by a single-type-import, mark it as
                // used (max. 1)
                for (int i = 0; i < importedClasses.length; i++) {
                    if (name == importedClasses[i].getClassIdent()) {
                        importedClasses[i].setUsed();
                        break;
                    }
                }

                return cl;
            }

            // Otherwise, look for a type declared in another compilation unit
            // of this package
            if (packageName.length() == 0) {

                /* Package is default package. */
                if (CTopLevel.hasClassFile(name)) {
                    cl = CClassType.lookup(name);
                }
            }
            else {

                String  temp = packageName + '/' + name;
                if (CTopLevel.hasClassFile(temp)) {
                    cl = CClassType.lookup(temp);
                }
            }

            if (cl != null) {
                loadedClasses.put(name, cl);
            }
            else {
                // Otherwise, look for a type declared by EXACTLY ONE
                // import-on-demand declaration
                for (int i = 0; i < importedPackages.length; i++) {
                    String qualifiedName = (importedPackages[i].getName()
                                            + '/' + name).intern();

                    if (CTopLevel.hasClassFile(qualifiedName)) {
                        CClassType type = (CClassType)loadedClasses.get(name);

                        if (type != null
                            && !type.getQualifiedName().equals(qualifiedName)){
                            // Oops, the name is ambiguous (declared by more
                            // than one import-on-demand declaration)
                            throw new UnpositionedError
                                (KjcMessages.CUNIT_RENAME2, name);
                        }
                        loadedClasses.put(name,
                                          CClassType.lookup(qualifiedName));
                        importedPackages[i].setClassUsed(name);
                    }
                }
            }

            /* Now the name must be unique and found. */
            if ((cl = (CClassType)loadedClasses.get(name)) == null) {
                // The javadoc says to return null.
                return null;
                // XXX-EJB-test-removed
                //throw new UnpositionedError(KjcMessages.CLASS_UNKNOWN, name);
            }

            return cl;
        }
        else {
            // 6.5.4.2 Qualified Type Names: look directly at top
            if (!CTopLevel.hasClassFile(name)) {
                // The javadoc says to return null.
                return null;
                // XXX-EJB-test-removed
                //throw new UnpositionedError(KjcMessages.CLASS_UNKNOWN, name);
            }

            return CClassType.lookup(name);
        }
    }
}
