/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;

/**
 * Special local context for code inside a loop.
 *
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CContext
 */
public class CLoopContext
    extends CAbstractBodyContext
{
    /**
     * The loop statement which provides the context for analyzing code
     * inside it.
     */
    private final JLoopStatement self;

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * @param parent
     * The parent context, it must be different than null except if called by
     * the top level.
     * @param self
     * The loop statement which provides the context for analyzing code
     * inside it.
     */
    public CLoopContext(CBodyContext parent, JLoopStatement self)
    {
        super(parent);

        this.self = self;

        flowState = parent.flowState | FLOW_INLOOP;
    }

    /**
     * Adds a break from within this context to the specified statement.
     */
    protected void addForwardedBreak(JStatement breakTo, CBodyContext from)
        throws PositionedError
    {
        if (breakTo == self) {
            self.addBreak(from);
        }
        else {
            super.addForwardedBreak(breakTo, from);
        }
    }

    /**
     * Adds a continue from within this context to the specified statement.
     */
    protected void addForwardedContinue(JStatement continueTo,
                                        CBodyContext from)
        throws PositionedError
    {
        if (continueTo == self) {
            self.addContinue(from);
        }
        else {
            super.addForwardedContinue(continueTo, from);
        }
    }

    /**
     * Creates a BodyContext with this LoopContext as its parent.
     */
    public CBodyContext createBodyContext()
    {
        return new CBodyContext(this);
    }

    /**
     * Returns the nearest (from this context outward) statement to which
     * an unlabeled break should jump. For a CLoopContext, that is the
     * loop statement that provides the context for the loop.
     */
    public JStatement getNearestBreakableStatement()
    {
        return self;
    }

    /**
     * Returns the nearest (from this context outward) statement that an
     * unlabeled continue should continue with. For a CLoopContext, that is
     * the loop statement that provides the context for the loop.
     */
    public JStatement getNearestContinuableStatement()
    {
        return self;
    }
}
