/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.Set;
import java.util.Iterator;

import at.dms.classfile.InvokeinterfaceInstruction;
import at.dms.classfile.MethodInfo;
import at.dms.classfile.MethodRefInstruction;

/**
 * This class represents a class method.
 */
public abstract class CMethod
    extends CMember
{
    /**
     * Return type of method.
     */
    private CType returnType;

    /**
     * Parameter types of method.
     */
    private CType[] parameters;

    /**
     * Types of exceptions that can be thrown from method.
     */
    private CClassType[] exceptions;

    /**
     * Constructs a method member.
     *
     * @param owner  the owner of this method
     * @param modifiers the modifiers on this method
     * @param ident  the ident of this method
     * @param returnType the return type of this method
     * @param parameters the parameters type of this method
     * @param exceptions a list of all exceptions of the throws list
     * @param deprecated is this method deprecated
     */
    public CMethod(CClass owner,
                   int modifiers,
                   String ident,
                   CType returnType,
                   CType[] parameters,
                   CClassType[] exceptions,
                   boolean deprecated)
    {
        super(owner, modifiers, ident, deprecated);

        this.returnType = returnType;
        this.parameters = parameters;
        this.exceptions = exceptions;
    }

    /**
     * @return the interface
     */
    public CMethod getMethod()
    {
        return this;
    }

    /**
     * @return the type of this field
     */
    public final CType getReturnType()
    {
        return returnType;
    }

    /**
     * @return the type of this field
     */
    public final CType[] getParameters()
    {
        return parameters;
    }

    /**
     * @return the type of this field
     */
    public String getSignature()
    {
        return CType.genMethodSignature(returnType, parameters);
    }

    /**
     * @return the exceptions that can be thrown by this method
     */
    public CClassType[] getThrowables()
    {
        return exceptions;
    }

    /**
     * This method is used by initializers that knows throwables exceptions
     * only after body was checked.
     * @param throwables the exceptions that can be thrown by this method
     */
    public void setThrowables(Set throwables)
    {
        if (throwables != null) {
            Iterator iterator = throwables.iterator();
            int  count = 0;

            exceptions = new CClassType[throwables.size()];
            while (iterator.hasNext()) {
                exceptions[count++] = ((CThrowableInfo)iterator.next())
                    .getThrowable();
            }
        }
    }

    /**
     * Returns true iff this method is native.
     */
    public boolean isNative()
    {
        return hasFlag(getModifiers(), ACC_NATIVE);
    }

    /**
     * Returns true iff this method is abstract.
     */
    public boolean isAbstract()
    {
        return hasFlag(getModifiers(), ACC_ABSTRACT);
    }

    /**
     * Returns true iff this method is a constructor.
     */
    public boolean isConstructor()
    {
        return getIdent() == JAV_CONSTRUCTOR;
    }

    /**
     * equals
     * search if two methods have same signature
     * @param params  an other method
     */
    public boolean equals(CMethod method)
    {
        if (parameters.length != method.parameters.length) {
            return false;
        }

        for (int i = 0; i < parameters.length; i++) {
            if (!parameters[i].equals(method.parameters[i])) {
                return false;
            }
        }
        if (getIdent() != method.getIdent()) {
            return false;
        }

        return true;
    }
  
    /**
     * isMatching
     * search if this matches with the provided type parameters
     * @param actuals  method parameters
     */
    public boolean isMatching(String ident, CType[] actuals)
    {
        return parameters.length == actuals.length &&
            ident == getIdent() &&      
            isMatchingParams(actuals);
    }

    /**
     * isMatching
     * search if this matches with the provided type parameters
     * @param actuals  method parameters
     */
    public boolean isMatchingParams(CType[] actuals)
    {
        CType[] formals = getParameters();

        for (int i = 0; i < actuals.length; i++) {
            /*
              if (formals[i] == CStdType.Char && actuals[i] != CStdType.Char) {
              return false;
              }
            */

            /*
             * method invocation conversion = assigment conversion without
             * literal narrowing we just look at the type and do not consider
             * literal special case
             */
            if (!actuals[i].isAssignableTo(formals[i])) {
                return false;
            }
        }

        return true;
    }

    /**
     * isMoreSpecificThan
     * search if this method is more specific than other
     * @param other  the method to compare to
     * @return true if this method is more specific than other
     */
    public boolean isMoreSpecificThan(CMethod other)
    {
        CType[] theseFormals = getParameters();
        CType[] otherFormals = other.getParameters();

        for (int i = 0; i < otherFormals.length; i++) {
            /*
             * method invocation conversion = assigment conversion without
             * literal narrowing we just look at the type and do not consider
             * literal special case.
             */
            if (!theseFormals[i].isAssignableTo(otherFormals[i])) {
                return false;
            }
        }

        return true;
    }

    public String toString()
    {
        return getIdent() + "(" + parameters.length + ")";
    }

    /**
     * Generates a sequence of bytecode
     * @param code  the code sequence
     * @param nonVirtual force non-virtual dispatching
     */
    public void genCode(CodeSequence code, boolean nonVirtual)
    {
        if (getOwner().isInterface()) {
            int size = 0;
      
            for (int i = 0; i < parameters.length; i++) {
                size += parameters[i].getSize();
            }
      
            code.plantInstruction
                (new InvokeinterfaceInstruction(getPrefixName(),
                                                getIdent(),
                                                getSignature(),
                                                size + 1));
        }
        else {
            int opcode;

            if (isStatic()) {
                opcode = opc_invokestatic;
            }
            else if (nonVirtual || isPrivate()) {
                // JDK1.2 does not like ... || isFinal()
                opcode = opc_invokespecial;
            }
            else {
                opcode = opc_invokevirtual;
            }

            code.plantInstruction(new MethodRefInstruction(opcode,
                                                           getPrefixName(),
                                                           getIdent(),
                                                           getSignature()));
        }
    }

    /**
     * Generate the code in a class file
     */
    public MethodInfo genMethodInfo()
    {
        fail();
        String[] exceptions = new String[this.exceptions.length];
        for (int i = 0; i < this.exceptions.length; i++) {
            exceptions[i] = this.exceptions[i].getSignature();
        }

        return new MethodInfo((short)getModifiers(),
                              getIdent(),
                              getSignature(),
                              exceptions,
                              null,
                              isDeprecated(),
                              false);
    }

}
