/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.FastStringBuffer;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents java and kopi numericals types
 * Such as byte, short, int, long, float, double
 */
public class CNumericType
    extends CType
{

    /**
     * Constructor
     * @param typeID  the ident (int value) of this type
     */
    protected CNumericType(int typeID)
    {
        super(typeID);
        assert(typeID == TID_BYTE ||
               typeID == TID_CHAR ||
               typeID == TID_DOUBLE ||
               typeID == TID_FLOAT ||
               typeID == TID_INT ||
               typeID == TID_LONG ||
               typeID == TID_SHORT);
    }

    /**
     * Transforms this type to a string
     */
    public String toString()
    {
        switch (type) {
        case TID_BYTE:
            return "byte";
        case TID_CHAR:
            return "char";
        case TID_DOUBLE:
            return "double";
        case TID_FLOAT:
            return "float";
        case TID_INT:
            return "int";
        case TID_LONG:
            return "long";
        case TID_SHORT:
            return "short";
        default:
            fail();
            return null;
        }
    }

    /**
     * Transforms this type to a string
     */
    public String getSignature()
    {
        switch (type) {
        case TID_BYTE:
            return "B";
        case TID_CHAR:
            return "C";
        case TID_DOUBLE:
            return "D";
        case TID_FLOAT:
            return "F";
        case TID_INT:
            return "I";
        case TID_LONG:
            return "J";
        case TID_SHORT:
            return "S";
        default:
            fail();
            return null;
        }
    }

    /**
     * Transforms this type to a string
     */
    protected void appendSignature(FastStringBuffer buff)
    {
        char c;

        switch (type) {
        case TID_BYTE:
            c = 'B';
            break;
        case TID_CHAR:
            c = 'C';
            break;
        case TID_DOUBLE:
            c = 'D';
            break;
        case TID_FLOAT:
            c = 'F';
            break;
        case TID_INT:
            c = 'I';
            break;
        case TID_LONG:
            c = 'J';
            break;
        case TID_SHORT:
            c = 'S';
            break;
        default:
            c = '!';
            fail();
        }
        buff.append(c);
    }

    /**
     * Returns the size used in stack by value of this type
     */
    public int getSize()
    {
        return (type == TID_LONG || type == TID_DOUBLE) ? 2 : 1;
    }

    /**
     * Check if a type is a numeric type
     * @return is it a numeric type ?
     */
    public boolean isNumeric()
    {
        return true;
    }

    public boolean isCodedHasInt()
    {
        switch (type) {
        case TID_BYTE:
        case TID_CHAR:
        case TID_INT:
        case TID_SHORT:
            return true;
        }
        return false;
    }

    /**
     * Check if a type is an integer type
     * @return is it a integer type ?
     */
    public boolean isOrdinal()
    {
        switch (type) {
        case TID_BYTE:
        case TID_CHAR:
        case TID_SHORT:
        case TID_INT:
        case TID_LONG:
            return true;
        default:
            return false;
        }
    }

    /**
     * Check if a type is an integer type
     * @return is it a real number type ?
     */
    public boolean isFloatingPoint()
    {
        switch (type) {
        case TID_FLOAT:
        case TID_DOUBLE:
            return true;
        default:
            return false;
        }
    }

    /**
     * check that type is valid
     * necessary to resolve String into java/lang/String
     * @exception UnpositionedError this error will be positioned soon
     */
    public void checkType(CContext context) throws UnpositionedError {}

    /**
     * isAssignableTo
     * @param dest  the destination type
     * @return  assignment is valid ?
     */
    public boolean isAssignableTo(CType dest)
    {
        /* Each type is assignable to itself. */
        if (dest == this) {
            return true;
        }
        if (dest.isNumeric()) {
            /* non-const byte (signed) is not assignable to a char (unsigned)
             */
            if (type == TID_BYTE && dest.type == TID_CHAR) {
                return false;
            }
            /* non-const short and char are not assignable to eachother.
             * (JLS 5.1.2)
             */
            if (type == TID_SHORT || type == TID_CHAR) {
                return dest.type >= TID_INT;
            }
            else {
                return type < dest.type;
            }
        }
        else {
            return false;
        }
    }

    /**
     * isCastableTo
     * @param dest  the destination type
     * @return  cast is valid ?
     */
    public boolean isCastableTo(CType dest) {
        return dest.isNumeric();
    }

    /**
     * implicitlyCastTo
     * @param dest the expression type to cast to
     * @return cast is valid ?
     */
    public boolean implicitlyCastTo(CType dest)
    {
        if (this == dest) {
            return true;
        }
        else if (!dest.isNumeric()) {
            return false;
        }
        else {
            /* Warning: explicit use of case-label fallthrough. */
            switch (type) {
            case TID_BYTE:
                /* Nb: a byte is not implicitly castable to a char. See JLS
                 * 2nd edition.
                 */
                if (dest.type == TID_SHORT) {
                    return true;
                }
            case TID_SHORT:
            case TID_CHAR:
                if (dest.type == TID_INT) {
                    return true;
                }
            case TID_INT:
                if (dest.type == TID_LONG) {
                    return true;
                }
            case TID_LONG:
                if (dest.type == TID_FLOAT) {
                    return true;
                }
            case TID_FLOAT:
                if (dest.type == TID_DOUBLE) {
                    return true;
                }
            case TID_DOUBLE:
                break;
            default:
                fail();
            }
            return false;
        }
    }

    /**
     * Search the type corresponding to the unification of the two types.
     *
     * @param t1 the first type
     * @param t2 the second type
     * @return the corresponding type or null
    public static CType unifyNumeric(CType t1, CType t2)
    {
        assert(t1.isNumeric() && t2.isNumeric());
        return CNumericType.unifyNumericIn
            (t1, CNumericType.unifyNumericIn(t2, t1));
    }

    /**
     * Internal method: search the type corresponding to the unification of
     * the two types, but simpler than the public method.
     *
     * @param t1 the first type
     * @param t2 the second type
     * @return the corresponding type or null
    private static CType unifyNumericIn(CType t1, CType t2)
    {
        if (t1 == CStdType.Short && t2 == CStdType.Byte) {
            return CStdType.Short;
        }
        else if (t1 == CStdType.Char
                 && (t2 == CStdType.Char || t2 == CStdType.Byte)) {

            return CStdType.Char;
        }
        else {
            return null;
        }
    }
     */

    /**
     * Returns the unary-promoted type for the argument <code>type</code>.
     * This is done according to the rules defined in JLS2 section 5.6.1,
     * titled "Unary Numeric Promotion."
     *
     * @param type the type to promote.
     * @return the unary-promoted type for <code>type</code>.
     */
    public static CType unaryPromote(CType type)
    {
        if (type == CStdType.Byte || type == CStdType.Short ||
            type == CStdType.Char) {

            return CStdType.Integer;
        }
        else {
            return type;
        }
    }

    /**
     * Returns the type to which both operands in a binary expression must be
     * promoted <i>before</i> the operation is performed. This must only
     * be used for binary epxressions that are required by the JLS to perform
     * binary numeric promotion (such as the multiplicative operations).
     * The resulting type of the expression is also the promoted type.
     * Binary Numeric Promotion is described in JLS 2nd edition,
     * paragraph 5.6.2.
     *
     * @param left the type of the left expression in a binary expression.
     * @param right the type of the right expression in a binary expression.
     * @return the promoted type.
     */
    public static CType binaryPromote(CType left, CType right)
    {
        assert(left.isNumeric() && right.isNumeric());

        if (left == CStdType.Double || right == CStdType.Double) {
            return CStdType.Double;
        }
        if (left == CStdType.Float || right == CStdType.Float) {
            return CStdType.Float;
        }
        if (left == CStdType.Long || right == CStdType.Long) {
            return CStdType.Long;
        }
        else {
            return CStdType.Integer;
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param dest  the destination type
     * @param code  the code list
     */
    public void genCastTo(CNumericType dest, CodeSequence code) {
        if (dest != this) {
            switch (dest.type) {
            case TID_BYTE:
                genCastTo(CStdType.Integer, code);
                code.plantInstruction(new NoArgInstruction(opc_i2b));
                break;

            case TID_CHAR:
                genCastTo(CStdType.Integer, code);
                code.plantInstruction(new NoArgInstruction(opc_i2c));
                break;

            case TID_SHORT:
                genCastTo(CStdType.Integer, code);
                code.plantInstruction(new NoArgInstruction(opc_i2s));
                break;

            case TID_INT:
                switch (type) {
                case TID_BYTE:
                case TID_CHAR:
                case TID_SHORT:
                    break;
                case TID_INT:
                    fail();
                case TID_LONG:
                    code.plantInstruction(new NoArgInstruction(opc_l2i));
                    break;
                case TID_FLOAT:
                    code.plantInstruction(new NoArgInstruction(opc_f2i));
                    break;
                case TID_DOUBLE:
                    code.plantInstruction(new NoArgInstruction(opc_d2i));
                    break;
                default:
                    fail();
                }
                break;

            case TID_LONG:
                switch (type) {
                case TID_BYTE:
                case TID_CHAR:
                case TID_SHORT:
                case TID_INT:
                    code.plantInstruction(new NoArgInstruction(opc_i2l));
                    break;
                case TID_LONG:
                    fail();
                case TID_FLOAT:
                    code.plantInstruction(new NoArgInstruction(opc_f2l));
                    break;
                case TID_DOUBLE:
                    code.plantInstruction(new NoArgInstruction(opc_d2l));
                    break;
                default:
                    fail();
                }
                break;
 
            case TID_FLOAT:
                switch (type) {
                case TID_BYTE:
                case TID_CHAR:
                case TID_SHORT:
                case TID_INT:
                    code.plantInstruction(new NoArgInstruction(opc_i2f));
                    break;
                case TID_LONG:
                    code.plantInstruction(new NoArgInstruction(opc_l2f));
                    break;
                case TID_FLOAT:
                    fail();
                case TID_DOUBLE:
                    code.plantInstruction(new NoArgInstruction(opc_d2f));
                    break;
                default:
                    fail();
                }
                break;
 
            case TID_DOUBLE:
                switch (type) {
                case TID_BYTE:
                case TID_CHAR:
                case TID_SHORT:
                case TID_INT:
                    code.plantInstruction(new NoArgInstruction(opc_i2d));
                    break;
                case TID_LONG:
                    code.plantInstruction(new NoArgInstruction(opc_l2d));
                    break;
                case TID_FLOAT:
                    code.plantInstruction(new NoArgInstruction(opc_f2d));
                    break;
                case TID_DOUBLE:
                    fail();
                default:
                    fail();
                }
                break;

            default:
                fail();
            }
        }
    }
}
