/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.Map;

import at.dms.classfile.ClassFileFormatException;
import at.dms.classfile.FieldRefInstruction;
import at.dms.compiler.Compiler;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents the exported members of a class (inner classes,
 * methods and fields). It is build from a parsed files so values are
 * accessibles differently after build and after interface checked
 */
public class CSourceClass
    extends CClass
{
    /**
     * Prefix for fields in inner classes that store the values of local
     * variables for those inner classes that are defined inside methods.
     */
    private final static String OUTERLOCAL_FIELDPREFIX = "var$";

    /**
     * Outer variables (in local classes): Maps JLocalVariables to field names.
     */
    private Map outerFieldNames;

    /**
     * Outer variables (in local classes): Lists JLocalVariables in a
     * consistent order.
     */
    private List outerVars;

    /**
     * Number of synthetic fields.
     */
    private int countSyntheticFields;

    /**
     * Whether the initializer has been checked.
     */
    private boolean initializerChecked;

    /**
     * The syntax-tree counterpart of this exported source class definition.
     */
    private JTypeDeclaration counterPart;

    /**
     * Used to create unique identifiers for synthetic members or types
     * by incrementing a number.
     */
    private int syntheticIndexCounter = 0;

    /**
     * Used to create unique identifiers for anonymous classes by incrementing
     * a number.
     */
    private int anonymousIndexCounter = 0;

    /**
     * Constructs a class export from source
     */
    public CSourceClass(Compiler compiler,
                        CClass owner,
                        JTypeDeclaration counterPart,
                        TokenReference where,
                        int modifiers,
                        String ident,
                        String qualifiedName,
                        boolean deprecated)
    {
        super(compiler, owner, where.getFile(), modifiers, ident,
              qualifiedName, null, deprecated);
        this.counterPart = counterPart;
    }

    /**
     * Ends the definition of this class
     */
    public void close(CClassType[] interfaces,
                      CClassType superClass,
                      CField[] fields,
                      CMethod[] methods)
    {
        setSuperClass(superClass);
        super.close(interfaces, fields, methods);
    }

    /**
     * Returns true if initialization pass is finished for this class
     */
    public boolean isInitializerChecked()
    {
        return initializerChecked;
    }

    /**
     * Sets initialization pass to be finished
     */
    public void setInitializerChecked()
    {
        initializerChecked = true;
    }

    /**
     * Returns the syntax-tree counterpart of this exported source class
     * definition.
     */
    public JTypeDeclaration getCounterPart()
    {
        return counterPart;
    }

    /**
     */
    public int getNextSyntheticIndex()
    {
        return ++syntheticIndexCounter;
    }

    /**
     */
    public int getNextAnonymousIndex()
    {
        return ++anonymousIndexCounter;
    }

    public CSourceMethod addJMethod(JMethodDeclaration method)
        throws PositionedError
    {
        CEmptyTopLevelContext context =
            new CEmptyTopLevelContext(getCompiler());
        CClassContext classContext = context.createClassContext(this);

        CSourceMethod sourceMethod = method.checkInterface(classContext);
        method.checkStatement(classContext);
        sourceMethod.setSynthetic(true);
        addMethod(sourceMethod);
        return sourceMethod;
    }

    /**
     * add synthetic parameters to method def
     */
    public CType[] genConstructorArray(CType[] params)
    {
        /* EJB: always generate outer this type in constructor param list!
         * So don't check if it is really used inside the inner class.
         * Outerwise the local var indexes in the constructor are wrong.
         */
        boolean hasOuterThis = isInnerClass() && !isStatic();

        int size = params.length;

        if (hasOuterThis) {
            size++;
        }

        if (outerVars != null) {
            size += outerVars.size();
        }

        if (params.length == size) {
            return params;
        }
        else {
            CType[] result = new CType[size];
            int pos = 0;
            
            if (hasOuterThis) {
                result[pos++] = getOwner().getType();
            }

            if (outerVars != null) {
                Iterator iterator = outerVars.iterator();
                while (iterator.hasNext()) {
                    JLocalVariable var = (JLocalVariable)iterator.next();
                    result[pos++] = var.getType();
                }
            }

            for (int i = 0; i < params.length; i++) {
                result[pos++] = params[i];
            }
      
            return result;
        }

    }

    /**
     * add synthetic parameters to method call
     */
    public void genSyntheticParams(CodeSequence code, JExpression explicitThis)
    {
        /* EJB: always except outer this as a parameter to a constructor call
         * of a non-static inner class.
         * So don't check if it is really used inside the inner class.
         * Outerwise it doesn't match the signature of the constructor.
         */
        if (isInnerClass() && !isStatic()) {
            if (explicitThis == null) {
                code.plantLoadThis();
            }
            else {
                explicitThis.genCode(code);
            }
        }

        if (outerVars != null) {
            Iterator iterator = outerVars.iterator();
            while (iterator.hasNext()) {
                JLocalVariable var = (JLocalVariable)iterator.next();

                new JLocalVariableExpression(TokenReference.NO_REF, var)
                    .genCode(code);
            }
        }
    }

    /**
     * add synthetic parameters to method call
     */
    public void genSyntheticParamsFromExplicitSuper(JExpression[] params,
                                                    CodeSequence code)
    {
        if (isInnerClass() && !isStatic()) {
            JGeneratedLocalVariable var;
            var = new JGeneratedLocalVariable(null, 0, getOwner().getType(),
                                              JAV_OUTER_THIS, null);

            var.setPosition(1);
            new JLocalVariableExpression(TokenReference.NO_REF, var)
                .genCode(code);
        }
    }

    /**
     * Returns the size of the outer-local variables in the variable table.
     */
    public int getOuterLocalsSize()
    {
        int size = 0;

        /* Outers are the local variables and formal parameters of the
         * method in which the inner class is defined (if it is defined
         * inside a method.)
         */
        if (outerVars != null) {
            Iterator iterator = outerVars.iterator();
            while (iterator.hasNext()) {
                JLocalVariable ovar = (JLocalVariable)iterator.next();
                size += ovar.getType().getSize();
            }
        }

        return size;
    }


    /**
     * Gets the code to access outer local vars.
     */
    public JExpression getOuterLocalAccess(TokenReference ref,
                                           final JLocalVariable var,
                                           final CMethod constructor)
    {
        String name;
        CSourceField field;

        if (outerVars == null) {
            outerVars = new ArrayList(4);
            outerFieldNames = new HashMap();
        }

        if ((name = (String)outerFieldNames.get(var)) == null) {
            // local vars are uniq
            name = OUTERLOCAL_FIELDPREFIX + var.getIdent();
            outerFieldNames.put(var, name);
            outerVars.add(var);
            field = new CSourceField(this, 0, name, var.getType(), false);
            field.setSynthetic(true);
            addField(field);
            countSyntheticFields++;
        }
        else {
            field = (CSourceField)getField(name);
        }

        if (constructor != null) {
            final CSourceField ffield = field;
            JGeneratedLocalVariable local =
                new JGeneratedLocalVariable(null, 0,
                                            var.getType(), var.getIdent(),
                                            null)
                    {
                        public int getPosition()
                        {
                            return ( 1 /*this*/
                                     + 1 /* outer-this */
                                     + outerVars.indexOf(var));
                        }
                    };

            return new JCheckedExpression(ref,
                                          new JLocalVariableExpression(ref,
                                                                       local));
        }
        else {
            return new JClassFieldExpression(ref,
                                             new JThisExpression(ref,
                                                                 getCClass()),
                                             name);
        }
    }

    /**
     * add synthetic parameters to method def
     */
    public void genInit(CodeSequence code, int countLocals)
    {

        /* instance-this is always present */
        countLocals++;
        
        if (isInnerClass() && !isStatic()) {
            int genPosition = 1;
            JGeneratedLocalVariable var;
            if (hasOuterThis()) {
                /* If the outer-this field isn't used inside the inner class,
                 * there is an outer-this parameter in the constructor
                 * declaration but it isn't used inside the method.
                 */
                var = new JGeneratedLocalVariable(null, 0,
                                                  getOwner().getType(),
                                                  JAV_OUTER_THIS, null);
                var.setPosition(genPosition);
                code.plantLoadThis();
                var.genLoad(code);
                FieldRefInstruction instr;
                String outerSignature = getOwner().getType().getSignature();
                instr = new FieldRefInstruction(opc_putfield,
                                                getQualifiedName(),
                                                JAV_OUTER_THIS,
                                                outerSignature);
                code.plantInstruction(instr);
            }

            /* Outer-this parameter is always present in constructor signature,
             * even if unused.
             */
            genPosition += 1;
            countLocals += 1;

            /* Outers are the local variables and formal parameters of the
             * method in which the inner class is defined (if it is defined
             * inside a method.)
             */
            if (outerVars != null) {
                Iterator iterator = outerVars.iterator();
                while (iterator.hasNext()) {
                    JLocalVariable ovar = (JLocalVariable)iterator.next();
                    var = new JGeneratedLocalVariable(null, 0,
                                                      ovar.getType(),
                                                      ovar.getIdent(), null);
                    var.setPosition(genPosition);
                    genPosition += var.getType().getSize();
                    countLocals += var.getType().getSize();
                    code.plantLoadThis();
                    var.genLoad(code);
                    FieldRefInstruction instr;
                    String varSignature = var.getType().getSignature();
                    instr = new FieldRefInstruction(opc_putfield,
                                                    getQualifiedName(),
                                                    OUTERLOCAL_FIELDPREFIX
                                                    + var.getIdent(),
                                                    varSignature);
                    code.plantInstruction(instr);
                }
            }
        }
    }

    /**
     * Collect and generate class files
     */
    public void genCode(String destination)
        throws IOException, ClassFileFormatException
    {
        try {
            genClassFile(destination);
        }
        catch (at.dms.util.InternalError e) {
            System.err.println("Ident = " + getIdent());
            throw e;
        }
    }
}
