/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.TokenReference;
import at.dms.compiler.PositionedError;

/**
 * This class represents an exported member of a class (fields)
 */
public class CSourceField
    extends CField
{
    /**
     * Position in the field list in the class.
     */
    private int pos;

    /**
     * Whether this field is used by any non-synthetic source code.
     */
    private boolean used;

    /**
     * Whether this field is "fully declared".i
     * XXX: lookup and copy definition of "fully declared".
     */
    private boolean fullyDeclared;

    /**
     * If this is a private field, and an inner class of the owner wants
     * to access it for reading, a getter method must be generated and
     * is stored in this field so it can be re-used.
     */
    private CSourceMethod innerGetter;

    /**
     * If this is a private field, and an inner class of the owner wants
     * to access it for writing, a setter method must be generated and
     * is stored in this field so it can be re-used.
     */
    private CSourceMethod innerSetter;

    /**
     * Constructs a field export.
     *
     * @param owner  the owner of this field
     * @param decl  the field declaration in source code
     * @param modifiers the modifiers on this field
     * @param ident  the name of this field
     * @param type  the type of this field
     * @param deprecated is this field deprecated
     */
    public CSourceField(CClass owner,
                        int modifiers,
                        String ident,
                        CType type,
                        boolean deprecated)
    {
        super(owner, modifiers, ident, type, deprecated);
    }

    public boolean isUsed()
    {
        return used || !isPrivate() || isSynthetic();
    }

    public void setUsed()
    {
        used = true;
    }


    public CSourceMethod getGetterAccessor()
        throws PositionedError
    {
        if (innerGetter != null) {
            /* Getter is already created. */
            return innerGetter;
        }

        /* static <field-type> access$xxx(<owner-type> instance)
         * {
         *     return instance.<field-name>;
         * }
         */ 
        CSourceClass owner = (CSourceClass)getOwner();

        TokenReference where = TokenReference.NO_REF;
        JExpression instanceExpression = null;

        JFormalParameter[] params;

        if (isStatic()) {
            params = new JFormalParameter[0];
        }
        else {
            JFormalParameter instanceParam =
                new JFormalParameter(where,
                                     0,
                                     owner.getType(),
                                     "access$instance",
                                     true);
            params = new JFormalParameter[] { instanceParam };
            instanceExpression = new JLocalVariableExpression(where,
                                                              instanceParam);
        }


        JStatement getField =
            new JReturnStatement(where,
                                 new JClassFieldExpression(where,
                                                           instanceExpression,
                                                           getIdent()),
                                 null);

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_STATIC,
                                   getType(), // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   params,
                                   new CClassType[0],
                                   new JBlock(where,
                                              new JStatement[] { getField },
                                              null),
                                   null, null);

        return innerGetter = owner.addJMethod(method);
    }

    public CSourceMethod getSetterAccessor()
        throws PositionedError
    {
        if (innerSetter != null) {
            /* Setter is already created. */
            return innerSetter;
        }

        /* static void access$xxx(<owner-type> instance, <field-type> value)
         * {
         *     instance.<field-name> = value;
         * }
         */ 
        CSourceClass owner = (CSourceClass)getOwner();
        TokenReference where = TokenReference.NO_REF;
        JExpression instanceExpression = null;

        JFormalParameter[] params;

        JFormalParameter valueParam =
            new JFormalParameter(where,
                                 0, getType(), "access$value",
                                 true);

        if (isStatic()) {
            params = new JFormalParameter[] { valueParam };
        }
        else {
            JFormalParameter instanceParam =
                new JFormalParameter(where,
                                     0,
                                     owner.getType(),
                                     "access$instance",
                                     true);
            params = new JFormalParameter[] { instanceParam, valueParam };
            instanceExpression = new JLocalVariableExpression(where,
                                                              instanceParam);
        }

        JExpression valueExpression = new JLocalVariableExpression(where,
                                                                   valueParam);

        JExpression setField =
            new JAssignmentExpression(where,
                                      new JClassFieldExpression
                                          (where,
                                           instanceExpression,
                                           getIdent()),
                                      valueExpression);

        JStatement setFieldStatement =
            new JExpressionStatement(where, setField, null);

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_STATIC, // quote-unquote "ACC_PACKAGE"
                                   CStdType.Void, // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   params,
                                   new CClassType[0],
                                   new JBlock(where,
                                              new JStatement[] {
                                                  setFieldStatement },
                                              null),
                                   null, null);

        return innerSetter = owner.addJMethod(method);
    }

    /**
     * Set this source field as "fully declared".
     */
    public void setFullyDeclared(boolean fullyDeclared)
    {
        this.fullyDeclared = fullyDeclared && !isFinal();
    }

    /**
     *
     */
    public boolean isFullyDeclared()
    {
        return fullyDeclared;
    }

    public int getPosition()
    {
        return pos;
    }

    public void setPosition(int pos)
    {
        this.pos = pos;
    }
}
