/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.MethodInfo;
import at.dms.classfile.CodeInfo;

/**
 * This class represents an exported method declaration of a class or
 * interface.
 */
public class CSourceMethod
    extends CMethod
{
    /**
     * The syntax tree of the code of the method.
     */
    private JBlock body;

    /**
     * Whether this method is used by non-synthetic code.
     */
    private boolean used;

    /**
     * How heavily the code in this method must be optimized.
     */
    private int optimizingLevel;

    /**
     * If this is a private method, and an inner class of the owner wants
     * to call it, an accessor method must be generated and
     * is stored in this field so it can be re-used.
     */
    private CSourceMethod innerAccessor;

    /**
     * If a member class of the owner wants to call the super-method
     * of this method, an accessor method must be generated and
     * is stored in this field so it can be re-used.
     */
    private CSourceMethod superAccessor;

    /**
     * Constructs a source method export.
     *
     * @param owner
     * The owner of this method.
     * @param modifiers
     * The modifiers on this method.
     * @param ident
     * The ident of this method.
     * @param returnType
     * The return type of this method.
     * @param paramTypes
     * The parameter types of this method.
     * @param exceptions
     * A list of all exceptions in the throws list.
     * @param deprecated
     * Is this method deprecated.
     * @param body
     * The source code.
     * @param optimizingLevel
     * The optimization level.
     */
    public CSourceMethod(CClass owner,
                         int modifiers,
                         String ident,
                         CType returnType,
                         CType[] paramTypes,
                         CClassType[] exceptions,
                         boolean deprecated,
                         JBlock body,
                         int optimizingLevel)
    {
        super(owner, modifiers, ident, returnType, paramTypes, exceptions,
              deprecated);

        this.body = body;
        this.optimizingLevel = optimizingLevel;
    }

    public boolean isUsed()
    {
        return used || !isPrivate() || isSynthetic();
    }

    public void setUsed()
    {
        used = true;
    }

    /**
     * Generate the code in a class file
     */
    public MethodInfo genMethodInfo()
    {
        CClassType[] excs = getThrowables();
        String[]  exceptions = new String[excs.length];
        for (int i = 0; i < excs.length; i++) {
            exceptions[i] = excs[i].getQualifiedName();
        }

        if (body != null) {
            return new CMethodInfo((short)getModifiers(),
                                   getIdent(),
                                   getSignature(),
                                   exceptions,
                                   this,
                                   isDeprecated(),
                                   false);
        }
        else {
            return new MethodInfo((short)getModifiers(),
                                  getIdent(),
                                  getSignature(),
                                  exceptions,
                                  null,
                                  isDeprecated(),
                                  false);
        }
    }

    /**
     * Generates or returns an already-generated method that can access
     * this method from classes that should be able to access this method
     * per the JLS, but can't because the JVM doesn't allow it.
     * <p>
     * This is the case for inner classes calling a private method in an
     * outer class, or an outer class calling a private method in an inner
     * class.
     * <p>
     * The accessor method will have package-only access rights, which is okay
     * for inner classes, since they always reside in the same package as
     * the class itself.
     */
    public CSourceMethod getAccessor()
        throws PositionedError
    {
        if (innerAccessor != null) {
            /* Accessor is already created. */
            return innerAccessor;
        }

        CSourceClass owner = (CSourceClass)getOwner();
        if (isConstructor()) {
            innerAccessor = owner.addJMethod(makeConstructorAccessor());
        }
        else {
            innerAccessor = owner.addJMethod(makeNormalAccessor());
        }

        return innerAccessor;
    }

    /**
     * Generates or returns an already-generated method that can access
     * the super-method of this method from classes that should be able to
     * access this method per the JLS, but can't because the JVM doesn't
     * implement it.
     * <p>
     * This is the case for inner classes calling a super-method of an
     * outer class.
     * <p>
     * The accessor method will have package-only access rights, which is okay
     * for inner classes, since they always reside in the same package as
     * the class itself.
     */
    public CSourceMethod getSuperAccessor()
        throws PositionedError
    {
        if (superAccessor != null) {
            /* Accessor is already created. */
            return superAccessor;
        }

        CSourceClass owner = (CSourceClass)getOwner();
        return superAccessor = owner.addJMethod(makeSuperAccessor());
    }

    /**
     * Creates an accessor method to access this method normally.
     * Normally means: not as a constructor and not the super method of this
     * method.
     */
    private JMethodDeclaration makeNormalAccessor()
        throws PositionedError
    {
        /* If the method is an ordinary method:
         *
         * static <method-return-type> access$xxx(<owner-type> instance,
         *                                        <params...>)
         * {
         *     return instance.<method-name>(<params....>);
         * }
         *
         * // or, for static methods:
         *
         * static <method-return-type> access$xxx(<params...>)
         * {
         *     return <method-name>(<params....>);
         * }
         */
        CSourceClass owner = (CSourceClass)getOwner();
        TokenReference where = TokenReference.NO_REF;
        CType[] accessedParameters = getParameters();

        JFormalParameter[] newParameters;
        boolean hasInstance = false;
        JExpression instanceExpression = null;

        // Fixed: owner.isStatic() -> isStatic() 
        if (!isStatic()) {
            JFormalParameter instanceParam =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     owner.getType(),
                                     "access$instance",
                                     true);
            newParameters =
                new JFormalParameter[accessedParameters.length + 1];
            newParameters[0] = instanceParam;
            instanceExpression =
                new JLocalVariableExpression(where, newParameters[0]);
            hasInstance = true;
        }
        else {
            newParameters =
                new JFormalParameter[accessedParameters.length];
        }


        for (int i = 0; i < accessedParameters.length; i++) {
            int j = i;
            if (hasInstance) {
                j++;
            }
            newParameters[j] =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     accessedParameters[i], "p$" + i,
                                     true);
        }

        JExpression[] args = new JExpression[accessedParameters.length];
        for (int i = 0; i < args.length; i++) {
            int j = i;
            if (hasInstance) {
                j++;
            }
            args[i] = new JLocalVariableExpression(where,
                                                   newParameters[j]);
        }

        JExpression callMethod =
            new JMethodCallExpression
                (where,
                 new JNameExpression(where,
                                     instanceExpression, // <instance>
                                     getIdent()),
                 args);
                                      

        JStatement callMethodStatement;

        if (getReturnType() == CStdType.Void) {
            callMethodStatement = new JExpressionStatement(where,
                                                           callMethod, null);
        }
        else {
            callMethodStatement = new JReturnStatement(where,
                                                       callMethod, null);
        }

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_STATIC,
                                   getReturnType(), // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   newParameters,
                                   getThrowables(),
                                   new JBlock(where,
                                              new JStatement[] {
                                                  callMethodStatement },
                                              null),
                                   null, null);

        return method;
    }

    /**
     * Creates an accessor method to access this the super-method of this
     * method. This construct is used for <code>ClassName.super</code>
     * constructs.
     */
    private JMethodDeclaration makeSuperAccessor()
        throws PositionedError
    {
        System.err.println("Making super-accessor!");
        /* If the method is an ordinary method:
         *
         * final <method-return-type> access$<owner-name>$xxx(<params...>)
         * {
         *     return super.<method-name>(<params....>);
         * }
         */
        CSourceClass owner = (CSourceClass)getOwner();
        TokenReference where = TokenReference.NO_REF;
        CType[] accessedParameters = getParameters();

        /* The only reason for this parameter is to avoid clashes with
         * method signatures in subclasses.
         */
        JFormalParameter selectorParam =
            new JFormalParameter(where,
                                 ACC_FINAL,
                                 owner.getType(),
                                 "access$selector",
                                 true);
        JFormalParameter[] newParameters =
            new JFormalParameter[accessedParameters.length + 1];
        newParameters[0] = selectorParam;


        for (int i = 0; i < accessedParameters.length; i++) {
            newParameters[i + 1] =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     accessedParameters[i], "p$" + i,
                                     true);
        }

        JExpression[] args = new JExpression[accessedParameters.length];
        for (int i = 0; i < args.length; i++) {
            args[i] = new JLocalVariableExpression(where,
                                                   newParameters[i + 1]);
        }

        JExpression callMethod =
            new JMethodCallExpression
                (where,
                 new JNameExpression(where,
                                     new JSuperExpression(where),
                                     getIdent()),
                 args);
                                      

        JStatement callMethodStatement;

        if (getReturnType() == CStdType.Void) {
            callMethodStatement = new JExpressionStatement(where,
                                                           callMethod, null);
        }
        else {
            callMethodStatement = new JReturnStatement(where,
                                                       callMethod, null);
        }

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_FINAL,
                                   getReturnType(), // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   newParameters,
                                   getThrowables(),
                                   new JBlock(where,
                                              new JStatement[] {
                                                  callMethodStatement },
                                              null),
                                   null, null);

        return method;
    }

    /**
     * Creates an accessor method to access this method as a constructor.
     * XXX-EJB: this doesn't work right; creating a static method sounds
     * nice, but it doesn't work if there is a subclass that needs to
     * access this private contstructor.
     * There are two solutions: make the constructor public or package-access,
     * or make the accessor a constructor with a special magic type as
     * an extra argument to ensure that no normal constructor signature could
     * clash with this one (this is how javac and jikes do it)
     */
    private JMethodDeclaration makeConstructorAccessor()
        throws PositionedError
    {
        /*
         * If the method is a constructor in an inner class:
         *
         * static <owner-type> access$xxx(<outer-type> outerInstance,
         *                                 <params...>)
         * {
         *     return outerInstance.new <owner-type>(<params....>);
         * }
         *
         * If the method is a constructor in a top-level class:
         * static <owner-type> access$xxx(<params...>)
         * {
         *     return new <owner-type>(<params....>);
         * }
         */ 
        CSourceClass owner = (CSourceClass)getOwner();
        CSourceClass outerClass = (CSourceClass)owner.getOwner();

        TokenReference where = TokenReference.NO_REF;

        CType[] origParameters = getParameters();
        JFormalParameter[] syntheticParameters;
        boolean hasExplicitThis = false;
        JExpression explicitThisExpression = null;

        if (owner.hasOuterThis()) {
            JFormalParameter outerInstanceParam =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     outerClass.getType(),
                                     "access$outerInstance",
                                 true);
            syntheticParameters =
                new JFormalParameter[origParameters.length + 1];
            syntheticParameters[0] = outerInstanceParam;
            explicitThisExpression =
                new JLocalVariableExpression(where, syntheticParameters[0]);
            hasExplicitThis = true;
        }
        else {
            syntheticParameters =
                new JFormalParameter[origParameters.length];
        }


        for (int i = 0; i < origParameters.length; i++) {
            int j = i;
            if (hasExplicitThis) {
                j++;
            }
            syntheticParameters[j] =
                new JFormalParameter(where,
                                     ACC_FINAL,
                                     origParameters[i], "p$" + i,
                                     true);
        }

        JExpression[] args = new JExpression[origParameters.length];

        for (int i = 0; i < args.length; i++) {
            int j = i;
            if (hasExplicitThis) {
                j++;
            }
            args[i] = new JLocalVariableExpression(where,
                                                   syntheticParameters[j]);
        }

        JExpression newObject =
            new JNewObjectExpression
                (where,
                 owner.getType(),
                 explicitThisExpression,
                 null,
                 args);

        JStatement returnNewObjectStatement =
            new JReturnStatement(where, newObject, null);

        JMethodDeclaration method = 
            new JMethodDeclaration(where,
                                   ACC_STATIC,
                                   owner.getType(), // return type
                                   ACCESSOR_PREFIX
                                   + owner.getNextSyntheticIndex(),
                                   syntheticParameters,
                                   getThrowables(),
                                   new JBlock(where,
                                              new JStatement[] {
                                                  returnNewObjectStatement },
                                              null),
                                   null, null);

        return method;
    }

    /**
     * @return the signature of this method.
     */
    public String getSignature()
    {
        CType[] params = getParameters();
        if (getOwner().isInnerClass() && isConstructor()) {
            params = ((CSourceClass)getOwner()).genConstructorArray(params);
        }
        return CType.genMethodSignature(getReturnType(), params);
    }

    /**
     * Generates code for this method.
     */
    public CodeInfo genCode()
    {
        CodeSequence code = CodeSequence.getCodeSequence();
    
        code.setLineNumber(body.getTokenReference().getLine());
        body.genCode(code);
    
        if (code.getPC() == 0 || 
            (code.getInstructionAt(code.getPC() - 1).getOpcode()
             != opc_return)
            && getReturnType() == CStdType.Void) {

            code.plantInstruction(new NoArgInstruction(opc_return));
        }

        CodeInfo info =
            new CodeInfo(code.getInstructionArray(),
                         code.getHandlers(),
                         optimizingLevel >  1 ? null : code.getLineNumbers(),
                         null);    //seq.getLocalVariableInfos());
        code.release();
        body = null;

        CType[] parameters = getParameters();
        int  paramCount = 0;

        for (int i = 0; i < parameters.length; i++) {
            paramCount += parameters[i].getSize();
        }
        paramCount += getReturnType().getSize();
        paramCount += isStatic() ? 0 : 1;

        info.setParameterCount(paramCount);

        if (optimizingLevel >= 10) {
            return at.dms.backend.Optimizer.optimize
                (null, info, new at.dms.backend.BackendOptions());
        }
        else if (optimizingLevel >= 1) {
            return at.dms.optimize.Optimizer.optimize(info, optimizingLevel);
        }
        else {
            return info;
        }
    }
}
