/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashSet;
import java.util.Set;

import at.dms.compiler.PositionedError;
import at.dms.compiler.UnpositionedError;

/**
 * Specialized local context during checkBody inside a switch self.
 *
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CContext
 */
public class CSwitchBodyContext
    extends CAbstractBodyContext
{
    /**
     * Whether the switch contains a default: label.
     */
    private boolean defaultExist;

    /**
     * The names of the case caseLabels inside the switch self.
     */
    private Set caseLabels = new HashSet();

    /**
     * The switch self that provides the context for this switch context.
     */
    private JSwitchStatement self;

    /**
     * The (ordinal) type of the switch selector, which cannt be bigger than
     * CStdType.Integer (either that one, or Byte, Char or Short)
     */
    private CType switchType;

    /**
     * Construct a block context, it supports local variable allocation
     * throw self and return self.
     *
     * @param parent
     * The parent context, it must be different from null except if called by
     * the top level.
     *
     * @param switchType
     * the type of switch (byte, short, char, int)
     */
    public CSwitchBodyContext(JSwitchStatement self,
                              CBodyContext parent,
                              CType switchType)
    {
        super(parent);
        this.switchType = switchType;
        this.self = self;
        flowState = parent.flowState & FLOW_INSWITCH;
    }

    /**
     * close.
     */
    public void close()
    {
    }

    /**
     * Adds a break from within this context to the specified self.
     */
    protected void addForwardedBreak(JStatement breakTo, CBodyContext from)
        throws PositionedError
    {
        if (breakTo == self) {
            self.addBreak(from);
        }
        else {
            super.addForwardedBreak(breakTo, from);
        }
    }

    /**
     * Add a default label to this switch.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void addDefault()
        throws UnpositionedError
    {
        if (defaultExist) {
            throw new UnpositionedError(KjcMessages.SWITCH_DEFAULT_DOUBLE);
        }
        defaultExist = true;
    }

    /**
     * Add a label to this switch and check it is a new one.
     *
     * @param lit
     * The literal value of this label.
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public void addLabel(JOrdinalLiteral lit)
        throws UnpositionedError
    {
        if (!lit.isAssignableTo(switchType)) {
            throw new UnpositionedError(KjcMessages.SWITCH_LABEL_OVERFLOW,
                                        lit.getValue());
        }

        if (!caseLabels.add(lit.getValue())) {
            throw new UnpositionedError(KjcMessages.SWITCH_LABEL_EXIST,
                                        lit.getValue());
        }
    }

    /**
     * @return true is a default label has already happen
     */
    public boolean defaultExists()
    {
        return defaultExist;
    }

    /**
     * remove a label from the list of caseLabels
     * @param label  the name to be removed
     */
    public JStatement getNearestBreakableStatement()
    {
        return self;
    }
}
