/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

import at.dms.classfile.ClassInfo;
import at.dms.classfile.ClassPath;
import at.dms.classfile.InnerClassInfo;
import at.dms.compiler.Compiler;
import at.dms.compiler.PositionedError;
import at.dms.util.Utils;

/**
 * This class implements the conceptual directory structure for .class files.
 */
public final class CTopLevel
    extends Utils
{
    /**
     * The compiler in use.
     */
    private static Compiler compiler;

    /**
     * The fake Compiler instance that is used as the 'Compiler' for binary
     * classes loaded from the classpath.
     */
    private static final Compiler BINARYCLASS_COMPILER =
        new Compiler() {
                public boolean run(String[] args)
                {
                    return true;
                }
                
                public void reportTrouble(PositionedError a)
                {
                }
                
                public boolean parseComments()
                {
                    return true;
                }
                
                public boolean verboseMode()
                {
                    return false;
                }
            };
    
    /**
     * All classes loaded in a compilation session.
     */
    private static Map allLoadedClasses = new HashMap(2000);

    /**
     * Loads class definition from .class file.
     *
     * @param fqName
     * The fully qualified name of the class.
     * @param innerInfo
     * An InnerClassInfo object providing more information about the class
     * to load. If <code>null</code>, a class at package level will be
     * loaded. An InnerClassInfo object contains the original name and
     * modifiers from the source code of an inner class, because these
     * are not the ones the are present in the class file, since the
     * JVM doesn't know about inner classes.
     */
    public static CClass loadClass(String fqName, InnerClassInfo innerInfo)
    {
        CClass clazz = (CClass)allLoadedClasses.get(fqName);

        if (clazz != null) {
            // look in cache
            if (clazz != CClass.CLS_UNDEFINED) {
                if (clazz instanceof CBinaryClass) {
                     InnerClassInfo clazzInfo =
                         ((CBinaryClass)clazz).getExtraInfo();

                     if (clazzInfo == null) {
                         if (innerInfo == null) {
                             return clazz;
                         }
                     }
                     else {
                         if (clazzInfo.equals(innerInfo) ) {
                             return clazz;
                         }
                     }
                    // XXX: use checked exception and be more informative.
                    throw new RuntimeException("InnerClassInfo does not match "
                                               + "with InnerClassInfo found "
                                               + "earlier for the same "
                                               + "class.");
                }
                return clazz;
            }
            else {
                return null;
            }
        }
        else {
            ClassInfo file = ClassPath.getClassInfo(fqName, true);

            if (file == null) {
                clazz = new CBadClass(fqName);
            }
            else {
                try {
                    clazz = new CBinaryClass(BINARYCLASS_COMPILER, file,
                                             innerInfo);
                }
                catch (RuntimeException r) {
                    System.err.println("Exception in class " + fqName);
                    throw r;
                }
            }

            allLoadedClasses.put(fqName, clazz);

            return clazz;
        }
    }

    /**
     * Reloads the specified class if this class has been defined during a
     * previous compilation session.
     */
    public static CClass reloadClass(CClass clazz)
    {
        if (clazz == null) {
            return null;
        }
        else if (clazz.getCompiler() == compiler) {
            // Same compilation session: keep it
            return clazz;
        }
        else {
            // Loaded in a previous compilation session: reload it
            return loadClass(clazz.getQualifiedName(), null);
        }
    }

    /**
     * Adds a source class to the set of all loaded classes.
     *
     * @param CClass
     * A class to add (must be a CSourceClass).
     * @return
     * false if name exists for source class as source class in an other file.
     */
    public static boolean addSourceClass(CClass cl)
    {
        assert(cl instanceof CSourceClass);

        CClass last = (CClass)allLoadedClasses.put(cl.getQualifiedName(), cl);
        return (last == null)
            || (cl.getOwner() != null)
            || !(last instanceof CSourceClass);
    }

    /**
     * Returns whether a class file that contain the named class named can
     * be found.
     *
     * @param name
     * The name of the class file.
     * @return
     * Whether a class file that contain the named class named can be found.
     */
    public static boolean hasClassFile(String name)
    {
        // already loaded or parsed ?
        if (allLoadedClasses.containsKey(name)) {
            return true;
        }
        else {
            return ClassPath.hasClassFile(name);
        }
    }

    /**
     * Removes all source classes.
     */
    public static void initSession(Compiler compiler)
    {
        CTopLevel.compiler = compiler;

        Iterator iterator = allLoadedClasses.values().iterator();
        while (iterator.hasNext()) {
            CClass clazz = (CClass)iterator.next();
            if (clazz instanceof CSourceClass) {
                iterator.remove();
            }
        }
    }

    /**
     * Ends session.
     */
    public static void endSession(Compiler compiler)
    {
        CTopLevel.compiler = null;
    }
}
