/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import at.dms.compiler.TokenReference;

/**
 * Special local context for checking code inside a try/catch statement.
 */
public class CTryContext
    extends CAbstractBodyContext
{
    /**
     * The throwables of the block containing the try statement, just 
     * before the try statement is checked.
     */
    private Set originalThrowables;

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     *
     * @param parent
     * The parent context, it must be different than null except if called by
     * the top level.
     */
    CTryContext(CBodyContext parent)
    {
        super(parent);

        originalThrowables = parent.getThrowables();
        throwables = new HashSet();
    }

    /**
     * Creates a BodyContext with this TryContext as its parent.
     */
    public CBodyContext createBodyContext()
    {
        return new CBodyContext(this);
    }

    /**
     * Closes this TryContext. This involves adding the throwables of
     * the try statement to the throwables of the containing block.
     */
    public void close(TokenReference ref)
    {
        if (throwables != null && throwables.size() > 0) {
            originalThrowables.addAll(throwables);
        }
        throwables = originalThrowables;

        super.close(ref);
    }

    /**
     * Adds throwables to the set that can be thrown from this try context.
     */
    public void addThrowables(Set newThrowables)
    {
        throwables.addAll(newThrowables);
    }

    /**
     * Clears the set of throwables that can be thrown from this try context.
     * Make sure that they are merged earlier with the outside context if
     * that is needed.
     */
    public void clearThrowables()
    {
        throwables = null;
    }

    /**
     * Replaces the set of throwables that can be thrown from this try context.
     * Necessary if a finally statement does not complete normally, so it
     * invalidates any exception throwables from the try clause.
     */
    public void replaceThrowables(Set throwables)
    {
        this.throwables = throwables;
    }
}
