/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Special local context for checking code inside a try/finally statement.
 */
public class CTryFinallyContext
    extends CTryContext
{
    /**
     * Set of statements that are breaked from inside the context to the try
     * statement itself or outer contexts of it.
     */
    private List breakedStatements = new ArrayList(4);

    /**
     * Set of statements that are continued from inside the context to the try
     * statement itself or outer contexts of it.
     */
    private List continuedStatements = new ArrayList(2);

    /**
     * Construct a specialized context for try/finally statement bodies.
     *
     * @param parent
     * The parent context, it must be different than null except if called by
     * the top level.
     */
    CTryFinallyContext(CBodyContext parent)
    {
        super(parent);
    }

   /**
     * Forward breaks from this try context to the parent context. This
     * is done when the breaks are not cancelled by a finally statement
     * that doesn't complete.
     */
    public void passBreakAndContinuesToParent()
        throws PositionedError
    {
        if (parent instanceof CAbstractBodyContext) {
            CAbstractBodyContext parent = (CAbstractBodyContext)this.parent;
            Iterator breaks = breakedStatements.iterator();
            while (breaks.hasNext()) {
                JStatement breakStatement = (JStatement)breaks.next();
                CBodyContext breakContext = (CBodyContext)breaks.next();
                parent.addForwardedBreak(breakStatement, breakContext);
            }
        }
        else {
            throw new RuntimeException("Assertion failed, no unforwarded "
                                       + "breaks should exist at top method "
                                       + "level.");
        }
        clearBreakAndContinues();
    }

    /**
     * Adds a break from within this context to the specified statement.
     */
    protected void addForwardedBreak(JStatement breakTo, CBodyContext from)
        throws PositionedError
    {
        breakedStatements.add(breakTo);
        breakedStatements.add(from);
    }

    /**
     * Adds a continue from within this context to the specified statement.
     */
    protected void addForwardedContinue(JStatement continueTo,
                                        CBodyContext from)
        throws PositionedError
    {
        continuedStatements.add(continueTo);
        continuedStatements.add(from);
    }

    /**
     * Clears the set of breaks from withing this context; this is used when
     * a finally clause invalidates the breaks within a try or catch clause.
     */
    public void clearBreakAndContinues()
    {
        breakedStatements.clear();
        continuedStatements.clear();
    }
}
