/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.ArrayList;
import java.util.List;

import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;
import at.dms.util.Utils;

/**
 * Root for type abstraction hierarchy.
 */
public abstract class CType
    extends Utils
    implements Constants
{
    /**
     * VM type signature parser instance.
     */
    protected final static SignatureParser signatureParser =
        new SignatureParser();

    /**
     * Type identifier.
     */
    protected int type;

    /**
     * Constructs a type baseclass with the general type identifier
     * <code>typeID</code>.
     */
    protected CType(int typeID)
    {
        this.type = typeID;
    }

    /**
     * Default equals for type abstractions.
     */
    public boolean equals(CType other)
    {
        return this == other;
    }

    /**
     * Transforms this type to a string.
     */
    public abstract String toString();

    /**
     * Transforms this type to a string
     * @return the class file form of thid type
     */
    public String getSignature()
    {
        FastStringBuffer buff = new FastStringBuffer();

        appendSignature(buff);
        String s = buff.toString();

        return s;
    }

    /**
     * Transforms this type to a string.
     */
    protected abstract void appendSignature(FastStringBuffer buff);

    /**
     * Returns the size used in stack by value of this type.
     */
    public abstract int getSize();

    /**
     * Check if a type is a numeric type.
     *
     * @return is it a numeric type?
     */
    public boolean isNumeric()
    {
        return false;
    }

    /**
     * ?
     */
    public boolean isCodedHasInt()
    {
        return false;
    }

    /**
     * Check if a type is an integer type.
     * @return is it a integer type?
     */
    public boolean isOrdinal()
    {
        return false;
    }

    /**
     * Check if a type is an integer type.
     *
     * @return is it a real number type?
     */
    public boolean isFloatingPoint()
    {
        return false;
    }

    /**
     * Check if a type is a class type.
     *
     * @return is it a subtype of ClassType?
     */
    public boolean isPrimitive()
    {
        return isNumeric() || (this == CStdType.Boolean);
    }

    /**
     * Check if a type is a class type.
     *
     * @return is it a subtype of ClassType?
     */
    public boolean isReference()
    {
        return false;
    }

    /**
     * Check if a type is a class type.
     *
     * @return is it a subtype of ClassType?
     */
    public boolean isClassType()
    {
        return false;
    }

    /**
     * @return is this type an array?
     */
    public boolean isArrayType()
    {
        return false;
    }

    /**
     * @return true if this type is valid.
     */
    public boolean checked()
    {
        return true;
    }

    /**
     * Returns the ID of this type.
     */
    public int getTypeID()
    {
        return type;
    }

    /**
     * check that type is valid.
     *
     * necessary to resolve String into java/lang/String
     *
     * @exception UnpositionedError
     * This error will be positioned soon.
     */
    public abstract void checkType(CContext context)
        throws UnpositionedError;

    /**
     * Can this type be converted to the specified type by casting conversion
     * (JLS 5.5) ?
     *
     * @param dest
     * The destination type.
     *
     * @return
     * true iff the casting conversion is valid.
     */
    public abstract boolean isCastableTo(CType dest);

    /**
     * implicitlyCastTo.
     *
     * @param dest the expression type to cast to
     * @return cast is valid ?
     */
    public boolean implicitlyCastTo(CType dest)
    {
        return dest.equals(this);
    }

    /**
     * Can this type be converted to the specified type by assignment
     * conversion (JLS 5.2) ?
     *
     * @param dest
     * The destination type.
     *
     * @return
     * true iff the casting conversion is valid.
     */
    public abstract boolean isAssignableTo(CType dest);

    /**
     * @return true if this type corrspond to a checked exception
     */
    public boolean isCheckedException()
    {
        return false;
    }

    /**
     * @return the object class of this type
     */
    public CClass getCClass()
    {
        fail();
        return null;
    }

    /**
     * Returns the opcode to load a local variable of this type.
     */
    public int getLoadOpcode()
    {
        switch (type) {
        case TID_CLASS:
        case TID_ARRAY:
            return opc_aload;
        case TID_DOUBLE:
            return opc_dload;
        case TID_FLOAT:
            return opc_fload;
        case TID_LONG:
            return opc_lload;
        case TID_BYTE:
        case TID_SHORT:
        case TID_CHAR:
        case TID_INT:
        case TID_BOOLEAN:
            return opc_iload;
        default:
            throw new RuntimeException("INTERNAL ERROR: " + type);
        }
    }

    /**
     * Returns the opcode to store a local variable of this type.
     */
    public int getStoreOpcode()
    {
        switch (type) {
        case TID_CLASS:
        case TID_ARRAY:
            return opc_astore;
        case TID_DOUBLE:
            return opc_dstore;
        case TID_FLOAT:
            return opc_fstore;
        case TID_LONG:
            return opc_lstore;
        case TID_BYTE:
        case TID_SHORT:
        case TID_CHAR:
        case TID_INT:
        case TID_BOOLEAN:
            return opc_istore;
        default:
            throw new RuntimeException("INTERNAL ERROR: " + type);
        }
    }

    /**
     * Returns the opcode to return a value of this type.
     */
    public int getReturnOpcode()
    {
        switch (type) {
        case TID_VOID:
            return opc_return;
        case TID_CLASS:
        case TID_ARRAY:
            return opc_areturn;
        case TID_DOUBLE:
            return opc_dreturn;
        case TID_FLOAT:
            return opc_freturn;
        case TID_LONG:
            return opc_lreturn;
        case TID_BYTE:
        case TID_SHORT:
        case TID_CHAR:
        case TID_INT:
        case TID_BOOLEAN:
            return opc_ireturn;
        default:
            throw new RuntimeException("INTERNAL ERROR: " + type);
        }
    }

    /**
     * Parse a java type signature.
     *
     * Attempts to parse the provided string as if it started with the Java
     * VM-standard signature for a type.
     */
    public static CType parseSignature(String signature)
    {
        return signatureParser.parseSignature(signature);
    }

    /**
     * Returns an array of types represented by the type signature.
     *
     * The return type is the last element of the array.
     */
    public static CType[] parseMethodSignature(String sig)
    {
        return signatureParser.parseMethodSignature(sig);
    }

    /**
     * Generate a method signature.
     */
    public static String genMethodSignature(CType returnType,
                                            CType[] parameters)
    {
        FastStringBuffer buff = new FastStringBuffer();

        buff.append('(');

        for (int i = 0; i < parameters.length; i++) {
            parameters[i].appendSignature(buff);
        }
        buff.append(')');

        returnType.appendSignature(buff);

        String s = buff.toString();
        return s;
    }
}
