/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;

import at.dms.classfile.*;
import at.dms.util.Utils;

/**
 * Code sequence is used as a bag to hold lists of instructions
 * until it is time to put them into a class file.
 *
 * Note: Local variables are currently not used.
 */
public final class CodeSequence
    extends Utils
    implements at.dms.classfile.Constants
{
    /**
     * Helper class to resolve labels.
     */
    class LabelResolvingTransformer
        implements AccessorTransformer
    {
        public InstructionAccessor transform(InstructionAccessor accessor,
                                             AccessorContainer container)
        {
            // the only accessors to resolve are labels
            return getInstructionAt(((CodeLabel)accessor).getAddress());
        }
    }

    /**
     * ...
     */
    private Instruction[] instructions;

    /**
     * ...
     */
    private List handlers;

    /**
     * ...
     */
    private List lines;

    /* LOCAL VARIABLES NOT USED
       private Map locals;
    LOCAL VARIABLES NOT USED */

    /**
     * ...
     */
    private int pc;

    /**
     * ...
     */
    private boolean discardValue;

    /**
     * Is a label after the last instruction?
     * ...
     */
    private boolean labelAtEnd;

    /**
     * ...
     */
    private int lineNumber;

    /**
     * ...
     */
    private int lastLine;

    /**
     * ...
     */
    private LinkedList contexts = new LinkedList();

    /**
     *
     */
    private static final LinkedList stack = new LinkedList();

    /**
     * Constructs a code attribute.
     */
    private CodeSequence()
    {
        instructions =
            new Instruction[at.dms.classfile.Constants.MAX_CODE_PER_METHOD];
        handlers = new ArrayList();
        lines = new ArrayList();
        /* LOCAL VARIABLES NOT USED
           locals = new HashMap();
           LOCAL VARIABLES NOT USED */
    }

    /**
     * Constructs a code sequence.
     */
    public static CodeSequence getCodeSequence()
    {
        CodeSequence seq;
        if (!Constants.ENV_USE_CACHE || stack.isEmpty()) {
            seq = new CodeSequence();
        }
        else {
            seq = (CodeSequence)stack.removeFirst();
        }
        seq.pc = 0;
        seq.discardValue = false;
        seq.labelAtEnd = false;
        seq.lineNumber = 0;
        seq.lastLine = -1;
        return seq;
    }

    /**
     * Release a code sequence.
     */
    public void release()
    {
        if (Constants.ENV_USE_CACHE) {
            stack.addFirst(this);
            handlers.clear();
            lines.clear();
            /* LOCAL VARIABLES NOT USED
               seq.locals.clear();
               LOCAL VARIABLES NOT USED */      
        }
    }

    /**
     * ...
     */
    public static void endSession()
    {
        while (!stack.isEmpty()) {
            stack.removeFirst();
        }
    }

    /**
     * Adds an instruction to the code of the current method.
     *
     * @param instruction  the instruction to append
     */
    public final void plantInstruction(Instruction instruction)
    {
        instructions[pc++] = instruction;

        discardValue = false;
        labelAtEnd = false;
        if (lineNumber != lastLine) {
            lastLine = lineNumber;
            lines.add(new LineNumberInfo((short)lineNumber, instruction));
        }
    }

    /**
     * Adds a local var instruction to the code of the current method.
     *
     * @param instruction  the instruction to append
     */
    public final void plantLocalVar(int opcode, JLocalVariable var)
    {
        LocalVarInstruction instruction
            = new LocalVarInstruction(opcode, var.getPosition());
        plantInstruction(instruction);
        addLocalVarInfo(instruction, var);
    }

    /**
     * Adds a load of this (local var 0)
     *
     * @param instruction  the instruction to append
     */
    public final void plantLoadThis()
    {
        plantInstruction(new LocalVarInstruction(opc_aload, 0));
    }

    /**
     * Adds an instruction to the code of the current method.
     *
     * @param instruction  the instruction to append
     */
    public final void plantLabelRef(int opcode, CodeLabel label)
    {
        plantLabelRef("unidentified construct", opcode, label);
    }

    public final void plantLabelRef(String debugIdent,
                                    int opcode, CodeLabel label)
    {
        plantInstruction(new JumpInstruction(opcode, label, debugIdent));
    }

    /**
     * Adds an instruction to the code of the current method.
     *
     * @param instruction  the instruction to append
     */
    public final void plantLabel(CodeLabel label)
    {
        label.setAddress(pc);
        labelAtEnd = true;
    }

    /**
     * @param lineNumber  the current line number in source code.
     */
    public final void setLineNumber(int lineNumber)
    {
        if (lineNumber != 0) {
            this.lineNumber = lineNumber;
        }
    }

    /**
     * @return an array of line number information.
     */
    public final LineNumberInfo[] getLineNumbers()
    {
        LineNumberInfo[] info = new LineNumberInfo[lines.size()];
        return (LineNumberInfo[])lines.toArray(info);
    }

    /**
     * @return an array of local vars information..
     */
    public final LocalVariableInfo[] getLocalVariableInfos()
    {
        return null;
        /* LOCAL VARIABLES NOT USED
           if (locals.size() == 0) {
           return null;
           }

           List compress = new ArrayList();

           Enumeration stacks = locals.elements();
           while (stacks.hasMoreElements()) {
           LinkedList current = (LinkedList)stacks.nextElement();
           while (!current.empty()) {
           compress.add(((JLocalVariableEntry)current.pop()).getInfo());
           }
           }
           LocalVariableInfo[] info = new LocalVariableInfo[compress.size()];
           return (LocalVariableInfo[])compress.toArray(info);

           LOCAL VARIABLES NOT USED */
    }

    /**
     * @param discardValue
     * Set if the value of subexpression should be discarded.
     */
    public final void setDiscardValue(boolean discardValue)
    {
        this.discardValue = discardValue;
    }

    /**
     * @return true if super expression use value of sub expression.
     */
    public final boolean discardValue()
    {
        return discardValue;
    }

    /**
     * Ask the code handler to generate the necessary code to call every
     * finally clause of all try statements. Doesn't not really insert
     * a return!
     */
    public final void plantReturn(JReturnStatement ret)
    {
        for (int i = contexts.size() - 1; i >= 0; i--) {
            JStatement statement = (JStatement)contexts.get(i);

            if (statement instanceof JTryFinallyStatement) {
                ((JTryFinallyStatement)statement).genFinallyCall(this, ret);
            }
            else if (statement instanceof JSynchronizedStatement) {
                ((JSynchronizedStatement)statement).genMonitorExit(this);
            }
        }
    }

    /**
     * Ask the code handler to generate the necessary code to call every
     * finally and monitorexit.
     */
    public final void plantBreak(JStatement top)
    {
        for (int i = contexts.size() - 1;
             i >= 0 && contexts.get(i) != top; i--) {

            JStatement statement = (JStatement)contexts.get(i);

            if (statement instanceof JTryFinallyStatement) {
                ((JTryFinallyStatement)statement).genFinallyCall(this, null);
            }
            else if (statement instanceof JSynchronizedStatement) {
                ((JSynchronizedStatement)statement).genMonitorExit(this);
            }
        }
    }

    /**
     * Informs the code handlers that we begin a portion of breakable code.
     */
    public final void pushContext(JStatement statement)
    {
        contexts.addFirst(statement);
    }

    /**
     * Informs the code handlers that we exit a breakable code.
     * Checks that contexts match.
     */
    public final void popContext(JStatement statement)
    {
        assert((JStatement)contexts.removeFirst() == statement);
    }

    /**
     * Add an exception handler to this code.
     */
    public final void addExceptionHandler(HandlerInfo info)
    {
        handlers.add(info);
    }

    /**
     * Returns an array of all exception handler.
     */
    public final HandlerInfo[] getHandlers()
    {
        HandlerInfo[] info = new HandlerInfo[handlers.size()];
        return (HandlerInfo[])handlers.toArray(info);
    }

    /**
     * Gets the location in code sequence.
     */
    public final int getPC()
    {
        return pc;
    }

    /**
     * Returns the actual size of code. (number of instruction)
     */
    public final int size()
    {
        return pc;
    }

    /**
     * Returns the instruction at a given position.
     */
    public final Instruction getInstructionAt(int pc)
    {
        return instructions[pc];
    }

    /**
     * Return the instruction as a list.
     *
     * WARNING: AFTER a call to release() this array will be reused
     *
     * @param instruction the instruction to append.
     */
    public Instruction[] getInstructionArray()
    {
        // we resolve the labels here, since the array will be reused
        resolveLabels();

        Instruction[] code = new Instruction[pc];
        System.arraycopy(instructions, 0, code, 0, pc);
        return code;
    }

    private void resolveLabels()
    {
        // if there is a label planted as last instruction, add a dummy
        // instruction at the end: it will never be reached
        if (labelAtEnd) {
            plantInstruction(new NoArgInstruction(Constants.opc_nop));
        }

        try {
            AccessorTransformer transformer = new LabelResolvingTransformer();
            for (int i = 0; i < pc; i++) {
                if (instructions[i] instanceof AccessorContainer) {
                    ((AccessorContainer)instructions[i])
                        .transformAccessors(transformer);
                }
            }
        } 
        catch (BadAccessorException e) {
            throw new RuntimeException("INTERNAL ERROR");
        }
    }

    /**
     * Add a local variable name information.
     */
    private final void addLocalVarInfo(LocalVarInstruction instruction,
                                       JLocalVariable var)
    {
        /* //  DISABLED, SLOW AND NOT USED AFTER
           Integer  pos = new Integer(var.getLocalIndex());
           LinkedList  elemsAtPos;
           JLocalVariableEntry entry = new JLocalVariableEntry(instruction, var);

           elemsAtPos  = (LinkedList)locals.get(pos);
           if (elemsAtPos == null) {
           locals.put(pos, elemsAtPos = new LinkedList());
           elemsAtPos.push(entry);
           } else if (!((JLocalVariableEntry)elemsAtPos.peek()).merge(entry)) {
           elemsAtPos.push(entry);
           }
        */
    }
}
