/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.UnpositionedError;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.util.MessageDescription;

/**
 * This class is an abstract root class for binary arithmetic expressions.
 * (In the JLS second edition also refered to as "Multiplicative Operators",
 * section 15.27)
 */
public abstract strictfp class JBinaryArithmeticExpression
    extends JBinaryExpression
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the location of this node in the source code.
     * @param left left operand
     * @param righ right operand
     */
    protected JBinaryArithmeticExpression(TokenReference where,
                                          JExpression left,
                                          JExpression right)
    {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context the context to evaluate this expression in.
     * @return a pure java expression including promote node
     * @exception PositionedError
     * If the check fails
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        JExpression expr = checkBinaryBody(context);
        
        /* ? */
        if (expr == this) {
            left = left.convertType(type, context);
            right = right.convertType(type, context);
        }

        return expr;
    }

    /**
     * Check expression and evaluate and alter context
     *
     * @param context the actual context of analyse
     *
     * @return  a pure java expression including promote node
     *
     * @exception PositionedError
     * Any error during this method, with information about position in the
     * source file.
     */
    public abstract JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError;

    /**
     * Default rule for the type of the result of the binary arithmetic
     * expression.
     *
     * @param leftType the type of left operand
     *
     * @param rightType the type of right operand
     *
     * @param badTypeMessage
     * The message to be shown to the user when the types of the operands are
     * bad.
     *
     * @return The possibly promoted type of the binary arithmetic expression.
     *
     * @exception UnpositionedError
     * This error will be positioned soon
     */
    public static CType computeType(CType leftType, CType rightType,
                                    MessageDescription badTypeMessage)
        throws UnpositionedError
    {
        if (leftType.isNumeric() && rightType.isNumeric()) {
            return CNumericType.binaryPromote(leftType, rightType);
        }
        else {
            /* Only numeric types can be used in binary arithmetic expressions.
             */
            throw new UnpositionedError(badTypeMessage, leftType, rightType);
        }
    }

    /**
     * Peforms the operation of the subclass implementation on constant
     * operands. This method must only be called when both operands are
     * constant.
     *
     * These are the default rules for constant folding in binary arithmetic
     * expressions.
     * This involves Binary Numeric Promotion (JLS2 5.6.2) on constants.
     *
     * @return a literal resulting of an operation over two literals.
     *
     * @exception UnpositionedError
     * This error will be positioned soon
     */
    public JExpression constantFolding()
        throws UnpositionedError
    {
        Number leftNumber = left.getNumberLiteral().getNumberValue();
        Number rightNumber = right.getNumberLiteral().getNumberValue();

        CType leftType = left.getType();
        CType rightType = right.getType();

        if (leftType == CStdType.Double || rightType == CStdType.Double) {

            double  result;
      
            result = compute(leftNumber.doubleValue(),
                             rightNumber.doubleValue());
            return new JRealLiteral(getTokenReference(),
                                    new Double(result), CStdType.Double);
        }
        else if (leftType == CStdType.Float || rightType == CStdType.Float) {

            float  result;
      
            result = compute(leftNumber.floatValue(),
                             rightNumber.floatValue());
            return new JRealLiteral(getTokenReference(),
                                    new Float(result), CStdType.Float);
        }
        else if (leftType == CStdType.Long || rightType == CStdType.Long) {

            long  result;
      
            result = compute(leftNumber.longValue(), rightNumber.longValue());
            return new JOrdinalLiteral(getTokenReference(),
                                       new Long(result), CStdType.Long);
        }
        else if (leftType.implicitlyCastTo(CStdType.Integer)
                 && rightType.implicitlyCastTo(CStdType.Integer)) {

            int  result;

            result = compute(leftNumber.intValue(), rightNumber.intValue());
            return new JOrdinalLiteral(getTokenReference(),
                                       new Integer(result), CStdType.Integer);
        }
        else {
            fail();
            return null;
        }
    }

    /**
     * Computes the result of the operation on two integer constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public abstract int compute(int left, int right)
        throws UnpositionedError;

    /**
     * Computes the result of the operation on two long constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public abstract long compute(long left, long right)
        throws UnpositionedError;

    /**
     * Computes the result of the operation on two float constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public abstract float compute(float left, float right)
        throws UnpositionedError;

    /**
     * Computes the result of the operation on two double constant values.
     *
     * @param left the left operand
     * @param right the right operand
     * @return the result of the operation
     */
    public abstract double compute(double left, double right)
        throws UnpositionedError;
}
