/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class is an abstract root class for binary expressions
 * Here are conversion method following JLS 5.6.2
 */
public abstract class JBinaryExpression
    extends JExpression
{

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where  the line of this node in the source code
     * @param p1  left operand
     * @param p2  right operand
     */
    public JBinaryExpression(TokenReference where,
                             JExpression left,
                             JExpression right)
    {
        super(where);
        this.left = left;
        this.right = right;
    }

    /**
     * Returns the type of the whole expression.
     * @return the type of this expression
     */
    public CType getType()
    {
        return type;
    }

    /**
     * Sets the type of the whole expression. Must only be done by
     * code that is responsible for determining the type.
     *
     * @return the type of this expression
     */
    protected void setType(CType type)
    {
        this.type = type;
    }

    /**
     * Returns the left side of this binary expression.
     */
    public JExpression getLeft()
    {
        return left;
    }

    /**
     * Sets the left side of the expression. Must only be done when this class
     * expects it to happen. This method is protected because only this
     * class and subclasses can reasonably know when this is.
     */
    protected void setLeft(JExpression left)
    {
        this.left = left;
    }

    /**
     * Returns the right side of this binary expression.
     */
    public JExpression getRight()
    {
        return right;
    }

    /**
     * Sets the right side of the expression. Must only be done when this class
     * expects it to happen. This method is protected because only this
     * class and subclasses can reasonably know when this is.
     */
    protected void setRight(JExpression right)
    {
        this.right = right;
    }

    /**
     * Apply binary numeric promotion (JLS2 5.6.2)
     *
     * @param context the context of the evaluation of this expression.
     *
     * @return the type after promotion
     *
     * @exception PositionedError an error occurred
     */
    public CType promote(CExpressionContext context)
        throws PositionedError
    {
        CType promoted;

        promoted = CNumericType.binaryPromote(left.getType(), right.getType());

        left = left.convertType(promoted, context);
        right = right.convertType(promoted, context);

        return promoted;
    }

    /**
     * Generates a sequence of bytecodes
     * @param code the code list
     */
    public void genBooleanResultCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        CodeLabel  okayLabel = new CodeLabel();
        CodeLabel  nextLabel = new CodeLabel();

        //  LEFT CODE IFNE okay
        genBranch(true, code, okayLabel);
        // FALSE
        code.plantInstruction(new PushLiteralInstruction(0));
        //  GOTO next
        code.plantLabelRef("binaryExpr:GOTO next", opc_goto, nextLabel);
        // okay:
        code.plantLabel(okayLabel);
        // TRUE
        code.plantInstruction(new PushLiteralInstruction(1));
        // next ...
        code.plantLabel(nextLabel);
    }

    /**
     * Generates a sequence of bytecodes to branch on a label
     * This method helps to handle heavy optimizables conditions
     * @param code  the code list
     */
    public void genBranch(boolean cond,
                          CodeSequence code,
                          CodeLabel label)
    {
        // here, we try to optimize
        left = left.unParenthesize();
        right = right.unParenthesize();

        genBranch(left, right, cond, code, label);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
                             JExpression right,
                             boolean cond,
                             CodeSequence code,
                             CodeLabel label)
    {
        genCode(code);
        code.plantLabelRef("binaryExpr:genBranch",
                           cond ? opc_ifne : opc_ifeq, label);
    }

    protected CType type;
    protected JExpression left;
    protected JExpression right;
}
