/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Block. (JLS 14.2)
 *
 * A block is a sequence of statements and local variable declaration
 * statements within braces.
 */
public class JBlock
    extends JStatement
{
    /**
     * Statements in the body of the block.
     */
    protected JStatement[] body;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JBlock(TokenReference where,
                  JStatement[] body,
                  JavaStyleComment[] comments)
    {
        super(where, comments);

        this.body = body;
    }

    /**
     * Returns statements in the body of the block.
     */
    public JStatement[] getBody()
    {
        return body;
    }

    /**
     * Installs a new body in this block.
     */
    public void setBody(JStatement[] body)
    {
        this.body = body;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        CBlockContext self = context.createBlockContext();
        CBodyContext inside = self.createBodyContext();

        for (int i = 0; i < body.length; i++) {
            if ( !inside.isReachable() ) {
                throw new CLineError(body[i].getTokenReference(),
                                     KjcMessages.STATEMENT_UNREACHABLE);
            }
            
            try {
                body[i].checkBody(inside);
            }
            catch (CLineError le) {
                context.reportTrouble(le);
            }
        }

        inside.close(getTokenReference());
        self.close(getTokenReference());
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitBlockStatement(this, body, getComments());
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        for (int i = 0; i < body.length; i++) {
            body[i].genCode(code);
        }
    }
}
