/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Break Statement. (JLS 14.13)
 *
 * A break statement transfers control out of an enclosing statement.
 */
public class JBreakStatement
    extends JStatement
{
    /**
     * Where to break to.
     */
    private CodeLabel breakLabel;

    /**
     * Name of label to break to, or <code>null</code> for nearest breakable
     * statement.
     */
    private String label;

    /**
     * Target statement.
     */
    private JStatement target;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code
     * @param label  the label of the enclosing labeled statement
     */
    public JBreakStatement(TokenReference where,
                           String label,
                           JavaStyleComment[] comments)
    {
        super(where, comments);
        this.label = label;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails.
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        if (label != null) {
            target = context.getLabeledStatement(label);
            check(context,
                  target != null,
                  KjcMessages.LABEL_UNKNOWN,
                  label);
        }
        else {
            target = context.getNearestBreakableStatement();
            check(context,
                  target != null,
                  KjcMessages.CANNOT_BREAK);
        }
        context.addBreak(target, context);
        context.addBreak();

        breakLabel = target.getBreakLabel();

    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);
        p.visitBreakStatement(this, label);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        code.plantBreak(target);

        code.plantLabelRef(opc_goto, breakLabel);

        breakLabel = null;
        target = null;
    }
}
