/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.ClassRefInstruction;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a cast expression '((byte)2)'
 */
public class JCastExpression
    extends JExpression
{
    /**
     * Type of "java/lang/ClassCastException"
     */
    private CClassType classCastException;

    /**
     * Constructs a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param expr
     * The expression to be casted.
     * @param dest
     * The type of this expression after cast.
     */
    public JCastExpression(TokenReference where,
                           JExpression expr,
                           CType dest)
    {
        super(where);
        this.expr = expr;
        this.dest = dest;
    }

    /**
     * Compute the type of this expression.
     *
     * @return
     * The type of this expression.
     */
    public CType getType()
    {
        return dest;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        if ( classCastException == null ) {
            classCastException =
                CClassType.lookup("java/lang/ClassCastException");
            
            try {
                classCastException.checkType(null);
            }
            catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }
        }
        
        expr = expr.checkExpression(context);

        try {
            dest.checkType(context);
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

        check(context,
              expr.getType().isCastableTo(dest),
              KjcMessages.CAST_CANT,
              expr.getType(), dest);

        if (expr.getType().isClassType()) {
            /* Casting something of class type can result in a
             * ClassCastException, so a catch clause for ClassCastException
             * would be reachable.
             */
            CBodyContext bodyContext = context.getBodyContext();
            CThrowableInfo info =
                new CThrowableInfo(classCastException, this,
                                   bodyContext.makeSnapshot(),
                                   false); // subTypePossible
            bodyContext.addThrowable(info);
        }

        if (!expr.getType().isPrimitive()
            && expr.getType().implicitlyCastTo(dest)
            && expr.getType() != CStdType.Null) {

            context.reportTrouble
                (new CWarning(getTokenReference(),
                              KjcMessages.UNNECESSARY_CAST,
                              expr.getType(), dest));
        }

        if (expr.isConstant()) {
            if (expr.getType().isPrimitive()) {
                return expr.convertType(dest, context);
            }
            else if (expr.getType().equals(CStdType.String)) {
                if (dest.equals(CStdType.String)) {
                    return expr;
                }
                else {
                    fail();
                    // The earlier check to see if casting is possible must
                    // have caught this case.
                }
            }
        }

        
        // WAS: expr.getType() / dest
        if (!dest.isAssignableTo(expr.getType())) { 
            return expr.convertType(dest, context);
        }

        return this;
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p) {
        p.visitCastExpression(this, expr, dest);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        expr.genCode(code);

        if (dest.isNumeric()) {
            ((CNumericType)expr.getType()).genCastTo((CNumericType)dest, code);
        }
        else if (dest instanceof CClassType) {
            code.plantInstruction(new ClassRefInstruction
                (opc_checkcast, ((CClassType)dest).getQualifiedName()));
        }
    }

    protected JExpression expr;
    protected CType dest;
}
