/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.HandlerInfo;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents a catch clause declaration in the syntax tree.
 */
public class JCatchClause
    extends JPhylum
{
    /**
     * The exception parameter of this catch clause.
     */
    private JFormalParameter exception;

    /**
     * The body of code of this catch clause.
     */
    private JBlock body;

    /**
     * Whether this catch clause has already been checked by a call to
     * checkBody.
     */
    private boolean bodyChecked;

    /**
     * The body of code of this catch clause.
     */
    private int catchPC;

    /**
     * Whether this clause is used. (there is an exception that can be
     * thrown that can be caught by this clause)
     */
    private boolean used;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     *
     * @param ident
     * The name of this variable.
     *
     * @param initializer
     * The initializer.
     */
    public JCatchClause(TokenReference where,
                        JFormalParameter exception,
                        JBlock body)
    {
        super(where);

        this.exception = exception;
        this.body = body;
    }

    /**
     * Returns the type of the exception caught by this catch clause.
     *
     * @return
     * The type of exception caught by this catch clause.
     */
    public CClassType getType()
    {
        return (CClassType)exception.getType();
    }

    /**
     * Returns the body of the catch clause.
     */
    public JBlock getBody()
    {
        return body;
    }

    /**
     * Returns whether this clause is used. (there is an exception that can be
     * thrown that can be caught by this clause)
     */
    public boolean isUsed()
    {
        return used;
    }

    /**
     * Sets whether this clause is used. (there is an exception that can be
     * thrown that can be caught by this clause)
     */
    public void setUsed(boolean flag)
    {
        this.used = flag;
    }

    /**
     * Returns the parameter declaration for the exception caught by this
     * catch clause.
     */
    public JFormalParameter getException()
    {
        return exception;
    }

    /**
     * Whether the checkBody method has already been called.
     */
    public boolean isBodyChecked()
    {
        return bodyChecked;
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public void checkInterface(CBodyContext context)
        throws PositionedError
    {
        /* Will also be checked in the checkBody method. */
        exception.checkInterface(context);
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public void checkBody(CBodyContext context)
        throws PositionedError
    {
        bodyChecked = true;

        /* Will also be checked in the checkInterface method. */
        exception.checkInterface(context.getClassContext());

        CBlockContext block = context.createBlockContext(1);
        CBodyContext inside = block.createBodyContext();

        inside.setReachable(true);
        exception.checkBody(inside);
        body.checkStatement(inside);

        inside.close(getTokenReference());
        block.close(getTokenReference());

        if (body.getBody().length == 0) {
            context.reportTrouble(new CWarning(getTokenReference(),
                                               KjcMessages.EMPTY_CATCH_BLOCK));
        }
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitCatchClause(this, exception, body);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        catchPC = code.getPC();
        exception.genStore(code);
        body.genCode(code);
    }

    /**
     * Adds exception handler code for this catch clause to the code sequence
     * being generated for this method.
     *
     * @param code
     * The code sequence.
     *
     * @param startPC
     * The index of the first bytecode instruction protected by this catch
     * clause.
     *
     * @param endPC
     * The index immediately after the last bytecode instruction protected by
     * this catch clause.
     */
    public void genExceptionHandler(CodeSequence code, int startPC, int endPC)
    {
        code.addExceptionHandler
            (new HandlerInfo(code.getInstructionAt(startPC),
                             code.getInstructionAt(endPC),
                             code.getInstructionAt(catchPC),
                             exception.getType()
                             .getCClass().getQualifiedName()));
    }
}
