/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a java class in the syntax tree.
 */
public class JClassDeclaration
    extends JTypeDeclaration
{
    /**
     * Assignment information for static fields.
     */
    private CVariableInfo staticInfo;

    /**
     * Context used for checking members of this class.
     */
    private CClassContext self;

    /**
     * Name of superclass.
     */
    protected String superName;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code
     * @param context  the context in which this class is defined
     * @param modifiers the list of modifiers of this class
     * @param ident  the short name of this class
     * @param superName the name of super class of this class
     * @param interfaces the name of this class's interfaces
     * @param javadoc is this class deprecated
     */
    public JClassDeclaration(TokenReference where,
                             int modifiers,
                             String ident,
                             String superName,
                             CClassType[] interfaces,
                             JFieldDeclaration[] fields,
                             JMethodDeclaration[] methods,
                             JTypeDeclaration[] inners,
                             JPhylum[] initializers,
                             JavadocComment javadoc,
                             JavaStyleComment[] comment)
    {
        super(where, modifiers, ident, interfaces, fields, methods,
              inners, initializers, javadoc, comment);
        this.superName = superName;
    }

    /**
     * Sets the super class.
     */
    public void setSuperClass(String superName)
    {
        this.superName = superName;
    }

    /**
     * Transforms this type before code generation.
     *
     * @exception PositionedError
     * An error with reference to the source file./
     */
    public void transformType(CContext context)
        throws PositionedError
    {
    }

    /**
     * Second pass (quick), check interface looks good.
     * This pass is just a tuning pass in order to create informations about
     * exported elements such as Classes, Interfaces, Methods, Constructors
     * and Fields.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInterface(final CContext context)
        throws PositionedError
    {
        boolean inInterface = false;
        CClassContext classContext = context.getClassContext();
        if (classContext != null) {
            inInterface = classContext.getCClass().isInterface();
        }

        if (inInterface) {
            setModifiers(getModifiers() | ACC_PUBLIC | ACC_STATIC);
            getCClass().setModifiers(getCClass().getModifiers() |
                                     ACC_PUBLIC | ACC_STATIC);
        }

        int modifiers = getModifiers();

        if (getOwner() != null) {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_PRIVATE + ACC_PROTECTED
                                 + ACC_ABSTRACT + ACC_FINAL + ACC_STATIC
                                 + ACC_STRICT),
                  KjcMessages.INVALID_CLASS_MODIFIERS);
        }
        else {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_ABSTRACT + ACC_FINAL
                                 + ACC_STRICT),
                  KjcMessages.INVALID_CLASS_MODIFIERS);
        }

        if (inInterface) {
            check(context,
                  !hasOtherFlags(modifiers,
                                 ACC_PUBLIC + ACC_ABSTRACT + ACC_STATIC),
                  KjcMessages.CLASS_FLAGS_IN_INTERFACE);
        }

        check(context,
              !hasMoreThanOneFlag(modifiers, ACC_FINAL + ACC_ABSTRACT),
              KjcMessages.MODIFIERS_ABSTRACT_AND_FINAL);

        check(context,
              !hasMoreThanOneFlag(modifiers,
                                  ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE),
              KjcMessages.MULTIPLE_ACCESS_MODIFIERS);

        CClass superClass = null;
        CClassType superType = null;

        if (superName == null) {
            if (getSourceClass().getType().equals(CStdType.Object)) {
                // java/lang/Object
                // superClass = null;
                // superType = null;
            }
            else {
                superClass = CStdType.Object.getCClass();
                superType = CStdType.Object;
            }
        }
        else {
            // find name
            try {
                superType = CClassType.lookup(superName);
                superType.checkType(context);
                superClass = superType.getCClass();
            }
            catch (UnpositionedError cue) {
                /* This must be worked out further. To allow checks for
                 * circular hierarchies, this must go, and it doesn't seem
                 * to ever be used in a valid program.
                */
                // System.err.println("Strange delayed super class lookup: "
                //                     + superName);
                // not yet defined, throw cue.addPosition(getTokenReference());
                superType = new CClassNameType(superName) {
                        public void checkType(CContext ctxt)
                            throws UnpositionedError
                        {
                            CClassType temp = CClassType.lookup(superName);

                            temp.checkType(context);
                            setClass(temp.getCClass());
                            setQualifiedName(null);
                        }
                    };
            }
        }

        // check access
        if (superClass != null) {
            if (!superClass.isInnerClass()) {
                /* Superclass is toplevel class */
                check(context,
                      superClass.isPublic()
                      || ( getSourceClass().getPackage()
                           == superClass.getPackage() ),
                      KjcMessages.CLASS_ACCESSPARENT,
                      superName);
            }
            else {
                if (superClass.isProtected() && getCClass().isInnerClass()
                    && getOwner().descendsFrom(superClass.getOwner())) {
                    // Okay!
                }
                else {
                    check(context,
                          superClass.isPublic()
                          || ( getSourceClass().getPackage()
                               == superClass.getPackage() ),
                          KjcMessages.CLASS_ACCESSPARENT,
                          superName);
                }
            }

            check(context,
                  !superClass.isFinal(),
                  KjcMessages.CLASS_PARENT_FINAL,
                  superName);

            check(context,
                  !superClass.isInterface(),
                  KjcMessages.CLASS_EXTENDS_INTERFACE,
                  superName);
        }

        CClassContext self = createContext(context);

        JInitializerDeclaration statInit = constructInitializers(true);
        if (statInit != null) {
            statInit.checkInterface(self);
            setStaticInitializer(statInit);
        }

        Iterator iterator = getMethodList().iterator();
        boolean hasConstructor = false;
        while (iterator.hasNext()) {
            JMethodDeclaration method = (JMethodDeclaration)iterator.next();
            if (method instanceof JConstructorDeclaration) {
                // usually constructors are at the top of the file
                // so its not as inefficient as it's look
                hasConstructor = true;
                break;
            }
        }

        if (!hasConstructor && getDefaultConstructor() == null) {
            setDefaultConstructor(constructDefaultConstructor());
        }

        JInitializerDeclaration instanceInit = constructInitializers(false);
        if (instanceInit != null) {
            instanceInit.checkInterface(self);
            setInstanceInitializer(instanceInit);
        }

        getCClass().setSuperClass(superType);
        super.checkInterface(context, superType);
    }

    /**
     * Check that initializers are correct.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkInitializers(CContext context)
        throws PositionedError
    {
        self = createContext(context);

        JInitializerDeclaration statInit = getStaticInitializer();

        if (statInit != null) {
            statInit.checkInitializer(self);
            staticInfo = self.getVariableInfo();
        }

        if (getCClass().getSuperClass() != null) {
            check(context,
                  !(getCClass().descendsFrom(getCClass().getSuperClass()) &&
                    getCClass().getSuperClass().descendsFrom(getCClass())),
                  KjcMessages.CLASS_CIRCULARITY,
                  getIdent());
        }

        // Check inners
        Iterator iterator = getInnerList().iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            inner.checkInitializers(self);
        }

        super.checkInitializers(context);
    }

    /**
     * Checks interface again after transformations have been applied.
     */
    public void checkTransformedInterface(CContext context)
        throws PositionedError
    {
        CClassContext self = createContext(context);

        Iterator iterator;

        /* Now we add the declaration of the generated methods
         * (generated constructors not allowed), so generated inner classes
         * can find those methods when the inner classes have their body
         * checked.
         */
        iterator = getGeneratedMethodList().iterator();
        while (iterator.hasNext()) {
            JMethodDeclaration method = (JMethodDeclaration)iterator.next();
            try {
                if (!(method instanceof JConstructorDeclaration)) {
                    /*
                    System.err.println("Adding method to source class: "
                                       + method.getIdent());
                    */
                    getSourceClass().addMethod(method.checkInterface(self));
                }
                else {
                    throw new RuntimeException("Constructor not allowed as "
                                               + "seperate-kept generated "
                                               + "method.");
                }
            }
            catch (CBlockError e) {
                context.reportTrouble(e);
            }
        }

        /* Check interfaces of generated inner classes.
         */
        iterator = getGeneratedInnerList().iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            try {
                inner.checkInterface(self);
            }
            catch (CBlockError e) {
                context.reportTrouble(e);
            }
        }
    }

    /**
     * Check only the parts of the class that have changed because
     * of the transformations. Only generated inner classes and methods
     * (no constructors) are checked, not fields. (Because AgentJava
     * doesn't generate fields in transformed classes, only in generated
     * inner classes.)
     */
    public void checkTransformedTypeBody(CContext context)
        throws PositionedError
    {
        CClassContext self = createContext(context);

        Iterator iterator;

        /* Check bodies of generated inner classes.
         */
        iterator = getGeneratedInnerList().iterator();
        while (iterator.hasNext()) {
            JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
            try {
                inner.checkInitializers(self);
                inner.checkTypeBody(self);
                // Add it to the list of inners in the class, and don't
                // keept it seperate: this ensures that the pretty printer
                // doesn't forget about this inner.
                addGeneratedInner(inner, false);
                // Remove to avoid clogging up memory.
                iterator.remove();
            }
            catch (CBlockError e) {
                context.reportTrouble(e);
            }
        }

        /* Now check bodies of generated methods.
         */
        iterator = getGeneratedMethodList().iterator();
        while (iterator.hasNext()) {
            JMethodDeclaration method = (JMethodDeclaration)iterator.next();
            try {
                method.checkStatement(self);
                // Add it to the list of methods in the class, and don't
                // keept it seperate: this ensures that the pretty printer
                // doesn't forget about this method.
                addGeneratedMethod(method, false);
                // Remove to avoid clogging up memory.
                iterator.remove();
            }
            catch (CBlockError e) {
                context.reportTrouble(e);
            }
        }
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     *
     * @exception PositionedError
     * An error with reference to the source file.
     */
    public void checkTypeBody(CContext context)
        throws PositionedError
    {
        try {
            CVariableInfo instanceInfo = null;

            /* Array of CVariableInfo objects resulting from constructors. */
            CVariableInfo[] constructorsInfo;

            /* List of CVariableInfo objects that is added to and is converted
             * to constructorsInfo when ready. */
            List constructorsInfoList = new ArrayList(5);

            /* Check the instance initializer. */
            JInitializerDeclaration instanceInit = getInstanceInitializer();
            if (instanceInit != null) {
                instanceInit.checkInitializer(self);
                instanceInfo = self.getVariableInfo();
            }

            /* Mark all fields as 'fully declared'. */
            Iterator iterator = getFieldList().iterator();
            while (iterator.hasNext()) {
                JFieldDeclaration declaration =
                    (JFieldDeclaration)iterator.next();
                ((CSourceField)declaration.getField()).setFullyDeclared(true);
            }

            /* Check inner classes. */
            iterator = getInnerList().iterator();
            while (iterator.hasNext()) {
                JTypeDeclaration inner = (JTypeDeclaration)iterator.next();
                try {
                    inner.checkTypeBody(self);
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            /* Compile the constructors. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                try {
                    /* Check if method is a constructor. */
                    if (method instanceof JConstructorDeclaration) {
                        /* Important: set assignment info from instance
                         * initializer before checking; getVariableInfo
                         * method clears old info!!
                         */
                        if (instanceInfo != null) {
                            self.setVariableInfo((CVariableInfo)instanceInfo.clone());
                        }
                        method.checkStatement(self);
                        constructorsInfoList.add(self.getVariableInfo());
                    }
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            if (getDefaultConstructor() == null) {
                constructorsInfo = (CVariableInfo[])
                    constructorsInfoList.toArray
                    (new CVariableInfo[constructorsInfoList.size()]);
            }
            else {
                getDefaultConstructor().checkStatement(self);
                constructorsInfo =
                    new CVariableInfo[] { self.getVariableInfo() };
            }

            /* For the purpose of compiling methods, all final fields
             * are marked as initialized. They are checked later to see
             * if they really were correctly initialized in the constructors.
             */
            self.markAllFieldsAsInitialized();

            /* Compile the methods. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                try {
                    if ( !(method instanceof JConstructorDeclaration) ) {
                        method.checkStatement(self);
                    }
                }
                catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            /* Check if methods are used. */
            iterator = getMethodList().iterator();
            while (iterator.hasNext()) {
                JMethodDeclaration method =
                    (JMethodDeclaration)iterator.next();
                CSourceMethod checkingMethod =
                    (CSourceMethod)method.getMethod();

                if (!checkingMethod.isUsed()
                    && !checkingMethod.getIdent().equals(JAV_CONSTRUCTOR)) {

                    context.reportTrouble
                        (new CWarning(method.getTokenReference(), 
                                      KjcMessages.UNUSED_PRIVATE_METHOD, 
                                      checkingMethod.getIdent()));
                }
            }

            /* Check if fields are used. */
            iterator = getFieldList().iterator();
            while (iterator.hasNext()) {
                JFieldDeclaration declaration =
                    (JFieldDeclaration)iterator.next();

                if (!((CSourceField)declaration.getField()).isUsed()) {

                    context.reportTrouble
                        (new CWarning(declaration.getTokenReference(), 
                                      KjcMessages.UNUSED_PRIVATE_FIELD, 
                                      declaration.getField().getIdent()));
                }
            }

            self.close(this, staticInfo, instanceInfo, constructorsInfo);

            super.checkTypeBody(context);
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
    
        self = null;
    }

    /**
     * Creates a CClassContext that takes this class as its class context.
     *
     * @param parent
     * The parent context or null.
     * @return
     * A CClassContext that represents this context.
     */
    public CClassContext createContext(CContext parent)
    {
        return new CClassContext(parent, getSourceClass(), this);
    }

    /**
     * Creates a default constructor for classes that don't have declared
     * constructors in their source code.
     */
    private JConstructorDeclaration constructDefaultConstructor()
    {
        JConstructorBlock defaultBlock =
            new JConstructorBlock(getTokenReference(), new JStatement[0]);

        return new JConstructorDeclaration(getTokenReference(),
                                           (getCClass().isPublic()
                                            ? ACC_PUBLIC : 0),
                                           getIdent(),
                                           JFormalParameter.EMPTY,
                                           CClassType.EMPTY,
                                           defaultBlock,
                                           null,
                                           null);
    }

    /**
     * Collects all initializers and builds a single method.
     * @param isStatic class or instance initializers ?
     */
    private JInitializerDeclaration constructInitializers(boolean isStatic)
    {
        List statementList = new ArrayList(5);
        boolean  needGen = false;

        Iterator iterator = getBodyList().iterator();
        while (iterator.hasNext()) {
            JPhylum element = (JPhylum)iterator.next();
            if (element instanceof JClassBlock) {
                JClassBlock classBlock = (JClassBlock)element;
                if (classBlock.isStaticInitializer() == isStatic) {
                    statementList.add(element);
                    needGen = true;
                }
            }
            else {
                if (element instanceof JFieldDeclaration) {
                    JFieldDeclaration declaration = (JFieldDeclaration)element;
                    if (declaration.getVariable().isStatic() == isStatic) {
                        needGen |= declaration.needInitialization();
                        statementList.add
                            (new JClassFieldDeclarator(getTokenReference(),
                                                       declaration));
                    }
                }
            }
        }

        if (statementList.isEmpty()) {
            return null;
        }

        JStatement[] statements = (JStatement[])
            statementList.toArray(new JStatement[statementList.size()]);

        return new JInitializerDeclaration
            (getTokenReference(),
             new JClassBlock(getTokenReference(), false, statements),
             isStatic, !needGen);
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        super.accept(p);

        p.visitClassDeclaration(this,
                                getCClass().getModifiers(),
                                getCClass().getIdent(),
                                superName, 
                                getInterfaces(),
                                getBody(),
                                getMethods(),
                                getInners());
    }

    /**
     * Generate the code in pure java form
     * It is useful to debug and tune compilation process
     * @param p  the printwriter into the code is generated
     */
    public void genInnerJavaCode(KjcPrettyPrinter p)
    {
        super.accept(p);

        p.visitInnerClassDeclaration(this,
                                     getModifiers(),
                                     getIdent(),
                                     superName,
                                     getInterfaces(),
                                     getInners(),
                                     getBody(),
                                     getMethods());
    }
}
