/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Implementation of the ?: operator.
 */
public class JConditionalExpression
    extends JExpression
{
    /**
     * Resulting type of expression.
     */
    private CType type;

    /**
     * Condition part of <code>condition ? left : right</code> expression.
     */
    private JExpression condition;

    /**
     * Left part of <code>condition ? left : right</code> expression.
     */
    private JExpression left;

    /**
     * Right part of <code>condition ? left : right</code> expression.
     */
    private JExpression right;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param condition
     * The condition operand.
     * @param left
     * The left operand.
     * @param right
     * The right operand.
     */
    public JConditionalExpression(TokenReference where,
                                  JExpression condition,
                                  JExpression left,
                                  JExpression right)
    {
        super(where);
        this.condition = condition;
        this.left = left;
        this.right = right;
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType()
    {
        return type;
    }

    /**
     * Check expression and evaluate and alter context
     *
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        condition = condition.checkExpression(context);
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        check(context,
              condition.getType() == CStdType.Boolean,
              KjcMessages.TRINARY_BADCOND);

        /* Now determine the type of the (a ? b : c) expression.
         * It depends on the types of b and c; if they're not equal,
         * then if possible, one of the expressions b and c has its result
         * promoted the the type of the other. Type promotion shouldn't result
         * in information loss, so can only go from types with a smaller range
         * to types with a bigger range.
         * See JLS paras 15.25 and 5.6.2 
         */
        CType leftType = left.getType();
        CType rightType = right.getType();

        if (leftType.equals(rightType)) {
            /* This case also catches the explicitly allowed (JLS) case where
             * leftType and rightType are both CStdType.Boolean.
             */
            type = leftType;
        }
        else if (leftType.isNumeric() && rightType.isNumeric()) {

            if (leftType == CStdType.Byte && rightType == CStdType.Short) {
                type = CStdType.Short; 
            }
            else if ( leftType == CStdType.Short &&
                      rightType == CStdType.Byte ) {
                type = CStdType.Short; 
            }

            /* Type unification in the case where one of the operands
             * is constant.
             */
            else if ( left.isConstant() &&
                      leftType == CStdType.Integer &&
                      ( rightType == CStdType.Byte ||
                        rightType == CStdType.Short ||
                        rightType == CStdType.Char
                        ) &&
                      left.isAssignableTo(rightType)) {

                type = rightType;
            }
            else if ( right.isConstant() &&
                      rightType == CStdType.Integer &&
                      ( leftType == CStdType.Byte ||
                        leftType == CStdType.Short ||
                        leftType == CStdType.Char
                        ) &&
                      right.isAssignableTo(leftType)) {

                type = leftType;
            }
            else {
                type = CNumericType.binaryPromote(leftType, rightType);
            }

            check(context,
                  type != null,
                  KjcMessages.TRINARY_INCOMP,
                  leftType, rightType);

            left = left.convertType(type, context);
            right = right.convertType(type, context);
        }
        else if (leftType.isReference() && rightType.isReference()) {

            if (leftType == CStdType.Null) {
                type = rightType;
            }
            else if (rightType == CStdType.Null) {
                type = leftType;
            }
            else if (leftType.isAssignableTo(rightType)) {
                type = rightType;
            }
            else if (rightType.isAssignableTo(leftType)) {
                type = leftType;
            }
            else {
                check(context,
                      false,
                      KjcMessages.TRINARY_INCOMP,
                      leftType, rightType);
            }

        }
        else {
            check(context,
                  false,
                  KjcMessages.TRINARY_INCOMP,
                  leftType, rightType);
        }

        /* The result of this expression is only constant if all three
         * operands are constant, even if one of the last two constants
         * can never be used.
         */
        if (condition.isConstant()) {
            if (condition.getBooleanLiteral().booleanValue()) {
                if (!right.isConstant()) {
                    left.setConstant(false);
                }
                return left;
            }
            else {
                if (!left.isConstant()) {
                    right.setConstant(false);
                }
                return right;
            }
        }

        return this;
    }

    /**
     * unifyNumeric
     * search the type corresponding to the unification of the two types
     * @param t1 the first type
     * @param t2 the second type
     * @return the corresponding type or null
     */
    private CType unifyNumeric(CType t1, CType t2)
    {
        if (t1 == CStdType.Short && t2 == CStdType.Byte) {
            return CStdType.Short;
        }
        else if (t2 == CStdType.Short && t1 == CStdType.Byte) {
            return CStdType.Short;
        }
        else if (t1 == CStdType.Char
                 && (t2 == CStdType.Char || t2 == CStdType.Byte)) {

            return CStdType.Char;
        }
        else if (t2 == CStdType.Char
                 && (t1 == CStdType.Char || t1 == CStdType.Byte)) {

            return CStdType.Char;
        }
        else {
            return null;
        }
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitConditionalExpression(this, condition, left, right);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        CodeLabel  rightLabel = new CodeLabel();
        CodeLabel  nextLabel = new CodeLabel();

        //  COND IFEQ right
        condition.genBranch(false, code, rightLabel);
        //  LEFT CODE
        left.genCode(code);
        //  GOTO next
        code.plantLabelRef("?::GOTO next", opc_goto, nextLabel);
        // right:
        code.plantLabel(rightLabel);
        //  RIGHT CODE
        right.genCode(code);
        // next ...
        code.plantLabel(nextLabel);
    }
}
