/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class implements the conditional or operation
 */
public class JConditionalOrExpression
    extends JBinaryExpression
{

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where  the line of this node in the source code
     * @param oper  the operator
     * @param left  the left operand
     * @param right  the right operand
     */
    public JConditionalOrExpression(TokenReference where,
                                    JExpression left,
                                    JExpression right) {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     * @exception UnpositionedError this error will be positioned soon
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        left = left.checkExpression(context);
        right = right.checkExpression(context);

        CType leftType = left.getType();
        CType rightType = right.getType();

        check(context,
              leftType == CStdType.Boolean &&  rightType == CStdType.Boolean,
              KjcMessages.OR_BADTYPE);

        type = CStdType.Boolean;

        if (left.isConstant() && right.isConstant()) {
            JBooleanLiteral leftLit = left.getBooleanLiteral();
            JBooleanLiteral rightLit = right.getBooleanLiteral();
            
            left = new JBooleanLiteral(getTokenReference(),
                                       leftLit.booleanValue()
                                       || rightLit.booleanValue());
            return left;
        }
        else if (left.isConstant()) {
            JBooleanLiteral leftLit = left.getBooleanLiteral();
            if (!leftLit.booleanValue()) {
                left = right;
            }
            return left;
        }
        else if (right.isConstant()) {
            // Big error: "left" can have side effects, but they are
            // elimitated by this code. This results in terrible java code.
            // If right is constant and true, then the expression should
            // be true, but left should still be evaluated to let the side
            // effects in it take its course.
            /* Wrong code:
            JBooleanLiteral rightLit = right.getBooleanLiteral();
            if (rightLit.booleanValue()) {
                left = right;
            }
            return left;
            */
            // The right code:
            right = right.getBooleanLiteral();
            return this;
        }
        else {
            return this;
        }
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitBinaryExpression(this, "||", left, right);
    }

    /**
     * Generates a sequence of bytecodes.
     * @param code  the code list.
     */
    public void genCode(CodeSequence code)
    {
        genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
                             JExpression right,
                             boolean cond,
                             CodeSequence code,
                             CodeLabel label)
    {
        if (cond) {
            left.genBranch(true, code, label);
            right.genBranch(true, code, label);
        }
        else {
            CodeLabel  skip = new CodeLabel();
            left.genBranch(true, code, skip);
            right.genBranch(false, code, label);
            code.plantLabel(skip);
        }
    }
}
