/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * A representation of the divide operation (JLS2 15.17.2)
 *
 * @author DMS, Erwin Bolwidt
 */
public strictfp class JDivideExpression
    extends JBinaryArithmeticExpression
{
    /**
     * Type of "java/lang/ArithmeticException".
     */
    private CClassType arithmeticException;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code
     * @param left the left operand
     * @param right the right operand
     */
    public JDivideExpression(TokenReference where,
                             JExpression left, JExpression right)
    {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * Any error during compilation, with position information.
     */
    public JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError
    {
        try {
            if (arithmeticException == null) {
                arithmeticException =
                    CClassType.lookup("java/lang/ArithmeticException");
                arithmeticException.checkType(null);
            }

            type = computeType(left.getType(), right.getType(),
                               KjcMessages.DIVIDE_BADTYPE);

            if (left.isConstant() && right.isConstant()) {
                return constantFolding();
            }
            else {
                /* An integer division may throw an ArithmeticException
                 * if the division is a division-by-zero. This will
                 * always be exactly the type java.lang.ArithmeticException.
                 */
                if (type.isOrdinal()) {
                    CBodyContext bodyContext = context.getBodyContext();
                    CThrowableInfo info = 
                        new CThrowableInfo(arithmeticException, this,
                                           bodyContext.makeSnapshot(),
                                           false); // subTypePossible
                    bodyContext.addThrowable(info);
                }
                return this;
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public int compute(int left, int right)
        throws UnpositionedError
    {
        if (right == 0) {
            throw new UnpositionedError(KjcMessages.DIVIDE_BY_ZERO,
                                        new Integer(left), new Integer(right));
        }
        return left / right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public long compute(long left, long right)
        throws UnpositionedError
    {
        if (right == 0l) {
            throw new UnpositionedError(KjcMessages.DIVIDE_BY_ZERO,
                                        new Long(left), new Long(right));
        }
        return left / right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public float compute(float left, float right)
    {
        return left / right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public double compute(double left, double right)
    {
        return left / right;
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitBinaryExpression(this, "/", left, right);
    }

    /**
     * @param type  the type of result
     * @return the type of opcode for this operation
     */
    public static int getOpcode(CType type)
    {
        switch (type.getTypeID()) {
        case TID_FLOAT:
            return opc_fdiv;
        case TID_LONG:
            return opc_ldiv;
        case TID_DOUBLE:
            return opc_ddiv;
        default:
            return opc_idiv;
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        left.genCode(code);
        right.genCode(code);

        code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
    }
}

