/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * For Statement. (JLS 14.10)
 *
 * The while statement executes an expression and a statement repeatedly
 * until the value of the expression is false.
 */
public class JForStatement
    extends JLoopStatement
{
    /**
     * Initializer in <code>for (initializer; condition; increment)</code>.
     */
    private JStatement init;

    /**
     * Condition in <code>for (initializer; condition; increment)</code>.
     */
    private JExpression condition;

    /**
     * Increment in <code>for (initializer; condition; increment)</code>.
     */
    private JStatement incr;

    /**
     * Body of the for loop.
     */
    private JStatement body;

    /**
     * Construct a node in the parsing tree
     *
     * @param where
     * The line of this node in the source code.
     * @param init
     * The init part.
     * @param condition
     * The condition part, or null if the condition part is empty, as
     * in <code>for (;;)</code>
     * @param incr
     * The increment part.
     * @param body
     * The loop body.
     */
    public JForStatement(TokenReference where,
                         JStatement init,
                         JExpression condition,
                         JStatement incr,
                         JStatement body,
                         JavaStyleComment[] comments)
    {
        super(where, comments);

        this.init = init;
        this.condition = condition;
        this.incr = incr;
        if (body instanceof JBlock) {
            this.body =
                new JCompoundStatement(where, ((JBlock)body).getBody());
        }
        else {
            this.body = body;
        }
    }

    /**
     * Check statement.
     *
     * @param context
     * The actual context of analyse.
     * @exception PositionedError
     * If the check fails
     */
    public void checkStatement(CBodyContext context)
        throws PositionedError
    {
        try {
            if (init != null) {
                init.checkBody(context);
            }

            int vars = 0;
            if (init instanceof JVariableDeclarationStatement) {
                vars = ((JVariableDeclarationStatement)init).getVars().length;
            }
            CBlockContext self = new CBlockContext(context, vars);
            CBodyContext inside = new CBodyContext(self);

            if (condition != null) {
                CExpressionContext ceContext = new CExpressionContext(inside);
                condition = condition.checkExpression(ceContext);
                check(context,
                      condition.getType() == CStdType.Boolean,
                      KjcMessages.FOR_COND_NOTBOOLEAN);
            }

            /* JLS14.20: It is not valid to have a constant false condition
             * in a for loop.
             */
            check(context,
                  condition == null || !condition.isConstant() ||
                  ((JBooleanLiteral)condition.getLiteral()).booleanValue(),
                  KjcMessages.FOR_COND_ISFALSE);

            /* If the condition is not guaranteed constant true, then the
             * loop may not execute at all, and all unassigned local variables
             * may still not have been assigned.
             */
            CBodyContext neverContext = null;
            if (condition != null && !condition.isConstant()) {
                neverContext = inside.cloneContext();
            }

            CLoopContext selfLoop = new CLoopContext(inside, this);
            CBodyContext inside2 = new CBodyContext(selfLoop);

            if (init instanceof JVariableDeclarationStatement) {
                ((JVariableDeclarationStatement)init).setIsInFor();
            }
            body.checkStatement(inside2);
            if (init instanceof JVariableDeclarationStatement) {
                ((JVariableDeclarationStatement)init).unsetIsInFor();
            }
            
            if (neverContext == null && hasBreak()) {
                inside2.adopt(getBreakContextSummary());
                /*
                System.err.println("Reachable-1: " + context.isReachable());
                System.err.println(" inside2.isReachable(): "
                                   + inside2.isReachable());
                System.err.println(" selfLoop.isReachable(): "
                                   + selfLoop.isReachable());
                System.err.println(" self.isReachable(): "
                                   + self.isReachable());
                System.err.println(" inside.isReachable(): "
                                   + inside.isReachable());
                */
            }

            inside2.close(getTokenReference());      
            selfLoop.close(getTokenReference());      
            // System.err.println("Reachable0: " + context.isReachable());
            // System.err.println(" self.isReachable(): "
            //                    + self.isReachable());
            // System.err.println(" inside.isReachable(): "
            //                   + inside.isReachable());


            if (incr != null) {
                incr.checkStatement(inside);
            }
            /*
            System.err.println("Reachable0.5: " + context.isReachable());
            System.err.println(" self.isReachable(): "
                               + self.isReachable());
            System.err.println(" inside.isReachable(): "
                               + inside.isReachable());
            */
            inside.close(getTokenReference());
            self.close(getTokenReference());

            if (neverContext != null) {
                context.merge(neverContext);
            }
            /*
            else if (!hasBreak()) {
                System.err.println("Setting reachable to false");
                context.setReachable(false);
            }
            */
            else {
                context.setReachable(hasBreak());
            }
        }
        catch (CBlockError e) {
            context.reportTrouble(e);
        }

        if (init instanceof JVariableDeclarationStatement) {
            ((JVariableDeclarationStatement)init).dereferenceVariable(context);
        }
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitForStatement(this, init, condition, incr, body);
    }

    /**
     * Generates a sequence of bytescodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        CodeLabel  startLabel = new CodeLabel();
        CodeLabel  condLabel = new CodeLabel();

        code.pushContext(this);

        if (init != null) {
            //  INIT
            init.genCode(code);
        }

        //  GOTO condition
        code.plantLabelRef("for:GOTO condition", opc_goto, condLabel);
        // start:
        code.plantLabel(startLabel);
        if (body != null) {
            //  BODY
            body.genCode(code);
        }

        // incr:
        code.plantLabel(getContinueLabel());
        if (incr != null) {
            //  INCR
            incr.genCode(code);
        }

        // condition:
        code.plantLabel(condLabel);
        if (condition != null) {
            //  COND IFNE start
            condition.genBranch(true, code, startLabel);
        }
        else {
            //  GOTO start
            code.plantLabelRef("for:GOTO start", opc_goto, startLabel);
        }

        // end:
        code.plantLabel(getBreakLabel());

        code.popContext(this);
    }
}

