/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Representation of an instance or static initializer in a class declaration.
 */
public class JInitializerDeclaration
    extends JMethodDeclaration
{
    /**
     * Whether this initializer declaration is used only to check code
     * and that it should not generate code.
     */
    private boolean isDummy;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JInitializerDeclaration(TokenReference where,
                                   JBlock body,
                                   boolean isStatic,
                                   boolean isDummy)
    {
        super(where,
              ACC_PRIVATE + (isStatic ? ACC_STATIC : 0),
              CStdType.Void,
              (isStatic ? JAV_STATIC_INIT : JAV_INIT),
              JFormalParameter.EMPTY,
              CClassType.EMPTY,
              body,
              null,
              null);

        this.isDummy = isDummy;
    }

    /**
     * Return true if this initializer declaration is used only to check code
     * and that it should not generate code.
     */
    public boolean isDummy()
    {
        return isDummy;
    }

    /**
     * createContext.
     *
     * @param parent
     * The parent context or null.
     * @return
     * A CMethodContext that to which this method provides the context.
     */
    public CMethodContext createContext(CClassContext parent)
    {
        return new CInitializerContext(parent, getMethod());
    }

    /**
     * Second pass (quick), check interface looks good
     * Exceptions are not allowed here, this pass is just a tuning
     * pass in order to create informations about exported elements
     * such as Classes, Interfaces, Methods, Constructors and Fields
     * @return true iff sub tree is correct enough to check code
     * @exception PositionedError an error with reference to the source file
     */
    public CSourceMethod checkInterface(CClassContext context)
        throws PositionedError
    {
        CSourceMethod method = super.checkInterface(context);
        /* A JInitializerDeclaration is always synthetic (otherwise it would
         * be a JClassBlock)
         */
        method.setSynthetic(true);
        return method;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context  the actual context of analyse.
     * @return a pure java expression including promote node.
     */
    public void checkStatement(CClassContext context)
    {
        // I've made it !
        // nothing to check.
    }

    protected void checkAccess(CClassContext context,
                               JMethodDeclaration superMethod)
    {
        // nothing to do here
    }

    protected void checkBody(CClassContext context)
    {
        // nothing to do here
    }

    protected void checkOverriding(CClassContext context, CMethod superMethod)
        throws PositionedError
    {
        // nothing to do here
    }
  
    /**
     * Check expression and evaluate and alter context
     * @param context  the actual context of analyse
     * @return a pure java expression including promote node
     * @exception PositionedError  Error catched as soon as possible
     * @exception UnpositionedError Error catched as soon as possible
     */
    public void checkInitializer(CClassContext context)
        throws PositionedError
    {
        if (!getMethod().isStatic() && !isDummy()) {
            context.addInitializer();
        }
        // we want to do that at check intitializers time
        super.checkStatement(context);
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        // Don t print initializers
    }
}
