/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * A LocalVariable used as an lvalue or rvalue expression.
 */
public class JLocalVariableExpression
    extends JExpression
    implements CInitializable
{
    /**
     * The local variable itself.
     */
    JLocalVariable variable;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JLocalVariableExpression(TokenReference where,
                                    JLocalVariable variable)
    {
        super(where);
        this.variable = variable;
    }

    /**
     * Returns true if this field accept assignment.
     */
    public boolean isLValue(CExpressionContext context)
    {
        return !variable.isFinal() || !isInitialized(context);
    }

    /**
     * Returns true if this field is already possible initialized.
     */
    public boolean isInitialized(CExpressionContext context)
    {
        int index = variable.getIndex();
        return !CVariableInfo.isDefinitelyUnassigned(context.getBodyContext()
                                                     .getVariableInfo(index));
        // isInitialized(context.getBodyContext().getVariableInfo(index));
    }

    /**
     * Marks this local variable as 'definately assigned' in this context.
     */
    public void initialize(CBodyContext context)
    {
        context.addVariableInfo(variable.getIndex(),
                                CVariableInfo.initialize());
    }

    /**
     * Returns the position of this variable in the sets of local vars.
     */
    public int getPosition()
    {
        return variable.getPosition();
    }

    /**
     * Compute the type of this expression. Ccalled after parsing.
     *
     * @return the type of this expression.
     */
    public CType getType()
    {
        return variable.getType();
    }

    /**
     * Returns the name of the local variable.
     */
    public String getIdent()
    {
        return variable.getIdent();
    }  

    /**
     * Returns true iff this expression is constant.
     */
    public boolean isConstant()
    {
        return variable.isConstant();
    }

    /**
     * Returns the literal value of this field
     */
    public JLiteral getLiteral()
    {
        return (JLiteral)variable.getValue();
    }  

    /**
     * Returns the local variable.
     */
    public JLocalVariable getVariable()
    {
        return variable;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of the check.
     *
     * @return
     * A pure java expression including promote node.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        CBodyContext bodyContext = context.getBodyContext();

        CMethodContext method = context.getMethodContext();

        /* Solve problem of not-yet known variable position in
         * constructors of local classes.
         */
        if (method instanceof CConstructorContext) {
            CClass clazz = method.getClassContext().getCClass();
            if (clazz.isInnerClass()) {
                variable.setConstructorClass(clazz);
            }
        }

        if (!context.isLeftSide() || !context.discardValue()) {
            variable.setUsed();
        } 
        if (context.isLeftSide()) {
            variable.setAssigned(getTokenReference(), bodyContext);
        }

        int index = variable.getIndex();
        check(context,
              CVariableInfo.isInitialized(bodyContext.getVariableInfo(index))
              || (context.isLeftSide() && context.discardValue()),
              KjcMessages.UNINITIALIZED_LOCAL_VARIABLE,
              variable.getIdent());

        if (variable.isConstant() && !context.isLeftSide()) {
            return variable.getValue();
        }
        
        return this;
    }

    /**
     * Accepts the specified visitor.
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitLocalVariableExpression(this, variable.getIdent());
    }

    public boolean equals(Object o)
    {
        return (o instanceof JLocalVariableExpression)
            && variable.equals(((JLocalVariableExpression)o).variable);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        if (!code.discardValue()) {
            variable.genLoad(code);
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genStartStoreCode(CodeSequence code)
    {
        // nothing to do here
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genEndStoreCode(CodeSequence code, boolean discardValue)
    {
        if (!discardValue) {
            int opcode;

            if (getType().getSize() == 2) {
                opcode = opc_dup2;
            }
            else {
                opcode = opc_dup;
            }
            code.plantInstruction(new NoArgInstruction(opcode));
        }
        variable.genStore(code);
    }
}
