/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * Loop Statement.
 *
 * Root class for loop statements.
 */
public abstract class JLoopStatement
    extends JStatement
{
    /**
     * ...
     */
    private CBodyContext breakContextSummary;

    /**
     * Whether there is a break to this loop statement.
     */
    private boolean breaked;

    /**
     * Whether there is a continue to this loop statement.
     */
    private boolean continued;

    /**
     * The position of the bytecode instruction to jump to for a continue.
     */
    private CodeLabel continueLabel = new CodeLabel();

    /**
     * The position of the bytecode instruction after the loop, which is also
     * to jump to for a break.
     */
    private CodeLabel endLabel = new CodeLabel();

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     */
    public JLoopStatement(TokenReference where, JavaStyleComment[] comments)
    {
        super(where, comments);
    }

    /**
     * Returns true iff this statement is a possible target for a break
     * statement.
     */
    public boolean acceptsBreak()
    {
        return true;
    }

    /**
     * Returns true iff this statement is a possible target for a continue
     * statement.
     */
    public boolean acceptsContinue()
    {
        return true;
    }

    /**
     * Informs statement that there is at least one break to it.
     */
    public void addBreak(CBodyContext context)
    {
        if (breakContextSummary == null) {
            breakContextSummary = context.cloneContext();
        }
        else {
            breakContextSummary.merge(context);
        }

        breaked = true;
    }

    /**
     * Informs statement that there is at least one continue to it.
     */
    public void addContinue(CBodyContext context)
    {
        continued = true;
    }

    /**
     * Returns true if this loop statement has a break to it.
     */
    public boolean hasBreak()
    {
        return breaked;
    }

    /**
     * Returns true if this loop statement has a continue to it.
     */
    public boolean hasContinue()
    {
        return continued;
    }

    /**
     * Returns a summary of all break context
     */
    public CBodyContext getBreakContextSummary()
    {
        return breakContextSummary;
    }

    /**
     * Return the end of this block (for break statement)
     */
    public CodeLabel getBreakLabel()
    {
        return endLabel;
    }

    /**
     * Return the beginning of this block (for continue statement)
     */
    public CodeLabel getContinueLabel()
    {
        return continueLabel;
    }
}
