/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements the modulo '%' operation.
 */
public strictfp class JModuloExpression
    extends JBinaryArithmeticExpression
{

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param where  the line of this node in the source code
     * @param left  the left operand
     * @param right  the right operand
     */
    public JModuloExpression(TokenReference where,
                             JExpression left,
                             JExpression right)
    {
        super(where, left, right);
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     *
     * @return
     * A pure java expression including promote node.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkBinaryBody(CExpressionContext context)
        throws PositionedError
    {
        try {
            type = computeType(left.getType(), right.getType(),
                               KjcMessages.MODULO_BADTYPE);

            if (left.isConstant() && right.isConstant()) {
                return constantFolding();
            }
            else {
                return this;
            }
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public int compute(int left, int right)
        throws UnpositionedError
    {
        if (right == 0) {
            throw new UnpositionedError(KjcMessages.DIVIDE_BY_ZERO,
                                        new Integer(left), new Integer(right));
        }
        return left % right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public long compute(long left, long right)
        throws UnpositionedError
    {
        if (right == 0l) {
            throw new UnpositionedError(KjcMessages.DIVIDE_BY_ZERO,
                                        new Long(left), new Long(right));
        }
        return left % right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public float compute(float left, float right)
        throws UnpositionedError
    {
        return left % right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param left  the first operand
     * @param right  the seconds operand
     * @return the result of the operation
     */
    public double compute(double left, double right)
        throws UnpositionedError
    {
        return left % right;
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitBinaryExpression(this, "%", left, right);
    }


    /**
     * @param type  the type of result
     * @return the type of opcode for this operation
     */
    public static int getOpcode(CType type)
    {
        switch (type.getTypeID()) {
        case TID_FLOAT:
            return opc_frem;
        case TID_LONG:
            return opc_lrem;
        case TID_DOUBLE:
            return opc_drem;
        default:
            return opc_irem;
        }
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());

        left.genCode(code);
        right.genCode(code);

        code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
    }
}
