/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * A name within an expression.
 */
public class JNameExpression
    extends JExpression
{

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code.
     */
    public JNameExpression(TokenReference where, String ident)
    {
        super(where);

        this.ident = ident;
        assert(ident.indexOf('.') == -1); // $$$
        assert(ident.indexOf('/') == -1);
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code.
     */
    public JNameExpression(TokenReference where,
                           JExpression prefix, String ident)
    {
        super(where);

        this.prefix = prefix;
        this.ident = ident;
        assert(ident.indexOf('.') == -1); // $$$
        assert(ident.indexOf('/') == -1);
    }

    /**
     * Construct a node in the parsing tree.
     *
     * @param where the line of this node in the source code.
     */
    public JNameExpression(TokenReference where,
                           String name,
                           boolean fullyQualified)
    {
        super(where);

        int pos = Math.max(name.lastIndexOf("."), name.lastIndexOf("/"));

        if (fullyQualified && pos >= 0) {
            this.prefix = new JNameExpression(where, name.substring(0, pos),
                                              true);
            this.ident = name.substring(pos + 1, name.length()).intern();
        } 
        else {
            this.ident = name.intern();
        }
    }

    /**
     * Compute the type of this expression (called after parsing).
     *
     * @return the type of this expression.
     */
    public CType getType()
    {
        return null;
    }

    /**
     * @return the name of this name expression.
     */
    public String getName()
    {
        return ident;
    }

    /**
     * @return the prefix of this name expression.
     */
    public JExpression getPrefix()
    {
        return prefix;
    }

    /**
     * Returns the longest name available.
     */
    public String getQualifiedName()
    {
        String str = getName();

        if (prefix == null) {
            return str;
        }
        else if (prefix instanceof JNameExpression) {
            return ((JNameExpression)prefix).getQualifiedName() + "." + str;
        }
        else if (prefix instanceof JTypeNameExpression) {
            return ((JTypeNameExpression)prefix)
                .getQualifiedName() + "." + str;
        }
        else {
            return str;
        }
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     *
     * @return
     * A pure java expression including promote node.
     *
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        TokenReference where = getTokenReference();
        try {
            // 6.5.2 Reclassification of Contextually Ambiguous Names
            // If the AmbiguousName is a simple name, consisting of a single
            // Identifier:
            if (prefix == null) {
                // If the Identifier appears within the scope of a local
                // variable declaration or parameter
                JLocalVariable var = context.lookupLocalVariable(ident);
                if (var != null) {
                    return new JLocalVariableExpression(where, var)
                        .checkExpression(context);
                }
 
                // Otherwise, consider the class or interface C within whose
                // declaration the Identifier occurs
                if (context.lookupField(ident) != null) {
                    return ( createClassField(where, ident)
                             .checkExpression(context) );
                }

                // If the Identifier appears within (an outer) scope of a
                // local variable declaration or parameter
                JExpression outer = ( context.getBlockContext()
                                      .lookupOuterLocalVariable(where, ident));
                if (outer != null) {
                    return outer.checkExpression(context);
                }
 
                // Otherwise, outers
                if (context.getClassContext().lookupOuterField(ident)
                    != null) {

                    return ( createClassField(where, ident)
                             .checkExpression(context) );
                }
 
                try {
                    // Otherwise: TypeName
                    CClassType type = CClassType.lookup(ident);
                    // CClassType type = context.lookupClass(ident);
                    type.checkType(context);
                    return new JTypeNameExpression(where, type);
                }
                catch (UnpositionedError cue) {
                    // Otherwise: PackageName
                    throw new CLineError(where,
                                         KjcMessages.VAR_UNKNOWN, ident);
                }
            }

            assert(prefix != null);
      
            try {
                prefix = prefix.checkExpression(context);
            }
            catch (CLineError cue) {
                if (cue.hasDescription(KjcMessages.VAR_UNKNOWN)) {
                    return convertToPackageName(cue, context);
                }
                else {
                    throw cue;
                }
            }

            // If the name to the left of the "." is reclassified as a TypeName
            if (prefix instanceof JTypeNameExpression) {
                JTypeNameExpression typePrefix = (JTypeNameExpression)prefix;
                if (typePrefix.getClassType().getCClass().lookupField(ident)
                    != null) {

                    // there is a field
                    return ( createClassField(where, prefix, ident)
                             .checkExpression(context) );
                }
                else {
                    // no field => should be a type name
                    return new JTypeNameExpression
                        (where, typePrefix.getQualifiedName() + "/" + ident);
                }
            }
      
            // If the name to the left of the "." is reclassified as an
            // ExpressionName
      
            if (ident == Constants.JAV_LENGTH) {
                // is it an array access ?
                if (prefix.getType().isArrayType()) {
                    JExpression result =
                        new JArrayLengthExpression(where, prefix);
                    return result.checkExpression(context);
                }
            }
      
            return ( createClassField(where, prefix, ident)
                     .checkExpression(context) );
        }
        catch (UnpositionedError e) {
            throw e.addPosition(where);
        }
    }

    /**
     * Try to convert to a package name
     */
    private JExpression convertToPackageName(CLineError cue,
                                             CExpressionContext context)
        throws PositionedError
    {
        TokenReference where = getTokenReference();
        try {
            if (prefix instanceof JNameExpression) {
                CClassType type =
                    CClassType.lookup(((JNameExpression)prefix).getName()
                                      + '/' + ident);
                type.checkType(context);
                return new JTypeNameExpression(where, type);
            }
            else if (!(prefix instanceof JTypeNameExpression)) {
                throw new CLineError(where, KjcMessages.VAR_UNKNOWN, ident);
            }
            else {
                return prefix;
            }
        }
        catch (UnpositionedError cue2) {
            // Otherwise: PackageName
            ident = ((JNameExpression)prefix).getName() + '/' +  ident;
            throw new CLineError(where, KjcMessages.VAR_UNKNOWN, ident);
        }
    }

    /**
     * Since class field may be overloaded in sub compiler, this method allows
     * you to specify the type of class field without needed to touch
     * the huge method above !
     */
    protected JClassFieldExpression createClassField(TokenReference ref,
                                                     JExpression prefix,
                                                     String ident)
    {
        return new JClassFieldExpression(ref, prefix, ident);
    }

    /**
     * Since class field may be overloaded in sub compiler, this method allows
     * you to specify the type of class field without needed to touch
     * the huge method above !
     */
    protected JClassFieldExpression createClassField(TokenReference ref,
                                                     String ident)
    {
        return new JClassFieldExpression(ref, ident);
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitNameExpression(this, prefix, ident);
    }

    /**
     * Generates a sequence of bytecodes
     * @param code  the code list
     */
    public void genCode(CodeSequence code)
    {
        fail();
    }

    private JExpression prefix;
    private String ident;
}
