/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.Constants;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a new allocation expression with a special class type:
 * an anonymous inner class.
 */
public class JNewAnonymousClassExpression
    extends JNewObjectExpression
{
    /**
     * The declaration of the anonymous inner class.
     */
    protected JClassDeclaration decl;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code
     * @param objectType the type of this object allocator
     * @param explicitThis a TypeName expression
     * @param args  parameters to be passed to constructor
     */
    public JNewAnonymousClassExpression(TokenReference where,
                                        CClassType objectType,
                                        JExpression explicitThis,
                                        String identifier,
                                        JExpression[] args,
                                        JClassDeclaration decl)
    {
        super(where, objectType, explicitThis, identifier, args);

        this.decl = decl;
    }

    /**
     * If the declaration in which this expression occurs is static, then the
     * class declaration of the anonymous inner class is also static; pass
     * it on!
     */
    public void setDeclarationIsStatic(boolean flag)
    {
        if (flag) {
            decl.setModifiers(decl.getModifiers() | Constants.ACC_STATIC);
        }
        else {
            decl.setModifiers(decl.getModifiers() & ~Constants.ACC_STATIC);
        }
    }

    /**
     * Checks types of this expression.
     */
    protected CType[] checkTypes(CExpressionContext context)
        throws PositionedError
    {
        CType[] argsTypes = super.checkTypes(context);
        CClass owner = context.getClassContext().getCClass();    

        if (type.getCClass().isInterface()) {
            decl.setInterfaces(new CClassType[] { type });
        }
        else {
            decl.setSuperClass(type.getCClass().getQualifiedName());
        }

        if (context.getMethodContext().getCMethod().isStatic()) {
            decl.setModifiers(decl.getModifiers() | Constants.ACC_STATIC);
        }

        String syntheticIdent =
            String.valueOf(context.getClassContext().getNextSyntheticIndex())
            .intern();

        decl.setIdent(syntheticIdent);

        decl.generateInterface((Main)context.getCompiler(),
                               owner, 
                               owner.getQualifiedName() + "$");
        decl.getCClass().setSynthetic(true);
        /* Because this construct goes completely against the normal order
         * of type checking, it is necessary to temporarily set the superclass
         * with the following line, to allow the look for the superclass-
         * constructor to find protected constructors.
         */
        decl.getCClass().setSuperClass(type);

        // add implicit constructor
        JConstructorDeclaration cstr;
        CMethod superCstr;
        try {
            superCstr = type.getCClass().lookupMethod(decl.getCClass(),
                                                      JAV_CONSTRUCTOR,
                                                      argsTypes);
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }

        if (superCstr == null) {
            throw new CMethodNotFoundError(getTokenReference(),
                                           null, type.toString(), argsTypes);
        }

        CType[] parameters = superCstr.getParameters();
        JFormalParameter[] fargs = new JFormalParameter[parameters.length];
        CClassType[] throwables = superCstr.getThrowables();

        for (int i = 0; i < parameters.length; i++) {
            fargs[i] = new JFormalParameter(getTokenReference(),
                                            JLocalVariable.DES_GENERATED,
                                            parameters[i],
                                            "p$" + i,
                                            true);
        }

        JExpression[] checkedParams = new JExpression[args.length];

        for (int i = 0; i < args.length; i++) {
            checkedParams[i] =
                new JLocalVariableExpression(getTokenReference(), fargs[i]);
        }

        cstr = new JConstructorDeclaration
            (getTokenReference(),
             ACC_PUBLIC,
             // syntheticIdent,
             decl.getCClass().getIdent(),
             fargs,
             throwables,
             new JConstructorBlock(getTokenReference(), new JStatement[] {
                 new JExpressionStatement
                     (getTokenReference(),
                      new JExplicitConstructorInvocation(getTokenReference(),
                                                         null, "super",
                                                         checkedParams),
                      null)
                     }), null, null);
        decl.setDefaultConstructor(cstr);
        decl.checkInterface(context.getCompilationUnit());
        decl.checkInitializers(context);
        decl.checkTypeBody(context);

        type = decl.getCClass().getType();

        return argsTypes;
    }

    /**
     * Accepts the specified visitor.
     *
     * @param p
     * The visitor.
     */
    public void accept(KjcVisitor p)
    {
        p.visitNewObjectExpression(this, type, args, decl);
    }
}
