/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.classfile.*;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a new allocation expression 'new toto(1)'
 */
public class JNewObjectExpression
    extends JExpression
{
    /**
     * If the class to instantiate is an inner class, an explicit-this
     * expression, if present, supplies the outer-class instance to the
     * inner class.
     */
    protected JExpression explicitThis;

    /**
     * If an explicit this is given, then the class to construct cannot
     * be resolved very early, since it must be resolved in the scope of the
     * type of the explicitThis expression; so then the identifier is stored
     * here and when the type of explicitThis is known, the identifier is
     * looked up as the name of an inner class in the type of explicitThis.
     */
    protected String identifier;

    /**
     * Arguments to the constructor.
     */
    protected JExpression[] args;

    /**
     * The class to construct.
     */
    protected CClassType type;

    /**
     * The class in which this new-object expression occurs.
     */
    protected CClass local;

    /**
     * The constructor to invoke.
     */
    protected CMethod constructor;

    /**
     * True if the constructor is not directly accessible but only by creating
     * an accessor method in the class that calls the real constructor.
     */
    private boolean mustUseAccessorMethod = false;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where
     * The line of this node in the source code.
     * @param objectType
     * The type of this object allocator.
     * @param explicitThis
     * An explicit this reference for outerInstance.new InnerClass() kind-of
     * invocations.
     * @param args
     * Parameters to be passed to constructor.
     */
    public JNewObjectExpression(TokenReference where,
                                CClassType objectType,
                                JExpression explicitThis,
                                String identifier,
                                JExpression[] args)
    {
        super(where);

        this.type = objectType;
        this.args = args;
        this.explicitThis = explicitThis;
        this.identifier = identifier;
    }

    /**
     * Compute the type of this expression. Called after parsing.
     *
     * @return the type of this expression.
     */
    public CType getType()
    {
        return type;
    }

    /**
     * Returns false, since a new-object expression alway returns objects
     * of exactly the type return in the <code>type</code> property of this
     * JNewObjectExpression.
     *
     * Returns whether this expression can result in subtypes of the type
     * returned by the <code>getType()</code> method, or whether it always
     * returns the exact type returned by <code>getType()</code>.
     * The return value of this method has no meaning unless getType() returns
     * a Class-type. (<code>CType.isClassType()</code>)
     */
    public boolean isSubTypePossible()
    {
        return false;
    }

    /**
     * Returns true iff this expression can be used as a statement.
     * (JLS 14.7)
     */
    public boolean isStatementExpression()
    {
        return true;
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The actual context of analyse.
     * @return
     * A pure java expression including promote node.
     * @exception PositionedError
     * If the check fails.
     */
    public JExpression checkExpression(CExpressionContext context)
        throws PositionedError
    {
        CType[] argsType = checkTypes(context);
        return checkCall(context, argsType);
    }

    /**
     * ...
     */
    protected JExpression checkCall(CExpressionContext context,
                                    CType[] argsType)
        throws PositionedError
    {
        context.setDiscardValue(false);

        check(context, !type.getCClass().isAbstract(),
              KjcMessages.NEW_ABSTRACT, type);

        check(context, !type.getCClass().isInterface(),
              KjcMessages.NEW_INTERFACE, type);

        CClass caller = context.getClassContext().getCClass();
        try {

            constructor = type.getCClass().lookupMethod(caller,
                                                        JAV_CONSTRUCTOR,
                                                        argsType);
        }
        catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

        if (constructor == null
            || constructor.getOwner() != type.getCClass()) {

            // May be an inner class
            if (type.getCClass().isInnerClass()) {
                CType[]  argsType2 = new CType[argsType.length + 1];
                System.arraycopy(argsType, 0, argsType2, 0, argsType.length);
                argsType2[argsType.length] =
                    context.getClassContext().getCClass().getType();
                try {
                    constructor =
                        type.getCClass().lookupMethod(caller,
                                                      JAV_CONSTRUCTOR,
                                                      argsType2);
                }
                catch (UnpositionedError e) {
                    throw e.addPosition(getTokenReference());
                }
            }
            if (constructor == null
                || constructor.getOwner() != type.getCClass()) {

                // do not want a super constructor !
                throw new CMethodNotFoundError(getTokenReference(), null,
                                               type.toString(), argsType);
            }
        }

        /* Check access */
        local = context.getClassContext().getCClass();
        check(context,
              constructor.isAccessible(local),
              KjcMessages.CONSTRUCTOR_NOACCESS, type);

        CClass constructClass = constructor.getOwner();
        CClass constructOuter = constructClass.getOwner();
        CClass localOuter = local.getOwner();

        if ( constructClass.hasOuterThis() && explicitThis == null ) {
            check(context,
                  !context.getMethodContext().getCMethod().isStatic(),
                  KjcMessages.INNER_INHERITENCE,
                  constructClass.getType(), local.getType());

            CClass outer = local;
            explicitThis = new JThisExpression(getTokenReference());
            explicitThis.checkExpression(context);

            while (outer != null) {
                if (outer.descendsFrom(constructOuter)) {
                    break;
                }
                explicitThis =
                    new JClassFieldExpression(getTokenReference(),
                                              explicitThis, JAV_OUTER_THIS);
                explicitThis.checkExpression(context);
                outer = outer.getOwner();
            }

            check(context,
                  outer != null,
                  KjcMessages.INNER_INHERITENCE,
                  constructClass.getType(), local.getType());
        }

        CClassType[] exceptions = constructor.getThrowables();
        CBodyContext bodyContext = context.getBodyContext();
        CBodyContext snapshot = bodyContext.makeSnapshot();
        for (int i = 0; i < exceptions.length; i++) {
            CThrowableInfo info = new CThrowableInfo(exceptions[i], this,
                                                     snapshot,
                                                     true); // subTypePossible
            bodyContext.addThrowable(info);
        }

        /* New-object expressions can always result in a RuntimeException,
         * methods and constructors don't have to declare RuntimeExceptions.
         */
        CThrowableInfo info = new CThrowableInfo(CStdType.RuntimeException,
                                                 this, snapshot,
                                                 true); // subTypePossible
        bodyContext.addThrowable(info);
    
        argsType = constructor.getParameters();

        for (int i = 0; i < args.length; i++) {
            args[i] = args[i].convertType(argsType[i], context);
        }

        if (constructor.mustUseAccessorMethod(local)) {
            /* Make sure that accessor method is generated in checking phase,
             * so it will end up in the class file.
             */
            ((CSourceMethod)constructor).getAccessor();
            this.mustUseAccessorMethod = true;
        }

        return this;
    }

    /**
     * ...
     */
    protected CType[] checkTypes(CExpressionContext context)
        throws PositionedError
    {
        try {
            if (explicitThis != null) {
                explicitThis = explicitThis.checkExpression(context);
                
                CType explicitType = explicitThis.getType();
                check(context,
                      explicitType.isClassType(),
                      KjcMessages.QUALIFIEDCREATE_NOTCLASS, explicitType);
                CClass explicitClass = ((CClassType)explicitType).getCClass();
                
                CClass newClass;
                
                newClass = explicitClass.lookupMemberClass(identifier);
                if (newClass == null) {
                    throw new UnpositionedError
                        ( KjcMessages.TYPE_UNKNOWN,
                          explicitClass.getQualifiedName() + "$" + identifier
                          );
                }
                type = newClass.getType();
                local = context.getClassContext().getCClass();

                check(context,
                      newClass.isAccessible(local),
                      KjcMessages.CLASS_NOACCESS, type);

                check(context,
                      !newClass.isStatic(),
                      // BAD!!! EJB Make new KjcMessage! Meaning:
                      // this construct cannot be used with static
                      // member classes.
                      KjcMessages.CLASS_NOACCESS, type);

            }
            else {

            }

            type.checkType(context);
            
            CType[] argsType = new CType[args.length];
            
            for (int i = 0; i < argsType.length; i++) {
                args[i] = args[i].checkExpression(context);
                argsType[i] = args[i].getType();
                assert(argsType[i] != null);
            }

            return argsType;
        }
        catch (UnpositionedError e) {
            // DEBUG - -XXX make depend on debug flag in compiler.
            e.printStackTrace();
            throw e.addPosition(getTokenReference());
        }
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitNewObjectExpression(this, type, args, null);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        boolean discardValue = code.discardValue();
        code.setLineNumber(getTokenReference().getLine());

        if (!mustUseAccessorMethod) {
            code.plantInstruction
                (new ClassRefInstruction(opc_new,
                                         type.getCClass().getQualifiedName()));
            if (!discardValue) {
                code.plantInstruction(new NoArgInstruction(opc_dup));
            }
        }


        code.setDiscardValue(false);

        CClass instantiateClass = constructor.getOwner();

        // Synthetic this for inner classes should be argument at index 0,
        // so generate first. Fixed EJB.

        // Check if constructor is inside an inner class.
        if (instantiateClass.isInnerClass()) {

            CClass outerClassOfExpression = local.getOwner();
            CClass outerClassOfInnerClass = instantiateClass.getOwner();

            /* [EJB] New rule: if explicit this is specified, use it.
             * But how about inner classes in methods that take parameters
             * from that method? They're not always anonymous.
             */
            if (explicitThis != null) {
                instantiateClass.genSyntheticParams(code, explicitThis); 
            }
            else if (outerClassOfExpression == null
                     || outerClassOfExpression != outerClassOfInnerClass) {

                /* Either the 'new' expression is in a class that is
                 * a toplevel class, or the 'new' expression is in an
                 * inner class that does NOT have the same outer class
                 * as the inner class being constructed.
                 */
                instantiateClass.genSyntheticParams(code, null); 
            }
            else if (!instantiateClass.isStatic()
                     && instantiateClass.hasOuterThis()) {

                /* So the class to instantiate is an inner sibling of the
                 * class that contains the 'new' expression, so instead of
                 * passing "this" to the inner class, JAV_OUTER_THIS
                 * must be passed to the inner class.
                 */
                if (outerClassOfExpression != null
                    && outerClassOfExpression == outerClassOfInnerClass) {

                    code.plantLoadThis();
                    String localSignature = local.getType().getSignature();
                    String signatureSubString =
                        localSignature.substring(1, localSignature.length()-1);

                    code.plantInstruction
                        (new FieldRefInstruction
                            (opc_getfield,
                             signatureSubString,
                             JAV_OUTER_THIS,
                             outerClassOfExpression.getType().getSignature()));
                }
            }
        }

        // Generate code for non-synthetic arguments for this 'new' call.
        for (int i = 0; i < args.length; i++) {
            args[i].genCode(code);
        }

        if (mustUseAccessorMethod) {
            try {
                CSourceMethod accessor =
                    ((CSourceMethod)constructor).getAccessor();
                accessor.genCode(code, false);
            }
            catch (PositionedError e) {
                // Shouldn't happen; no invalid constructor accessors should
                // be generated.
                throw new RuntimeException(e.toString());
            }

            if (discardValue) {
                code.plantInstruction(new NoArgInstruction(opc_pop));
            }
        }
        else {
            code.plantInstruction
                (new MethodRefInstruction(opc_invokespecial,
                                          constructor.getPrefixName(),
                                          constructor.getIdent(),
                                          constructor.getSignature()));
        }
    }

}
