/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.kjc;

import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.TokenReference;

/**
 * Represenation of a string constant.
 *
 * @author DMS, Erwin Bolwidt
 */
public class JStringLiteral
    extends JLiteral
{
    /**
     * The String-value of this string literal.
     */
    private String value;

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param image  the string representation of this literal
     */
    public JStringLiteral(TokenReference where, String image)
    {
        this(where, image, false);
    }

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param image  the string representation of this literal
     * @param quoted  there is quote around image
     */
    public JStringLiteral(TokenReference where, String image, boolean quoted)
    {
        super(where);

        if (image == null) {
            throw new RuntimeException();
        }

        if (quoted) {
            StringBuffer s = new StringBuffer();
            for (int i = 0; i < image.length(); i++) {
                char c = image.charAt(i);
                if (c == '\\') {
                    if (i + 1 < image.length() - 1) {
                        i++;
                        c = image.charAt(i);
                        switch (c) {
                        case 'n' : c = '\n'; break;
                        case 'r' : c = '\r'; break;
                        case 't' : c = '\t'; break;
                        case 'b' : c = '\b'; break;
                        case 'f' : c = '\f'; break;
                        case '"' : c = '\"'; break;
                        case '\'' : c = '\''; break;
                        case '\\' : c = '\\'; break;
                        }
                    }
                }
                s.append(c);
            }
            value = s.toString();
            value = value.substring(1, value.length() - 1);
        }
        else {
            value = image;
        }
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType()
    {
        return CStdType.String;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before checkBody
     */
    public Object getValue()
    {
        return value;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before checkBody
     */
    public String stringValue()
    {
        return value;
    }

    /**
     * Returns the literal value of an expression.
     */
    public boolean isDefault()
    {
        return false;
    }

    /**
     * Check expression and evaluate and alter context
     * @param context the actual context of analyse
     * @return  a pure java expression including promote node
     * @exception PositionedError  if the check fails
     */
    public JExpression checkExpression(CExpressionContext context)
    {
        return this;
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        StringBuffer s = new StringBuffer();
        for (int i = 0; i < value.length(); i++) {
            char c = value.charAt(i);
            switch (c) {
            case '\n' : s.append("\\n"); break;
            case '\r' : s.append("\\r"); break;
            case '\t' : s.append("\\t"); break;
            case '\b' : s.append("\\b"); break;
            case '\f' : s.append("\\f"); break;
            case '\"' : s.append("\\\""); break;
            case '\'' : s.append("\\\'"); break;
            case '\\' : s.append("\\\\"); break;
            default:
                s.append(c);
            }
        }
        value = s.toString();
        p.visitStringLiteral(value);
    }

    /**
     * Generates a sequence of bytecodes.
     *
     * @param code
     * The code list.
     */
    public void genCode(CodeSequence code)
    {
        code.setLineNumber(getTokenReference().getLine());
        code.plantInstruction(new PushLiteralInstruction(value));
    }
}
