/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a local variable definition in the syntax tree.
 */
public class JVariableDefinition
    extends JLocalVariable
{
    /**
     * Whether the variable was declared inside a switch. If that's the
     * case, it's not constant. See jacks test for JLS2 14.10; this is
     * quite complex and not directly explained in the JLS2, but a logical
     * conclusion of it: final vars in a switch body cannot be really constant
     * because declaration is on the main execution path, but initialization
     * can occur independantly on different execution paths.
     */
    private boolean declaredInSwitch;

    /**
     * Construct a node in the parsing tree.
     *
     * @param where  the line of this node in the source code.
     * @param ident  the name of this variable.
     * @param initializer the initializer.
     */
    public JVariableDefinition(TokenReference where,
                               int modifiers,
                               CType type,
                               String ident,
                               JExpression initializer)
    {
        super(where, modifiers, DES_LOCAL_VAR, type, ident, initializer);
        assert(type != null);
    }

    /**
     * hasInitializer
     * @return true if there is an initializer
     */
    public boolean hasInitializer()
    {
        return getValue() != null;
    }

    /**
     * @return the initial value
     */
    public JExpression getValue()
    {
        return initializer;
    }

    /**
     * Second pass (quick), check interface looks good.
     * Exceptions are not allowed here, this pass is just a tuning
     * pass in order to create informations about exported elements
     * such as Classes, Interfaces, Methods, Constructors and Fields
     *
     * @param context  the current context
     * @return true iff sub tree is correct enought to check code
     */
    public void checkInterface(CClassContext context)
    {
        try {
            type.checkType(context);
        }
        catch (UnpositionedError cue) {
            /*checkbody will do it*/
        }
    }

    /**
     * Return true if this variable is constant.
     * Like in: <code>int (final int i) {}</code>
     * @return true if this variable is constant
     */
    public boolean isConstant()
    {
        return ( !declaredInSwitch &&
                 hasFlag(getModifiers(), Constants.ACC_FINAL)
                 && initializer != null && initializer.isConstant() );
    }

    /**
     * Check expression and evaluate and alter context.
     *
     * @param context
     * The context of the check.
     * @exception PositionedError
     * Error caught as soon as possible.
     */
    public void checkBody(CBodyContext context)
        throws PositionedError
    {
        try {
            type.checkType(context);

            if (initializer != null) {
                /* special case for array initializers */
                if (initializer instanceof JArrayInitializer) {
                    ((JArrayInitializer)initializer).setType(type);
                }

                /* Special case for anonymous inner class instantiation
                 * expressions: if the declaration is static, then
                 * so is the inner class!
                 */
                if (initializer instanceof JNewAnonymousClassExpression) {
                    ((JNewAnonymousClassExpression)initializer)
                        .setDeclarationIsStatic(isStatic());
                }

                CExpressionContext expressionContext =
                    context.createExpressionContext();
                initializer = initializer.checkExpression(expressionContext);
                if (initializer instanceof JTypeNameExpression) {
                    JTypeNameExpression typeName =
                        (JTypeNameExpression)initializer;

                    check(context,
                          false,
                          KjcMessages.VAR_UNKNOWN,
                          typeName.getQualifiedName());
                }

                /* The following violates the rules on assignment with
                 * constant expressions. (Where the type of expression doesn't
                 * matter if the value is in range of the type of lvalue.
                 * EJB: sorry, no they don't since
                 * JOrdinalLiteral.isAssignebleTo now correctly follows the
                 * JLS.
                 */
                if (!initializer.isAssignableTo(type)) {
                    check(context,
                          false,
                          KjcMessages.VAR_INIT_BADTYPE,
                          getIdent(), initializer.getType());
                }
                else {
                    initializer = initializer.convertType(type,
                                                          expressionContext);
                }

                CContext lookForSwitch = context;
                while (lookForSwitch != null
                       && (lookForSwitch instanceof CBodyContext)) {
                    /* Skip the body contexts, there may be a few between
                     * here and a context type that is important to us.
                     */
                    lookForSwitch = lookForSwitch.getParentContext();
                }

                if (lookForSwitch instanceof CSwitchBodyContext) {
                    // instanceof doesn't succeed if the left argument is
                    // null, so this is safe.
                    declaredInSwitch = true;
                }
            }
        }
        catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
    }

    /**
     * Accepts the specified visitor
     * @param p  the visitor
     */
    public void accept(KjcVisitor p)
    {
        p.visitVariableDefinition(this, modifiers, type, getIdent(),
                                  initializer);
    }
}
