/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import antlr.Token;

import at.dms.compiler.CToken;
import at.dms.compiler.Compiler;
import at.dms.compiler.InputBuffer;
import at.dms.compiler.Scanner;
import at.dms.compiler.PositionedError;
import at.dms.compiler.CompilerMessages;

/**
 * This is an handwritten scanner available for speed reasons.
 */
public class KjcScanner
    extends Scanner
{

    /**
     * Constructs a java scanner.
     */
    public KjcScanner(Compiler compiler, InputBuffer buffer)
    {
        super(compiler, buffer);
    }
  
    // Tokens are: Identifiers, Keywords, Literals, Separators, Operators.
    protected Token getToken()
        throws PositionedError
    {
        switch (buffer.data[buffer.pos]) {
            // separators: (period is a special case)
        case '(':
            buffer.pos++;
            return LPAREN;
        case ')':
            buffer.pos++;
            return RPAREN;
        case '{':
            buffer.pos++;
            return LCURLY;
        case '}':
            buffer.pos++;
            return RCURLY;
        case '[':
            buffer.pos++;
            return LBRACK;
        case ']':
            buffer.pos++;
            return RBRACK;
        case ';':
            buffer.pos++;
            return SEMI;
        case ',':
            buffer.pos++;
            return COMMA;
      
            // Operators (single-char):
        case '~':
            buffer.pos++;
            return BNOT;
        case '?':
            buffer.pos++;
            return QUESTION;
        case ':':
            buffer.pos++;
            return COLON;
      
        case '=':
        case '>':
        case '<':
        case '!':
        case '&':
        case '|':
        case '+':
        case '-':
        case '*':
        case '^':
        case '%':
            return getOperator();
        case '/':
            if (readJavaComment()) {
                return null;
            }
            else {
                return getOperator();
            }
        case '\'':
            return getCharLiteral();
        case '\"':
            return getStringLiteral();
      
            // a period is a special case:
        case '.':
            if (Character.digit(buffer.data[buffer.pos + 1],10) != -1) {
                return getNumericLiteral();
            }
            else {
                buffer.pos++;
                return DOT;
            }
      
        default:
            break;
        }
        if (isJavaIdentifierStart(buffer.data[buffer.pos])) {
            return getIdentifier();
        }
        if (Character.isDigit(buffer.data[buffer.pos])) {
            return getNumericLiteral();
        }
        throw new PositionedError(getTokenReference(),
                                  CompilerMessages.ILLEGAL_CHAR,
                                  "" + buffer.data[buffer.pos]);
    }

    /**
     * Returns next token as identifier.
     */  
    protected Token getIdentifier()
        throws PositionedError
    {
        int  start = buffer.pos;
        Token tok = null;
    
        if (readIdentifier()) {
            tok = lookupKeyword(buffer.data, start, buffer.pos - start);
        }
   
        if (tok != null) {
            return tok;
        }
        else {
            return CToken.lookupToken(KjcTokenTypes.IDENT,
                                      buffer.data, start, buffer.pos - start);
        }
    }
  
    protected Token lookupKeyword(char[] data, int start, int length)
    {
        return KjcKeywords.lookup(data, start, length);
    }
  
    protected boolean readIdentifier()
        throws PositionedError
    {
        int  mayBeAKeyword = 0;
        char c;
    
        while (true) {
            c = buffer.data[++buffer.pos];
            if (c >= 'a' && c <= 'z') {
                continue;
            }
            mayBeAKeyword++;
            if (c >= 'A' && c <= 'Z') {
                continue;
            }
            if (c == ' ' || c == '\t'
                || c == ';'
                || !Character.isJavaIdentifierPart(buffer.data[buffer.pos])) {
                break;
            }
        }
    
        return mayBeAKeyword == 1;
    }
  
    protected Token getNumericLiteral()
    {
        int i;
        // leading decimal indicates float.
        if (buffer.data[buffer.pos] == '.') {
            return getFloatingPointLiteral();
        }
        // 0x indicates Hex.
        if (buffer.data[buffer.pos]=='0' &&
            (buffer.data[buffer.pos + 1]=='x' ||
             buffer.data[buffer.pos + 1]=='X')) {
            return getIntegerLiteral(2, 16);
        }
        // otherwise scan to first non-numeric
        for (i = buffer.pos; Character.digit(buffer.data[i], 10) != -1; ) {
            i++;
        }
        switch (buffer.data[i]) { // discriminate based on first non-numeric
        case '.':
        case 'f':
        case 'F':
        case 'd':
        case 'D':
        case 'e':
        case 'E':
            return getFloatingPointLiteral();
        case 'L':
        case 'l':
        default:
            if (buffer.data[buffer.pos]=='0') {
                return getIntegerLiteral(1, 8);
            }
            return getIntegerLiteral(0, 10);
        }
    }
  
    protected Token getIntegerLiteral(int prefix, int radix)
    {
        int start = buffer.pos;
        buffer.pos += prefix;
        while (Character.digit(buffer.data[buffer.pos], radix) != -1) {
            buffer.pos++;
        }
        if (buffer.data[buffer.pos] == 'l' || buffer.data[buffer.pos] == 'L') {
            buffer.pos++;
        }
    
        return new CToken(KjcTokenTypes.INTEGER_LITERAL,
                          new String(buffer.data, start, buffer.pos - start));
    }
  
    protected Token getFloatingPointLiteral()
    {
        int start = buffer.pos;
        while (Character.digit(buffer.data[buffer.pos], 10) != -1) {
            buffer.pos++;
        }
        if (buffer.data[buffer.pos] == '.') {
            buffer.pos++;
            while (Character.digit(buffer.data[buffer.pos], 10) != -1) {
                buffer.pos++;
            }
        }
        if (buffer.data[buffer.pos] == 'e' || buffer.data[buffer.pos] == 'E') {
            buffer.pos++;
            if (buffer.data[buffer.pos]=='+'
                || buffer.data[buffer.pos] == '-') {
                buffer.pos++;
            }
            while (Character.digit(buffer.data[buffer.pos], 10) != -1) {
                buffer.pos++;
            }
        }
        switch (buffer.data[buffer.pos]) {
        case 'f':
        case 'F':
        case 'd':
        case 'D':
            buffer.pos++;
        }
        return new CToken(KjcTokenTypes.REAL_LITERAL,
                          new String(buffer.data, start, buffer.pos - start));
    }
  
    protected Token getOperator()
    {
        char first = buffer.data[buffer.pos++];
        char second= buffer.data[buffer.pos];
    
        switch (first) {
            // doubled operators
        case '+':
            if (second == first) {
                buffer.pos++;
                return INC;
            }
            break;
        case '-':
            if (second == first) {
                buffer.pos++;
                return DEC;
            }
            break;
        case '&':
            if (second == first) {
                buffer.pos++;
                return LAND;
            }
            break;
        case '|':
            if (second == first) {
                buffer.pos++;
                return LOR;
            }
            break;
        default:
            break;
        }
        // Check for trailing '='
        if (second=='=') {
            buffer.pos++;
            switch (first) {
            case '=':
                return EQUAL;
            case '!':
                return NOT_EQUAL;
            case '&':
                return BAND_ASSIGN;
            case '|':
                return BOR_ASSIGN;
            case '+':
                return PLUS_ASSIGN;
            case '-':
                return MINUS_ASSIGN;
            case '*':
                return STAR_ASSIGN;
            case '/':
                return SLASH_ASSIGN;
            case '^':
                return BXOR_ASSIGN;
            case '%':
                return PERCENT_ASSIGN;
            case '<':
                return LE;
            case '>':
                return GE;
            default:
                throw new RuntimeException("BAD OP");
            }
        }
        // Special-case '<<', '>>' and '>>>'
        if (first=='<' && second=='<') {
            buffer.pos++;
            if (buffer.data[buffer.pos]=='=') {
                buffer.pos++;
                return SL_ASSIGN;
            }
            else {
                return SL;
            }
        }
        else if (first=='>' && second=='>') {
            buffer.pos++;
            if (buffer.data[buffer.pos] == '>') {
                buffer.pos++;
                if (buffer.data[buffer.pos] == '=') {
                    buffer.pos++;
                    return BSR_ASSIGN;
                }
                else {
                    return BSR;
                }
            }
            else if (buffer.data[buffer.pos] == '=') {
                buffer.pos++;
                return SR_ASSIGN;
            }
            else {
                return SR;
            }
        }
    
        switch (first) {
        case '=':
            return ASSIGN;
        case '!':
            return LNOT;
        case '&':
            return BAND;
        case '|':
            return BOR;
        case '+':
            return PLUS;
        case '-':
            return MINUS;
        case '*':
            return STAR;
        case '/':
            return SLASH;
        case '^':
            return BXOR;
        case '%':
            return PERCENT;
        case '>':
            return GT;
        case '<':
            return LT;
        default:
            throw new RuntimeException("BAD OP" + first + second);
        }
    }
 
    /**
     * Parses a character literal.
     */ 
    protected Token getCharLiteral()
        throws PositionedError
    {
        char firstquote = buffer.data[buffer.pos++];
        char val;
        switch (buffer.data[buffer.pos]) {
        case '\\':
            val = scanEscapeSequence();
            break;
        case '\'':
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR,
                                      "" + buffer.data[buffer.pos]);
        case '\n':
        case '\r':
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR,
                                      "" + buffer.data[buffer.pos]);
        default:
            val = buffer.data[buffer.pos++];
        }
        char secondquote = buffer.data[buffer.pos++];
        if (firstquote != '\'' || secondquote != '\'') {
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR,
                                      "" + secondquote);
        }
    
        return new CToken(KjcTokenTypes.CHARACTER_LITERAL,
                          String.valueOf(val));
    }
  
    protected Token getStringLiteral()
        throws PositionedError
    {
        int   start = buffer.pos;
        char  openquote = buffer.data[buffer.pos++];
    
        while (buffer.data[buffer.pos] != '\"') {
            switch (buffer.data[buffer.pos]) {
            case '\\':
                buffer.pos = start;
                return getSpecialStringLiteral();
            case '\n':
            case '\r':
                throw new PositionedError(getTokenReference(),
                                          CompilerMessages.ILLEGAL_CHAR,
                                          "\\n");
            default:
                buffer.pos++;
            }
        }
        char closequote = buffer.data[buffer.pos++];
        if (openquote != '\"' || closequote != '\"') {
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR, "\"");
        }
        return new CToken(KjcTokenTypes.STRING_LITERAL,
                          String.valueOf(buffer.data, start + 1,
                                         buffer.pos - start - 2));
    }

    /**
     * Reads a string literal containing escape sequences and thus needs
     * a temporary buffer.
     */
    private Token getSpecialStringLiteral()
        throws PositionedError
    {
        StringBuffer strbuf = new StringBuffer();
        char  openquote = buffer.data[buffer.pos++];
    
        while (buffer.data[buffer.pos] != '\"') {
            switch (buffer.data[buffer.pos]) {
            case '\\':
                strbuf.append(scanEscapeSequence());
                break;
            case '\n':
            case '\r':
                throw new PositionedError(getTokenReference(),
                                          CompilerMessages.ILLEGAL_CHAR,
                                          "\\n");
            default:
                strbuf.append(buffer.data[buffer.pos++]);
                break;
            }
        }
        char closequote = buffer.data[buffer.pos++];
        if (openquote != '\"' || closequote != '\"') {
            throw new PositionedError(getTokenReference(),
                                      CompilerMessages.ILLEGAL_CHAR, "\"");
        }
    
        return new CToken(KjcTokenTypes.STRING_LITERAL, strbuf.toString());
    }
  
    // OPERATORS
    private static final
        CToken ASSIGN = new CToken(KjcTokenTypes.ASSIGN, "=");
    private static final
        CToken BAND = new CToken(KjcTokenTypes.BAND, "&");
    private static final
        CToken BAND_ASSIGN = new CToken(KjcTokenTypes.BAND_ASSIGN, "&=");
    private static final
        CToken BNOT = new CToken(KjcTokenTypes.BNOT, "~");
    private static final
        CToken BOR = new CToken(KjcTokenTypes.BOR, "|");
    private static final
        CToken BOR_ASSIGN = new CToken(KjcTokenTypes.BOR_ASSIGN, "|=");
    private static final
        CToken BSR = new CToken(KjcTokenTypes.BSR, ">>>");
    private static final
        CToken BSR_ASSIGN = new CToken(KjcTokenTypes.BSR_ASSIGN, ">>>=");
    private static final
        CToken BXOR = new CToken(KjcTokenTypes.BXOR, "^");
    private static final
        CToken BXOR_ASSIGN = new CToken(KjcTokenTypes.BXOR_ASSIGN, "^=");
    private static final
        CToken COLON = new CToken(KjcTokenTypes.COLON, ":");
    private static final
        CToken COMMA = new CToken(KjcTokenTypes.COMMA, ",");
    private static final
        CToken DEC = new CToken(KjcTokenTypes.DEC, "--");
    private static final
        CToken DOT = new CToken(KjcTokenTypes.DOT, ".");
    private static final
        CToken EQUAL = new CToken(KjcTokenTypes.EQUAL, "==");
    private static final
        CToken GE = new CToken(KjcTokenTypes.GE, ">=");
    private static final
        CToken GT = new CToken(KjcTokenTypes.GT, ">");
    private static final
        CToken INC = new CToken(KjcTokenTypes.INC, "++");
    private static final
        CToken LAND = new CToken(KjcTokenTypes.LAND, "&&");
    private static final
        CToken LBRACK = new CToken(KjcTokenTypes.LBRACK, "[");
    private static final
        CToken LCURLY = new CToken(KjcTokenTypes.LCURLY, "{");
    private static final
        CToken LE = new CToken(KjcTokenTypes.LE, "<=");
    private static final
        CToken LNOT = new CToken(KjcTokenTypes.LNOT, "!");
    private static final
        CToken LOR = new CToken(KjcTokenTypes.LOR, "||");
    private static final
        CToken LPAREN = new CToken(KjcTokenTypes.LPAREN, "(");
    private static final
        CToken LT = new CToken(KjcTokenTypes.LT, "<");
    private static final
        CToken MINUS = new CToken(KjcTokenTypes.MINUS, "-");
    private static final
        CToken MINUS_ASSIGN = new CToken(KjcTokenTypes.MINUS_ASSIGN, "-=");
    private static final
        CToken NOT_EQUAL = new CToken(KjcTokenTypes.NOT_EQUAL, "!=");
    private static final
        CToken PERCENT = new CToken(KjcTokenTypes.PERCENT, "%");
    private static final
        CToken PERCENT_ASSIGN = new CToken(KjcTokenTypes.PERCENT_ASSIGN, "%=");
    private static final
        CToken PLUS = new CToken(KjcTokenTypes.PLUS, "+");
    private static final 
        CToken PLUS_ASSIGN = new CToken(KjcTokenTypes.PLUS_ASSIGN, "+=");
    private static final
        CToken QUESTION = new CToken(KjcTokenTypes.QUESTION, "?");
    private static final
        CToken RBRACK = new CToken(KjcTokenTypes.RBRACK, "]");
    private static final
        CToken RCURLY = new CToken(KjcTokenTypes.RCURLY, "}");
    private static final
        CToken RPAREN = new CToken(KjcTokenTypes.RPAREN, ")");
    private static final
        CToken SEMI = new CToken(KjcTokenTypes.SEMI, ";");
    private static final
        CToken SL = new CToken(KjcTokenTypes.SL, "<<");
    private static final
        CToken SLASH = new CToken(KjcTokenTypes.SLASH, "/");
    private static final
        CToken SLASH_ASSIGN = new CToken(KjcTokenTypes.SLASH_ASSIGN, "/=");
    private static final
        CToken SL_ASSIGN = new CToken(KjcTokenTypes.SL_ASSIGN, "<<=");
    private static final
        CToken SR = new CToken(KjcTokenTypes.SR, ">>");
    private static final
        CToken SR_ASSIGN = new CToken(KjcTokenTypes.SR_ASSIGN, ">>=");
    private static final
        CToken STAR = new CToken(KjcTokenTypes.STAR, "*");
    private static final
        CToken STAR_ASSIGN = new CToken(KjcTokenTypes.STAR_ASSIGN, "*=");
}
