/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import at.dms.classfile.ClassFileFormatException;
import at.dms.compiler.Compiler;
import at.dms.compiler.CompilerMessages;
import at.dms.compiler.CWarning;
import at.dms.compiler.InputBuffer;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.WarningFilter;

/**
 * The entry point of the compiler, containing the <code>main</code> method.
 */
public class Main
    extends Compiler
{
    /**
     * Files to compiler.
     */
    protected List infiles = new ArrayList();

    /**
     * Was there at least one error.
     */
    protected boolean errorFound;

    /**
     * Command-line options.
     */
    protected KjcOptions options;

    /**
     * Warning filter.
     */
    private WarningFilter filter = new DefaultFilter();

    /**
     * All generated classes.
     */
    private List classes = new ArrayList(100);

    /**
     * Runs a compilation session.
     *
     * @param args
     * The command line arguments.
     */
    public boolean run(String[] args)
    {
        if (!parseArguments(args)) {
            return false;
        }

        setDebug(options.debug);

        if (options.Xstdout) {
            System.setErr(System.out);
            setErr(System.out);
        }

        initialize();

        if (infiles.size() == 0) {
            options.usage();
            inform(KjcMessages.NO_INPUT_FILES);
            return false;
        }

        if (verboseMode()) {
            inform(CompilerMessages.COMPILATION_STARTED,
                   new Integer(infiles.size()));
        }

        try {
            infiles = verifyFiles(infiles);
        }
        catch (UnpositionedError e) {
            inform(e);
            return false;
        }

        options.destination = checkDestination(options.destination);

        CTopLevel.initSession(this);
        JCompilationUnit[] tree = new JCompilationUnit[infiles.size()];

        if (options.proc > tree.length) {
            options.proc = tree.length;
        }

        if (options.multi) {
            parseMultiProc(tree);
        }
        else {
            for (int count = 0; count < tree.length; count++) {
                tree[count] = parseFile((File)infiles.get(count));
                if (options.debug) {
                    if (tree[count] == null) {
                        System.err.println("Parsefile failed for: "
                                           + infiles.get(count));
                    }
                }
            }
        }

        infiles = null;

        if (errorFound) {
            return false;
        }

        if (!options.beautify) {
            if (!doChecksAndTransform(tree)) {
                return false;
            }
        }

        if (!options.nowrite) {
            if (options.java || options.beautify) {
                if (options.multi) {
                    acceptMultiProc(tree);
                }
                else {
                    for (int count = 0; count < tree.length; count++) {
                        tree[count].accept(this, options.destination);
                        tree[count] = null;
                    }
                }
            }
            else {
                genCode(options.destination);
            }
        }

        if (verboseMode()) {
            inform(CompilerMessages.COMPILATION_ENDED);
        }

        CodeSequence.endSession();

        return true;
    }

    private boolean doChecksAndTransform(JCompilationUnit[] tree)
    {
        if (!doChecks(tree)) {
            return false;
        }

        for (int count = 0; count < tree.length; count++) {
            transformTypes(tree[count]);
        }

        if (!doChecksOnTransformedClasses(tree)) {
            return false;
        }

        if (!options.java && !options.beautify) {
            for (int count = 0; count < tree.length; count++) {
                tree[count] = null;
            }
        }
        return true;
    }

    private boolean doChecks(JCompilationUnit[] tree)
    {
        long lastTime = System.currentTimeMillis();
        for (int count = 0; count < tree.length; count++) {
            checkInterface(tree[count]);
        }

        if (verboseMode()) {
            inform(CompilerMessages.INTERFACES_CHECKED,
                   new Long(System.currentTimeMillis() - lastTime));
        }
        
        if (errorFound) {
            return false;
        }
        
        for (int count = 0; count < tree.length; count++) {
            checkInitializers(tree[count]);
        }
        
        if (errorFound) {
            return false;
        }
       
        if (options.multi) {
            checkBodyMultiProc(tree);
        }
        else {
            for (int count = 0; count < tree.length; count++) {
                checkBody(tree[count]);
            }
        }
        
        if (errorFound) {
            return false;
        }

        return true;
    }

    private boolean doChecksOnTransformedClasses(JCompilationUnit[] tree)
    {
        for (int count = 0; count < tree.length; count++) {
            checkTransformedInterfaces(tree[count]);
        }

        for (int count = 0; count < tree.length; count++) {
            checkTransformedTypesBody(tree[count]);
        }
        
        return !errorFound;
    }

    /**
     * Parse the argument list.
     */
    public boolean parseArguments(String[] args)
    {
        options = new KjcOptions();
        return options.parseCommandLine(args, infiles);
    }

    /**
     * Generates the code from an array of compilation unit and
     * a destination
     *
     * @param destination the directory where to write classfiles
     */
    public void genCode(String destination)
    {
        CSourceClass[] classes = getClasses();

        this.classes.clear();

        int count = 0;
        try {
            if (options.multi) {
                genCodeMultiProc(classes, destination);
            }
            else {
                for (count = 0; count < classes.length; count++) {
                    long lastTime = System.currentTimeMillis();

                    classes[count].genCode(destination);
                    if (verboseMode() && !classes[count].isInnerClass()) {
                        inform(CompilerMessages.CLASSFILE_GENERATED,
                               classes[count].getQualifiedName().replace('/',
                                                                         '.'),
                               new Long(System.currentTimeMillis()-lastTime));
                    }
                    classes[count] = null;
                }
            }
        }
        catch (ClassFileFormatException e) {
            if (options.debug) {
                System.err.println("count = " + count);
                e.printStackTrace();
            }
            inform(e);
        }
        catch (IOException e) {
            inform(e);
        }
    }

    /**
     * Initialize the compiler (read classpath, check classes.zip)
     */
    protected void initialize()
    {
        if (options.classpath == null) {
            options.classpath = System.getProperty("java.class.path");
            if (options.classpath == null) {
                options.classpath = ".";
            }
        }
        /* Initialize classpath.
         * Take note of the special nosystemlib option that allows you to
         * compile alternate system libraries like classpath and libgcj.
         */
        at.dms.classfile.ClassPath.init(options.classpath,
                                        !options.nosystemlib);
        CStdType.init(this);
    }

    /**
     * @return true if compilation in verbose mode
     */
    public boolean verboseMode()
    {
        return options.verbose;
    }

    /**
     * @return the optimization level
     */
    public int optimizeCode()
    {
        return options.optimize;
    }

    /**
     * Shows an error/warning and set error flag to true
     */
    public KjcPrettyPrinter getPrettyPrinter(String fileName)
    {
        return new KjcPrettyPrinter(fileName);
    }

    /**
     * Entry point
     *
     * @param args  the command line arguments
     */
    public static void main(String[] args)
    {
        String repeatProperty = System.getProperty("repeat");
        boolean success;

        if (repeatProperty != null) {
            int repeat = Integer.parseInt(repeatProperty);
       
            success = false;
            while (repeat > 0) {
                long start = System.currentTimeMillis();
                success = compile(args);
                long duration = System.currentTimeMillis() - start;
                System.err.println("Duration " + repeat + ": " + duration
                                   + " ms.");
                repeat--;
            }
        }
        else {
            success = compile(args);
        }

        System.exit(success ? 0 : 1);
    }

    /**
     * Second entry point
     */
    public static boolean compile(String[] args)
    {
        return new Main().run(args);
    }

    /**
     * parse the given file and return a compilation unit.
     * side effect: increment error number.
     * @param file  the name of the file. (assert exists)
     * @return the compilation unit defined by this file.
     */
    protected JCompilationUnit parseFile(File file)
    {
        InputBuffer buffer;

        try {
            buffer = new InputBuffer(file);
        }
        catch (IOException e) {
            inform(e);
            return null;
        }

	KjcScanner scanner;
        KjcParser  parser; 
        JCompilationUnit unit;
        long  lastTime = System.currentTimeMillis();

	scanner = new KjcScanner(this, buffer);
        parser = new KjcParser(this, scanner);

        try {
            unit = parser.jCompilationUnit();
        }
        catch (antlr.RecognitionException e) {
            if (options.debug) {
                e.printStackTrace();
            }
            inform(parser.beautifyParseError(e));
            errorFound = true;
            unit = null;
        }
        catch (Exception e) {
            if (options.debug) {
                e.printStackTrace();
            }
            errorFound = true;
            unit = null;
        }

        if (verboseMode()) {
            inform(CompilerMessages.FILE_PARSED,
                   file.getPath(),
                   new Long(System.currentTimeMillis() - lastTime));
        }

        buffer.release();

        return unit;
    }

    /**
     * check that interface of the transformed types in a given compilation
     * unit is correct
     * side effect: increment error number
     * @param cunit  the compilation unit
     */
    protected void checkTransformedInterfaces(JCompilationUnit cunit)
    {
        try {
            cunit.checkTransformedInterfaces(this);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }
    }

    /**
     * check that interface of a given compilation unit is correct
     * side effect: increment error number
     * @param cunit  the compilation unit
     */
    protected void checkInterface(JCompilationUnit cunit)
    {
        try {
            cunit.checkInterface(this);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }
    }

    /**
     * Transform classes before code generation.
     * side effect: increment error number
     * @param cunit  the compilation unit
     */
    protected void transformTypes(JCompilationUnit cunit)
    {
        try {
            if (cunit == null) {
                throw new NullPointerException("cunit?!?");
            }
            cunit.transformTypes(this);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }
    }

    /**
     * check that interface of a given compilation unit is correct
     * side effect: increment error number
     * @param cunit  the compilation unit
     */
    protected void checkInitializers(JCompilationUnit cunit)
    {
        try {
            cunit.checkInitializers(this, classes);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }
    }

    /**
     * check that body of a given compilation unit is correct
     * side effect: increment error number
     * @param cunit  the compilation unit
     */
    protected void checkBody(JCompilationUnit cunit)
    {
        long lastTime = System.currentTimeMillis();
        try {
            cunit.checkBody(this, classes);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }

        if (verboseMode()) {
            inform(CompilerMessages.BODY_CHECKED, cunit.getFileName(),
                   new Long(System.currentTimeMillis() - lastTime));
        }
    }

    /**
     * check that body of a given compilation unit is correct
     * side effect: increment error number
     * @param cunit  the compilation unit
     */
    protected void checkTransformedTypesBody(JCompilationUnit cunit)
    {
        long lastTime = System.currentTimeMillis();
        try {
            cunit.checkTransformedTypesBody(this, classes);
        }
        catch (PositionedError e) {
            reportTrouble(e);
        }

        if (verboseMode()) {
            inform(CompilerMessages.BODY_CHECKED,
                   cunit.getFileName(),
                   new Long(System.currentTimeMillis() - lastTime));
        }
    }

    /**
     * generate the source code of parsed compilation unit
     * @param cunit  the compilation unit
     */
    protected void generateJavaCode(JCompilationUnit cunit)
    {
        long lastTime = System.currentTimeMillis();
    
        cunit.accept(this, options.destination);
        if (verboseMode()) {
            inform(CompilerMessages.JAVA_CODE_GENERATED,
                   cunit.getFileName(),
                   new Long(System.currentTimeMillis() - lastTime));
        }
    }

    /**
     * Parse each file in multi thread
     */
    protected void parseMultiProc(final JCompilationUnit[] tree)
    {
        try {
            Thread[]  parsers = new Thread[options.proc];
            int  length = tree.length / options.proc;

            for (int i = 0; i < options.proc; i++) {
                parsers[i] = new ThreadedParser(this,
                                                infiles,
                                                tree, i * length,
                                                ( i == options.proc - 1
                                                  ? tree.length
                                                  : (i + 1) * length));
                parsers[i].start();
            }

            for (int i = 0; i < options.proc; i++) {
                parsers[i].join();
            }
        }
        catch (java.lang.InterruptedException ie) {
            if (options.debug) {
                ie.printStackTrace();
            }
        }
    }

    /**
     *
     */
    protected void checkBodyMultiProc(final JCompilationUnit[] tree)
    {
        try {
            Thread[]  checkers = new Thread[tree.length];
            int  length = tree.length / options.proc;

            for (int i = 0; i < options.proc; i++) {
                checkers[i] = new ThreadedChecker(this,
                                                  tree, i * length,
                                                  (i == options.proc - 1
                                                   ? tree.length
                                                   : (i + 1) * length));
                checkers[i].start();
            }

            for (int i = 0; i < options.proc; i++) {
                checkers[i].join();
            }
        }
        catch (java.lang.InterruptedException ie) {
            if (options.debug) {
                ie.printStackTrace();
            }
        }
    }

    /**
     *
     */
    protected void acceptMultiProc(final JCompilationUnit[] tree)
    {
        try {
            Thread[]  writers = new Thread[tree.length];
            int  length = tree.length / options.proc;

            for (int i = 0; i < options.proc; i++) {
                writers[i] = new ThreadedVisitor(options.destination,
                                                 tree, i * length,
                                                 (i == options.proc - 1
                                                  ? tree.length
                                                  : (i + 1) * length),
                                                 this);
                writers[i].start();
            }

            for (int i = 0; i < options.proc; i++) {
                writers[i].join();
            }
        }
        catch (java.lang.InterruptedException ie) {
            if (options.debug) {
                ie.printStackTrace();
            }
        }
    }

    /**
     *
     */
    protected void genCodeMultiProc(final CSourceClass[] tree,
                                    final String destination)
    {
        try {
            Thread[]  writers = new Thread[tree.length];
            int  length = tree.length / options.proc;

            for (int i = 0; i < options.proc; i++) {
                writers[i] = new ThreadedGenerator(tree, destination,
                                                   i * length,
                                                   (i == options.proc - 1
                                                    ? tree.length
                                                    : (i + 1) * length));
                writers[i].start();
            }

            for (int i = 0; i < options.proc; i++) {
                writers[i].join();
            }
        }
        catch (java.lang.InterruptedException ie) {
            if (options.debug) {
                ie.printStackTrace();
            }
        }
    }

    /**
     * Reports a trouble (error or warning).
     *
     * @param trouble  a description of the trouble to report.
     */
    public void reportTrouble(PositionedError trouble)
    {
        if (trouble instanceof CWarning) {
            if (options.warning != 0 && filterWarning((CWarning)trouble)) {
                inform(trouble);
            }
        }
        else {
            if (trouble.getTokenReference() != TokenReference.NO_REF) {
                inform(trouble);
                errorFound = true;
            }
        }
    }

    protected boolean filterWarning(CWarning warning)
    {
        WarningFilter filter = getFilter();
        int   value = filter.filter(warning);

        switch (value) {
        case WarningFilter.FLT_REJECT:
            return false;
        case WarningFilter.FLT_FORCE:
            return true;
        case WarningFilter.FLT_ACCEPT:
            return warning.getSeverityLevel() <= options.warning;
        default:
            throw new RuntimeException("INTERNAL ERROR");
        }
    }

    protected WarningFilter getFilter()
    {
        if (filter == null) {
            if (options.filter != null) {
                try {
                    filter = (WarningFilter)Class.forName(options.filter)
                        .newInstance();
                }
                catch (Exception e) {
                    inform(KjcMessages.FILTER_NOT_FOUND, options.filter);
                }
            }
            if (filter == null) {
                filter = new DefaultFilter();
            }
        }

        return filter;
    }

    /**
     * Returns true iff comments should be parsed (false if to be skipped)
     */
    public boolean parseComments()
    {
        return options.deprecation || (options.beautify && !options.nowrite);
    }

    /**
     * Returns the classes to generate
     */
    public CSourceClass[] getClasses()
    {
        CSourceClass[] array = new CSourceClass[classes.size()];
        return (CSourceClass[])classes.toArray(array);
    }

    /**
     * Helper class that implements multithreaded parsing.
     */
    static class ThreadedParser extends Thread 
    {
        private Main compiler;
        private int start;
        private int end;
        private JCompilationUnit[] tree;
        private List infiles;

        ThreadedParser(Main compiler,
                       List infiles,
                       JCompilationUnit[] tree,
                       int start, int end)
        {
            this.compiler = compiler;
            this.start = start;
            this.end = end;
            this.tree = tree;
            this.infiles = infiles;
        }

        public void run()
        {
            for (int count = start; count < end; count++) {
                tree[count] = compiler.parseFile((File)infiles.get(count));
            }
        }
    }

    static class ThreadedChecker extends Thread
    {
        private Main   compiler;
        private int   start;
        private int   end;
        private JCompilationUnit[] tree;

        ThreadedChecker(Main compiler,
                        JCompilationUnit[] tree,
                        int start, int end)
        {
            this.compiler = compiler;
            this.start = start;
            this.end = end;
            this.tree = tree;
        }

        public void run()
        {
            for (int count = start; count < end; count++) {
                compiler.checkBody(tree[count]);
                tree[count] = null;
            }
        }
    }

    static class ThreadedVisitor extends Thread
    {
        private int start;
        private int end;
        private JCompilationUnit[] tree;
        private String destination;
        private Main compiler;

        ThreadedVisitor(String destination,
                        JCompilationUnit[] tree,
                        int start, int end,
                        Main compiler)
        {
            this.start = start;
            this.end = end;
            this.tree = tree;
            this.destination = destination;
            this.compiler = compiler;
        }

        public void run()
        {
            for (int count = start; count < end; count++) {
                tree[count].accept(compiler, destination);
            }
        }
    }

    static class ThreadedGenerator extends Thread
    {
        private int start;
        private int end;
        private CSourceClass[] tree;
        private String destination;

        ThreadedGenerator(CSourceClass[] tree,
                          String destination,
                          int start, int end)
        {
            this.start = start;
            this.end = end;
            this.tree = tree;
            this.destination = destination;
        }

        public void run()
        {
            try {
                for (int count = start; count < end; count++) {
                    tree[count].genCode(destination);
                    tree[count] = null;
                }
            }
            catch (IOException e) {
                System.err.println(e.getMessage());
            }
            catch (ClassFileFormatException e) {
                System.err.println(e.getMessage());
            }
        }
    }
}
