/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 * Copyright (C) 2001    Erwin Bolwidt, Haarlem, The Netherlands.
 *
 * The original author of this file is DMS; refactorings, bug fixes and
 * extensions were done by Erwin Bolwidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.kjc;

import java.util.ArrayList;
import java.util.List;

import at.dms.util.Utils;

/**
 * Parser for VM type and method signatures.
 *
 * Signatures are part of the Java Virtual Machine standard and describe the
 * parameter types and returns types of methods.
 */
public class SignatureParser
    extends Utils
{
    /**
     * Used while parsing a signature.
     */
    private List typeCache = new ArrayList();

    /**
     * Parse a java type signature.
     *
     * Attempts to parse the provided string as if it started with the Java
     * VM-standard signature for a type.
     */
    public final CType parseSignature(String signature)
    {
        return parseSignature(signature, 0, signature.length());
    }

    /**
     * Parse a java type signature from the substring of
     * <code>signature</code> that starts at index <code>from</code> and
     * ends just before <code>to</code>.
     */
    public CType parseSignature(String signature, int from, int to)
    {
        CType type;
        int cursor = from;
        while (signature.charAt(cursor) == '[') {
            cursor++;
        }

        switch (signature.charAt(cursor)) {
        case 'V':
            type = CStdType.Void;
            break;
        case 'B':
            type = CStdType.Byte;
            break;
        case 'C':
            type = CStdType.Char;
            break;
        case 'D':
            type = CStdType.Double;
            break;
        case 'F':
            type = CStdType.Float;
            break;
        case 'I':
            type = CStdType.Integer;
            break;
        case 'J':
            type = CStdType.Long;
            break;
        case 'L':
            type = CClassType.lookup(signature.substring(cursor + 1,
                                                         to - 1));
            break;
        case 'S':
            type = CStdType.Short;
            break;
        case 'Z':
            type = CStdType.Boolean;
            break;
        default:
            fail();
            return null;
        }

        cursor -= from;
        if (cursor > 0) {
            type = new CArrayType(type, cursor);
        }
        return type;
    }

    /**
     * Returns an array of types represented by the type signature
     * For methods, the return type is the last element of the array
     */
    public synchronized CType[] parseMethodSignature(String signature)
    {
        // assert(sig.charAt(0) == '(');
        char[] sig = signature.toCharArray();
        int current = 1;

        while (sig[current] != ')') {
            int end = current;
            while (sig[end] == '[') {
                end += 1;
            }
            if (sig[end] != 'L') {
                end += 1;
            } else {
                while (sig[end] != ';') {
                    end += 1;
                }
                end += 1;
            }
            typeCache.add(parseSignature(signature, current, end));
            current = end;
        }
        typeCache.add(parseSignature(signature, current + 1,
                                     signature.length()));

        // store into array
        CType[] arr = new CType[typeCache.size()];

        arr = (CType[])typeCache.toArray(arr);
        typeCache.clear();

        return arr;
    }
}
