/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package at.dms.ksm;

import antlr.RecognitionException;
import at.dms.compiler.TokenReference;
import at.dms.util.Message;
import at.dms.util.MessageDescription;

/**
 * Error thrown on problems encountered while running the assembler itself.
 */
public class KsmError
    extends RecognitionException
{
    /**
     * Location in source where the error happened.
     */
    protected TokenReference where;

    /**
     * Error message.
     */
    protected Message message;

    /**
     * An error with a formatted message as argument.
     *
     * @param where       the reference to token where error happened.
     * @param message     the formatted message.
     */
    public KsmError(TokenReference where, Message message)
    {
        super(message.getDescription().getFormat());
        this.where = where;
        this.message = message;
    }

    /**
     * An error with with an arbitrary number of parameters.
     *
     * @param where       the reference to token where error happened.
     * @param desc        the message ident to be displayed.
     * @param params      the array of parameters.
     */
    public KsmError(TokenReference where, MessageDescription desc,
                    Object[] params)
    {
        this(where, new Message(desc, params));
    }

    /**
     * An error with two parameters
     * @param where       the reference to token where error happened.
     * @param desc        the message ident to be displayed.
     * @param param1      the first parameter.
     * @param param2      the second parameter.
     */
    public KsmError(TokenReference where, MessageDescription desc,
                    Object param1, Object param2)
    {
        this(where, desc, new Object[] { param1, param2 });
    }

    /**
     * An error with one parameter.
     *
     * @param where       the reference to token where error happened.
     * @param desc        the message ident to be displayed.
     * @param param       the parameter.
     */
    public KsmError(TokenReference where, MessageDescription desc,
                    Object param)
    {
        this(where, desc, new Object[] { param });
    }

    /**
     * An error without parameters.
     *
     * @param where       the reference to token where error happened.
     * @param desc        the message ident to be displayed.
     */
    public KsmError(TokenReference where, MessageDescription desc)
    {
        this(where, desc, null);
    }

    /**
     * ...
     */
    public MessageDescription getMessageDescription()
    {
        return message.getDescription();
    }

    /**
     * Return the string explaining the error and the token reference.
     */
    public String getErrorMessage()
    {
        if (where != null) {
            return where.getFile() + ":" + where.getLine() + ": "
                + message.getMessage();
        }
        else {
            return message.getMessage();
        }
    }

    /**
     * ...
     */
    public String getMessage()
    {
        return getErrorMessage();
    }
}
