/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package at.dms.msggen;

import java.io.PrintWriter;
import java.util.Map;

import at.dms.util.MessageDescription;

/**
 * A parsed definition of a compiler message.
 */
class MessageDefinition
{
    /**
     * A Java identifier for this message.
     */
    private final String identifier;

    /**
     * java.text.MessageFormat - formatted pattern used to generate the
     * compiler message (by the Java compiler, not here) when it is output
     * to the user.
     */
    private final String format;

    /**
     * A reference to the Java Language Specification for this message, or
     * if this is not a message related to the structure of the java program,
     * it may be a reference to some other document.
     */
    private final String jlsReference;

    /**
     * Warning or error level of the message.
     */
    private final int level;

    /**
     * Constructs a message definition.
     *
     * @param identifier
     * The java identifier that is used by the compiler to refer to this
     * message.
     *
     * @param format
     * The textual message format (with placeholders)
     *
     * @param jlsReference
     * The document describing the reason for this message
     *
     * @param level
     * The severity level of this message 
     *
     */
    public MessageDefinition(String identifier, String format,
                             String jlsReference, int level)
    {
        this.identifier = identifier;

        /* strip leading and trailing quotes */
        if (jlsReference != null) {
            this.jlsReference =
                jlsReference.substring(1, jlsReference.length() - 1);
        }
        else {
            this.jlsReference = null;
        }

        if (format != null) {
            this.format = format.substring(1, format.length() - 1);
        }
        else {
            throw new NullPointerException("format");
        }

        this.level = level;
    }

    /**
     * Checks for duplicate identifiers.
     * @param  identifiers a table of all token identifiers
     * @param  prefix      the literal prefix
     * @param  id          the id of the token
     * @param  sourceFile  the file where the token is defined
     */
    public void checkIdentifiers(Map identifiers, String sourceFile)
        throws MsggenError
    {
        String stored = (String)identifiers.get(identifier);

        if (stored != null) {
            throw new MsggenError(MsggenMessages.DUPLICATE_DEFINITION, 
                                  new Object[] { identifier, sourceFile,
                                                 stored });
        }
        identifiers.put(identifier, sourceFile);
    }

    /**
     * Prints the token definition to interface file (java)
     * @param   out     the output stream
     * @param   prefix      the literal prefix
     */
    public void printInterface(PrintWriter out, String prefix)
    {
        String referenceText = "null";
        if (jlsReference != null) {
            referenceText = "\"" + jlsReference + "\"";
        }

        out.println("    /**");
        out.println("     * Message of level '"
                    +       MessageDescription.getLevelText(level)
                    +       "': <i>" + format + "</i>");
        out.println("     */");
        out.println("    public static final MessageDescription "
                    +    identifier + " =");
        out.println("        new MessageDescription(\"" + format + "\", "
                    +        referenceText + ", " + level + ");");
        out.println();
    }
}
